#include <linux/fs.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/mutex.h>	
#include <linux/poll.h>	
#include <linux/kref.h>
#include <linux/cdev.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "host.h"
#include "sgs.h"
#include "sgt.h"
#include "immap_85xx.h"

#ifndef IRQF_SHARED 
#define IRQF_SHARED SA_SHIRQ
#endif

#define DRIVER_VERSION		"v3.0"
#define DRIVER_AUTHOR		"Stefani Seibold <stefani@seibold.net>"
#define DRIVER_DESC		"SGS Host Device Driver"
#define DRIVER_LICENSE		"GPL"

static int device_major;

struct range_t {
	unsigned long addr;
	unsigned long offset;
	unsigned long size;
};

static struct pci_device_id sg_idtab[] = {
	{
		.vendor =	SGS_PCI_VENDOR_ID,
		.device =	SGS_PCI_DEVICE_ID1,
		.subvendor =	SGS_VENDOR_SUB,
		.subdevice =	SGS_PRODUCT_SUB,
		.driver_data =	(unsigned long)&sgs_device,
	},
	{
		.vendor =	SGS_PCI_VENDOR_ID,
		.device =	SGS_PCI_DEVICE_ID2,
		.subvendor =	SGS_VENDOR_SUB,
		.subdevice =	SGS_PRODUCT_SUB,
		.driver_data =	(unsigned long)&sgs_device,
	},
	{
		.vendor =	SGS_PCI_VENDOR_ID,
		.device =	SGS_PCI_DEVICE_ID1,
		.subvendor =	SGS_VENDOR_SUB,
		.subdevice =	SGS_PRODUCT_SUB_OLD,
		.driver_data =	(unsigned long)&sgs_device_old,
	},
	{
		.vendor=	SGS_PCI_VENDOR_ID,
		.device=	SGS_PCI_DEVICE_ID2,
		.subvendor=	SGS_VENDOR_SUB,
		.subdevice=	SGS_PRODUCT_SUB_OLD,
		.driver_data =	(unsigned long)&sgs_device_old,
	},
	{
		.vendor=	SGT_PCI_VENDOR_ID,
		.device=	SGT_PCI_DEVICE_ID,
		.subvendor=	SGS_VENDOR_SUB,
		.subdevice=	SGT_PRODUCT_SUB,
		.driver_data =	(unsigned long)&sgt_device,
	},
	{
		0,
	}
};

MODULE_DEVICE_TABLE(pci, sg_idtab);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
static struct class_simple *sg_class;
#else
static struct class *sg_class;
#endif

static struct file_operations fops;

static struct devdata * devices[64];
static spinlock_t	drv_lock;

int debug;

module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Debug enabled or not");

static inline struct range_t *range_setup(struct range_t *range, unsigned long addr, unsigned long offset, unsigned long size)
{
	range->addr = addr;
	range->offset = offset;
	range->size = size;
	return range;
}

static unsigned long range_fix(struct range_t *range, unsigned long offset, unsigned long size)
{
	if (!((offset >= range->offset) && ((offset < range->offset + range->size))))
		return 0;

	if (range->size > offset + size - range->offset)
		range->size = offset + size - range->offset;
	if (range->offset < offset) {
		unsigned long fix;

		fix = range->offset - offset;
		range->size -= fix;
		range->offset += fix;
		range->addr += fix;
	}
	return range->size;
}

static void set_devdata_minor(unsigned int minor, struct devdata * devdata)
{
	unsigned long flags;

	spin_lock_irqsave(&drv_lock, flags);
	devices[minor] = devdata;
	spin_unlock_irqrestore(&drv_lock, flags);
}

static int is_removed(struct devdata * devdata)
{
	unsigned long flags;
	int ret;

	spin_lock_irqsave(&drv_lock, flags);
	ret = !devices[devdata->minor];
	spin_unlock_irqrestore(&drv_lock, flags);
	return ret;
}

static void sg_delete(struct kref *kref)
{
	struct devdata * devdata = container_of(kref, struct devdata, kref);

	if (!devdata)
		return;

	iounmap(devdata->fpgamsg);
	iounmap(devdata->pciexp);
	iounmap(devdata->iomem);
	kfree(devdata);
}

static void sg_remove(struct pci_dev *dev)
{
	struct devdata * devdata = (struct devdata *)pci_get_drvdata(dev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
	class_simple_device_remove(MKDEV(device_major, devdata->minor));
#else
	device_destroy(sg_class, MKDEV(device_major, devdata->minor));
#endif

	set_devdata_minor(devdata->minor, NULL);

	free_irq(dev->irq, devdata);

	//sg_dev->reset(devdata); // Wenn stecker gezugen, dann ist das device nicht mehr da!

	pci_set_drvdata(dev, NULL);

	pci_disable_device(dev);

	pci_release_regions(dev);

	wake_up_all(&devdata->read_wq);
	wake_up_all(&devdata->write_wq);
	wake_up_all(&devdata->io_wq);
	cdev_del(devdata->sgsCdev);
	kref_put(&devdata->kref, sg_delete);
}

static int sg_probe(struct pci_dev *dev, const struct pci_device_id *id)
{
	struct devdata *devdata;
	unsigned int minor;
	int ret = -ENOMEM;
	struct sg_dev * sg_dev = (struct sg_dev *)id->driver_data;

	PRINTK_DEBUG("%s:%d\n", __FUNCTION__, __LINE__);

	for (minor = 0; devices[minor]; minor++) {
		if (minor >= ARRAY_SIZE(devices)) {
			PRINTK_ERR("cannot allocate a minor number");
			return -ENXIO;
		}
	}

	devdata = kzalloc(sizeof(*devdata), GFP_KERNEL);

	if (!devdata) {
		PRINTK_ERR("cannot allocate dev memory");
		return -ENOMEM;
	}

	mutex_init(&devdata->read_mutex);
	mutex_init(&devdata->write_mutex);
	spin_lock_init(&devdata->lock);
	init_waitqueue_head(&devdata->read_wq);
	init_waitqueue_head(&devdata->write_wq);
	init_waitqueue_head(&devdata->io_wq);

	devdata->dev = dev;
	devdata->minor = minor;
	devdata->linked = -1;
	devdata->sg_dev = sg_dev;

	ret = -ENOMEM;

	ret = pci_request_regions(dev, DEVICE_NAME);
	if (ret < 0) {
		PRINTK_ERR("Memory address conflict for device");
		goto err1;
	}

	if (id->subdevice == SGT_PRODUCT_SUB ) {
	    devdata->iobar = pci_resource_start(dev, 1);
        devdata->iomem = ioremap_nocache(devdata->iobar, pci_resource_len(dev, 1));
	}
	else {
        devdata->iobar = pci_resource_start(dev, 0);
        devdata->iomem = ioremap_nocache(devdata->iobar, pci_resource_len(dev, 0));
	}
	if (devdata->iomem == NULL) {
		PRINTK_ERR("cannot map iomem");
		goto err2;
	}

	if (id->subdevice == SGT_PRODUCT_SUB ) {
        devdata->pciexbar = pci_resource_start(dev, 0);
        devdata->pciexp = ioremap_nocache(devdata->pciexbar, pci_resource_len(dev, 0));
	}
	else {
	    devdata->pciexbar = pci_resource_start(dev, 1);
        devdata->pciexp = ioremap_nocache(devdata->pciexbar, pci_resource_len(dev, 1));
	}
	if (devdata->pciexp == NULL) {
		PRINTK_ERR("cannot map pci express EP registers");
		goto err3;
	}

	devdata->fpgabar = pci_resource_start(dev, 2);
	devdata->fpgamsg = ioremap_nocache(devdata->fpgabar, pci_resource_len(dev, 2));
	if (devdata->fpgamsg == NULL) {
		PRINTK_ERR("cannot map FPGA message register");
		goto err4;
	}


	ret = pci_enable_device(dev);
	if (ret) {
		PRINTK_ERR("cannot enable pcie device");
		goto err5;
	}

	sg_dev->reset(devdata);
	sg_dev->setup(devdata);
	
	rfifo_reset(devdata->host2dev);

	pci_set_drvdata(dev, devdata);
	pci_set_master(dev);

	ret = request_irq(dev->irq, sg_dev->isr, IRQF_SHARED, DEVICE_NAME, devdata);
	if (ret) {
		ret = request_irq(dev->irq, sg_dev->isr, IRQF_SHARED | IRQF_NO_THREAD, DEVICE_NAME, devdata);
		if (ret) {
			PRINTK_ERR("IRQ %d not free.", dev->irq);
			goto err6;
		}
	}


	kref_init(&devdata->kref);

	set_devdata_minor(minor, devdata);

	devdata->sgsCdev = cdev_alloc();
	cdev_init(devdata->sgsCdev, &fops);
	devdata->sgsCdev->owner = THIS_MODULE;
	devdata->sgsCdev->ops =  &fops;
	ret = cdev_add(devdata->sgsCdev, MKDEV(device_major, minor), 1);
	if (ret) {
		PRINTK_ERR("cdev_add failed: %d", ret);
		cdev_del(devdata->sgsCdev);
		goto err2;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
	class_simple_device_add(sg_class, MKDEV(device_major, minor), &dev->dev, DEVICE_NAME "%d", minor);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27)
	device_create(sg_class, &dev->dev, MKDEV(device_major, minor), DEVICE_NAME "%d", minor);
#else
	device_create(sg_class, &dev->dev, MKDEV(device_major, minor), devdata, DEVICE_NAME "%d", minor);
#endif
#endif


	return 0;
err6:
	pci_disable_device(dev);
err5:
	iounmap(devdata->fpgamsg);
err4:
	iounmap(devdata->pciexp);
err3:
	iounmap(devdata->iomem);
err2:
	pci_release_regions(dev);
err1:
	kfree(devdata);

	return ret;
}

static void sg_vma_open(struct vm_area_struct *vma)
{
	__module_get(THIS_MODULE);
}

static void sg_vma_close(struct vm_area_struct *vma)
{
	module_put(THIS_MODULE);
}

static struct vm_operations_struct sg_vm_ops = {
	.open =		sg_vma_open,
	.close =	sg_vma_close,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
int remap_pfn_range(struct vm_area_struct *vma, unsigned long from, unsigned long pfn, unsigned long size, pgprot_t prot)
{
	return remap_page_range(vma, from, pfn << PAGE_SHIFT, size, prot);
}
#endif

static int _sg_mmap(struct vm_area_struct *vma, unsigned long vmoff, struct range_t *range)
{
	if (!range_fix(range, vmoff, vma->vm_end - vma->vm_start))
		return 0;

	if (remap_pfn_range(vma, vma->vm_start + range->offset - vmoff, range->addr >> PAGE_SHIFT, range->size, vma->vm_page_prot))
		return -EAGAIN;

	return 0;
}

static int sg_mmap(struct file *file, struct vm_area_struct *vma)
{
	int ret;
	struct range_t range;
	struct devdata *devdata;
	unsigned long vmoff;
	devdata = (struct devdata *)file->private_data;
// check this!
	if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT))
		return -EINVAL;

	vmoff = vma->vm_pgoff << PAGE_SHIFT;
	
	if (vmoff + vma->vm_end - vma->vm_start > 3 * PAGE_SIZE)
		return -EINVAL;

	vma->vm_ops = &sg_vm_ops;
	vma->vm_flags |= VM_IO|VM_DONTEXPAND|VM_LOCKED;
	vma->vm_private_data = devdata;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	ret = _sg_mmap(vma, vmoff, range_setup(&range, devdata->iobar + offsetof(struct sgpcie_iomem, reg_data), 0, PAGE_SIZE));
	if (ret)
		return ret;

	ret = _sg_mmap(vma, vmoff, range_setup(&range, devdata->fpgabar, PAGE_SIZE * 2, PAGE_SIZE));
	if (ret)
		return ret;

	ret = _sg_mmap(vma, vmoff, range_setup(&range, devdata->iobar + offsetof(struct sgpcie_iomem, reg_data_ep), PAGE_SIZE * 3, PAGE_SIZE));
	if (ret)
		return ret;

	sg_vma_open(vma);
	return 0;
}

static ssize_t sg_read(struct file *file, char __user *buffer, size_t length, loff_t *pos)
{
	int ret;
	struct devdata *devdata;
	unsigned int copied;
	devdata = (struct devdata *)file->private_data;
	if (is_removed(devdata))
		return -ENODEV;

	mutex_lock(&devdata->read_mutex);

	while (rfifo_len(devdata->dev2host) == 0) {
		u32 irqs;

		if (file->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			goto exit;
		}

		spin_lock_irq(&devdata->lock);
		irqs = devdata->read_irqs;
		writel(rfifo_recsize(0), devdata->host_read_wait);
		spin_unlock_irq(&devdata->lock);

		mutex_unlock(&devdata->read_mutex);

		ret = wait_event_interruptible(devdata->read_wq, (irqs != devdata->read_irqs) || is_removed(devdata));
		if (ret)
			return ret;
		if (is_removed(devdata))
			return -ENODEV;

		mutex_lock(&devdata->read_mutex);
	};

	ret = rfifo_to_user(devdata->dev2host, buffer, length, &copied);
	if (ret)
		goto exit;

	if (readl(devdata->dev_write_wait)) {
		if (rfifo_unused(devdata->dev2host) >= readl(devdata->dev_write_wait))
			devdata->sg_dev->gen_inbound_mbox_irq(devdata);
	}

	ret = copied;
exit:
	mutex_unlock(&devdata->read_mutex);

	return ret;
}

static ssize_t sg_write(struct file * file, const char __user *buffer, size_t length, loff_t *pos)
{
	int ret;
	struct devdata *devdata;
	unsigned int copied;
	devdata = (struct devdata *)file->private_data;
	if (is_removed(devdata))
		return -ENODEV;

	mutex_lock(&devdata->write_mutex);

	length = rfifo_max_r(length);

	while (rfifo_avail(devdata->host2dev) < length) {
		u32 irqs;

		if (file->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			goto exit;
		}

		spin_lock_irq(&devdata->lock);
		irqs = devdata->write_irqs;
		if ((readl(devdata->host_write_wait) - 1) >= length)
			writel(rfifo_recsize(length), devdata->host_write_wait);
		spin_unlock_irq(&devdata->lock);

		mutex_unlock(&devdata->write_mutex);

		ret = wait_event_interruptible(devdata->write_wq, (irqs != devdata->write_irqs) || is_removed(devdata));
		if (ret)
			return ret;
		if (is_removed(devdata))
			return -ENODEV;

		mutex_lock(&devdata->write_mutex);
	}

	ret = rfifo_from_user(devdata->host2dev, buffer, length, &copied);
	if (ret)
		goto exit;

	if (readl(devdata->dev_read_wait)) {
		if (rfifo_len(devdata->host2dev) >= readl(devdata->dev_read_wait)){
		    //PRINTK_ERR("%s:%d\n", __FUNCTION__, __LINE__);
		    devdata->sg_dev->gen_inbound_mbox_irq(devdata);
		}

	}

	ret = copied;
exit:
	mutex_unlock(&devdata->write_mutex);

	return ret;
}

static unsigned int sg_poll(struct file *file, poll_table * wait)
{
    struct devdata *devdata;
	unsigned long flags;
	int ret = 0;

	devdata = (struct devdata *)file->private_data;
	poll_wait(file, &devdata->read_wq, wait);
	poll_wait(file, &devdata->write_wq, wait);

	spin_lock_irqsave(&devdata->lock, flags);

	if (!rfifo_is_empty(devdata->dev2host)) {
		ret |= POLLIN | POLLRDNORM;
		writel(0, devdata->host_read_wait);
	}
	else
		writel(rfifo_recsize(0), devdata->host_read_wait);

	if (rfifo_avail(devdata->host2dev) > RFIFO_BUF_SIZE / 2) {
		ret |= POLLOUT | POLLWRNORM;
		writel(0, devdata->host_write_wait);
	}
	else {
		if ((readl(devdata->host_write_wait) - 1) >= RFIFO_BUF_SIZE / 2)
			writel(RFIFO_BUF_SIZE / 2, devdata->host_write_wait);
	}

	spin_unlock_irqrestore(&devdata->lock, flags);

	return ret;
}

static int sg_link_dev(struct devdata *devdata, unsigned long arg)
{
	struct devdata *linkdev;
	struct sg_dev * sg_dev = devdata->sg_dev;

	if (!sg_dev->link)
		return -ENOTSUPP;

	if (devdata->linked != -1)
		return -EBUSY;
	if (arg >= ARRAY_SIZE(devices))
		return -ENODEV;

	linkdev = devices[arg];
	if (!linkdev)
		return -ENODEV;

	if (!linkdev->sg_dev->link)
		return -ENOTSUPP;
	if (linkdev->linked != -1)
		return -EBUSY;

	if (sg_dev != linkdev->sg_dev)
		return -EINVAL;

	linkdev->linked = devdata->minor;
	devdata->linked = linkdev->minor;

	sg_dev->link(devdata, linkdev);
	linkdev->sg_dev->link(linkdev, devdata);

	return 0;
}

static int sg_unlink_dev(struct devdata *devdata)
{
	struct devdata *linkdev;
	int linked;
	struct sg_dev * sg_dev = devdata->sg_dev;

	linked = devdata->linked;
	if (linked == -1)
		return -EINVAL;

	devdata->linked = -1;
	sg_dev->unlink(devdata);

	linkdev = devices[linked];
	if (!linkdev)
		return 0;

	linkdev->linked = -1;
	linkdev->sg_dev->unlink(linkdev);

	return 0;
}

static long sg_pcie_wait_mask(struct devdata *devdata, unsigned long timeout, u32 mask)
{
	long ret;

	if (timeout) {
		if (timeout != -1) {
			unsigned long jiffies;

			jiffies = (timeout * HZ) / 1000UL;

			ret = wait_event_interruptible_timeout(devdata->io_wq, devdata->reg_bits & mask, jiffies);
			if (!ret)
				return -ETIMEDOUT;
		}
		else
			ret = wait_event_interruptible(devdata->io_wq, devdata->reg_bits & mask);

		if (ret < 0)
			return ret;
	}

	spin_lock_irq(&devdata->lock);
	ret = devdata->reg_bits & mask;
	devdata->reg_bits &= ~mask;
	spin_unlock_irq(&devdata->lock);

	return ret;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35)
static int sg_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
#else
static long sg_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#endif
{
    struct devdata *devdata;
	int ret = 0;
	devdata = (struct devdata *)file->private_data;
	if (is_removed(devdata))
		return -ENODEV;

	switch(cmd) {
	case SGS_WAIT_IRQ:
		ret = sg_pcie_wait_mask(devdata, arg, ~0);
		break;
	case SGS_WAIT_IRQ_MASK:
	 {
		struct wait_mask_t wait_mask;

		if (copy_from_user(&wait_mask, (void *)arg, sizeof(wait_mask)))
			return -EFAULT;

		if (!wait_mask.mask)
			return -EINVAL;

		//PRINTK_ERR("Waiting for Mask 0x%08x\n", wait_mask.mask);

		ret = sg_pcie_wait_mask(devdata, wait_mask.timeout, wait_mask.mask);
		break;
	 }
	case SGS_LINK:
		spin_lock_irq(&drv_lock);
		ret = sg_link_dev(devdata, arg);
		spin_unlock_irq(&drv_lock);
		break;
	case SGS_UNLINK:
		spin_lock_irq(&drv_lock);
		ret = sg_unlink_dev(devdata);
		spin_unlock_irq(&drv_lock);
		break;
	case SGS_SET_DEVDATA:
		if (copy_from_user(&devdata->iomem->new.dev_data, (void *)arg, sizeof(devdata->iomem->new.dev_data)))
			return -EFAULT;
		break;
	case SGS_GET_DEVDATA:
		if (copy_to_user((void *)arg, &devdata->iomem->new.dev_data, sizeof(devdata->iomem->new.dev_data)))
			return -EFAULT;
		break;
	default:
		ret = -EINVAL;
		break;
	}
	return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
static long sg_compat_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35)
	return sg_ioctl(NULL, file, cmd, arg);
#else
	return sg_ioctl(file, cmd, arg);
#endif
}
#endif

static int sg_open(struct inode *inode, struct file *file)
{
	struct devdata *devdata;

	if (iminor(inode) >= ARRAY_SIZE(devices))
		return -ENODEV;

	spin_lock_irq(&drv_lock);
	devdata = devices[iminor(inode)];
	if (!devdata) {
		spin_unlock_irq(&drv_lock);
		return -ENODEV;
	}

	kref_get(&devdata->kref);
	spin_unlock_irq(&drv_lock);

	file->private_data = devdata;

	return 0;
}

static int sg_release(struct inode *inode, struct file *file)
{
    struct devdata *devdata;

	devdata = (struct devdata *)file->private_data;
	kref_put(&devdata->kref, sg_delete);

	return 0;
}

static struct file_operations fops = {
	.owner =		THIS_MODULE,	/* owner own */
	.read =			sg_read,
	.write =		sg_write,
	.poll =			sg_poll,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35)
	.ioctl =		sg_ioctl,
#else
	.unlocked_ioctl =	sg_ioctl,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
	.compat_ioctl =		sg_compat_ioctl,
#endif
	.open =			sg_open,
	.release =		sg_release,
	.mmap =			sg_mmap,
	
};

static struct pci_driver pci_sg = {
	.name =		DEVICE_NAME,
	.id_table =	sg_idtab,
	.probe =	sg_probe,
	.remove =	sg_remove,
};

static int sg_init_module(void)
{
	int ret;
	dev_t chrDev;

	spin_lock_init(&drv_lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
	sg_class = class_simple_create(THIS_MODULE, DEVICE_NAME);
#else
	sg_class = class_create(THIS_MODULE, DEVICE_NAME);
#endif
	if (IS_ERR(sg_class)) {
		PRINTK_ERR("Unable to allocate class");
		ret = PTR_ERR(sg_class);
		goto err1;
	}

	ret = alloc_chrdev_region(&chrDev, 0, 32, DEVICE_NAME);
	if (ret < 0) {
	    PRINTK_ERR("Get major device number failed.\n");
	    goto err2;
	}
	device_major = MAJOR(chrDev);
	printk(KERN_ALERT "Got major device number %d.\n",device_major);

	ret = pci_register_driver(&pci_sg);
	if (ret < 0) {
		PRINTK_ERR("pci_register_driver failed: %d", ret);
		goto err3;
	}

//	PRINTK_INFO(DRIVER_DESC " (" DRIVER_VERSION ") compiled at "__TIME__ " " __DATE__);

	return 0;

err3:
//	unregister_chrdev(device_major, DEVICE_NAME);
err2:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
	class_simple_destroy(sg_class);
#else
	class_destroy(sg_class);
#endif
err1:
	return ret;
}
module_init(sg_init_module);

#ifdef MODULE
static void sg_exit_module(void)
{
    unregister_chrdev_region(MKDEV(device_major,0), 32  );
//	unregister_chrdev(device_major, DEVICE_NAME);
	pci_unregister_driver(&pci_sg);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
	class_simple_destroy(sg_class);
#else
	class_destroy(sg_class);
#endif
}
module_exit(sg_exit_module);
#endif

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC " " DRIVER_VERSION);
MODULE_LICENSE(DRIVER_LICENSE);


