/*
 * A remote FIFO implementation
 *
 * Copyright (C) 2009/2010 Stefani Seibold <stefani@seibold.net>
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/err.h>
#include <linux/version.h>
#include <asm/io.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
#include <asm/uaccess.h>
#else
#include <linux/uaccess.h>
#endif

#include "rfifo.h"

static const size_t recsize = 2;

#if 0
static void dump(rfifo fifo, const char *str, unsigned int n)
{
	unsigned int i = 0;

	printk("%s(%p in:%u out:%u):\n", str, fifo, le32_to_cpu(fifo->in), le32_to_cpu(fifo->out));

	for(i = 0; i != n; i++) 
		printk("%02x%c", fifo->buf[i], (i & 0xf) == 0xf ? '\n' : ' ');

	if (i & 0xf)
		printk("\n");
}
#endif

unsigned int rfifo_avail(rfifo fifo)
{
	unsigned int avail;

	avail = rfifo_unused(fifo);

	if (avail < recsize)
		return 0;

	return avail - recsize;
}

unsigned int rfifo_recsize(unsigned int length)
{
	return length + recsize;
}

static inline unsigned long rfifo_copy_from_user(rfifo fifo, 
	const void __user *from, unsigned int len, unsigned int off,
	unsigned int *copied)
{
	unsigned int l;
	unsigned long ret;

	off &= RFIFO_BUF_SIZE - 1;
	l = min(len, RFIFO_BUF_SIZE - off);

	ret = copy_from_user(__io_virt(fifo->buf) + off, from, l);
	if (unlikely(ret))
		ret = ret + len - l;
	else
		ret = copy_from_user(__io_virt(fifo->buf), from + l, len - l);

	*copied = len - ret;

	return ret;
}

static inline unsigned long rfifo_copy_to_user(rfifo fifo, void __user *to,
		unsigned int len, unsigned int off, unsigned int *copied)
{
	unsigned int l;
	unsigned long ret;

	off &= RFIFO_BUF_SIZE - 1;
	l = min(len, RFIFO_BUF_SIZE - off);

	ret = copy_to_user(to, __io_virt(fifo->buf) + off, l);
	if (unlikely(ret))
		ret = ret + len - l;
	else
		ret = copy_to_user(to + l, __io_virt(fifo->buf), len - l);

	*copied = len - ret;

	return ret;
}

unsigned int rfifo_max_r(unsigned int len)
{
	const unsigned int max = (1 << (recsize << 3)) - 1;

	if (len > max)
		return max;

	return len;
}

#define	__RFIFO_PEEK(data, out, mask) \
	readb(&((data)[(out) & (mask)]))
/*
 * rfifo_peek_n internal helper function for determinate the length of
 * the next record in the fifo
 */
static inline unsigned int rfifo_peek_n(rfifo fifo, unsigned int out)
{
	unsigned int l;

	l = __RFIFO_PEEK(fifo->buf, out, RFIFO_BUF_SIZE - 1);
	l |= __RFIFO_PEEK(fifo->buf, out + 1, RFIFO_BUF_SIZE - 1) << 8;

	return l;
}

#define	__RFIFO_POKE(data, in, mask, val) \
	writeb(val, &((data)[(in) & (mask)]))

static inline void rfifo_poke_n(rfifo fifo, unsigned int in, unsigned int n)
{

	__RFIFO_POKE(fifo->buf, in, RFIFO_BUF_SIZE - 1, n);
	__RFIFO_POKE(fifo->buf, in + 1, RFIFO_BUF_SIZE - 1, n >> 8);
}

int rfifo_from_user(rfifo fifo, const void __user *from,
	unsigned long len, unsigned int *copied)
{
	unsigned long ret;
	unsigned int in;

	len = rfifo_max_r(len);

	if (len > rfifo_avail(fifo)) {
		*copied = 0;
		return -EAGAIN;
	}

	in = readl(&fifo->in);

	rfifo_poke_n(fifo, in, len);

	ret = rfifo_copy_from_user(fifo, from, len, in + recsize, copied);
	if (unlikely(ret)) {
		*copied = 0;
		return -EFAULT;
	}
	writel(in + len + recsize, &fifo->in);

	return 0;
}

int rfifo_to_user(rfifo fifo, void __user *to,
	unsigned long len, unsigned int *copied)
{
	unsigned long ret;
	unsigned int n;
	unsigned int out;

	if (rfifo_is_empty(fifo)) {
		*copied = 0;
		return -EAGAIN;
	}

	out = readl(&fifo->out);

	n = rfifo_peek_n(fifo, out);
	if (len > n)
		len = n;

	ret = rfifo_copy_to_user(fifo, to, len, out + recsize, copied);
	if (unlikely(ret)) {
		*copied = 0;
		return -EFAULT;
	}
	writel(out + n + recsize, &fifo->out);

	return 0;
}

