/*
 * A remote FIFO implementation
 *
 * Copyright (C) 2009/2010 Stefani Seibold <stefani@seibold.net>
 */

#ifndef __RFIFO_H
#define __RFIFO_H

#include <linux/kernel.h>
#include <linux/io.h>

#ifndef __io_virt
#define __io_virt(x) ((void __force *) (x))
#endif

#define	RFIFO_BUF_SIZE	(1 << 17)	/* 128 KiloByte */

#define	_align(n) __attribute__ ((aligned (n)))

struct rfifo {
	u32 _align(32)	in;
	u32 _align(32)	out;
	u8 _align(32)	buf[RFIFO_BUF_SIZE];
} _align(32);

typedef struct rfifo __iomem *rfifo;

/**
 * INIT_RFIFO - Initialize a fifo declared by DECLARE_RFIFO
 * @fifo: name of the declared fifo datatype
 */
#define INIT_RFIFO(fifo) \
(void)({ \
	rfifo rfifo = &(fifo); \
	writel(0, &rfifo->in); \
	writel(0, &rfifo->out); \
})

/**
 * DEFINE_RFIFO - macro to define and initialize a fifo
 * @fifo: name of the declared fifo datatype
 * @size: the number of elements in the fifo, this must be a power of 2
 *
 * Note: the macro can be used for global and local fifo data type variables.
 */
#define DEFINE_RFIFO(fifo, size) \
	DECLARE_RFIFO(fifo, size) = \
	(struct rfifo) { \
		.in	= 0, \
		.out	= 0, \
	}


/**
 * rfifo_reset - removes the entire fifo content
 * @fifo: address of the fifo to be used
 *
 * Note: usage of rfifo_reset() is dangerous. It should be only called when the
 * fifo is exclusived locked or when it is secured that no other thread is
 * accessing the fifo.
 */
static inline void rfifo_reset(rfifo fifo)
{
	writel(0, &fifo->in);
	writel(0, &fifo->out);
}

/**
 * rfifo_reset_out - skip fifo content
 * @fifo: address of the fifo to be used
 *
 * Note: The usage of rfifo_reset_out() is safe until it will be only called
 * from the reader thread and there is only one concurrent reader. Otherwise
 * it is dangerous and must be handled in the same way as rfifo_reset().
 */
static inline void rfifo_reset_out(rfifo fifo)
{
	writel(readl(&fifo->in), &fifo->out);
}

/**
 * rfifo_len - returns the number of used elements in the fifo
 * @fifo: address of the fifo to be used
 */
static inline unsigned int rfifo_len(rfifo fifo)
{
	return readl(&fifo->in) - readl(&fifo->out);
}

/**
 * rfifo_is_empty - returns true if the fifo is empty
 * @fifo: address of the fifo to be used
 */
static inline int rfifo_is_empty(rfifo fifo)
{
	return readl(&fifo->in) == readl(&fifo->out);
}

/**
 * rfifo_is_full - returns true if the fifo is full
 * @fifo: address of the fifo to be used
 */
static inline int rfifo_is_full(struct rfifo __iomem *fifo)
{
	return rfifo_len(fifo) >= RFIFO_BUF_SIZE;
}

/**
 * rfifo_unused - returns the number of unused elements in the fifo
 * @fifo: address of the fifo to be used
 */
static inline unsigned int rfifo_unused(rfifo fifo)
{
	return RFIFO_BUF_SIZE - rfifo_len(fifo);
}

/**
 * rfifo_avail - returns the number of unused elements in the fifo
 * @fifo: address of the fifo to be used
 */
extern unsigned int rfifo_avail(struct rfifo __iomem *fifo);

/**
 * rfifo_recsize - returns the number of bytes used for the record
 * @fifo: address of the fifo to be used
 */
extern unsigned int rfifo_recsize(unsigned int lenght);

/**
 * rfifo_max_r - returns the max. size of a record
 * @fifo: address of the fifo to be used
 */
extern unsigned int rfifo_max_r(unsigned int len);

/**
 * rfifo_from_user - puts some data from user space into the fifo
 * @fifo: address of the fifo to be used
 * @from: pointer to the data to be added
 * @len: the length of the data to be added
 * @copied: pointer to output variable to store the number of copied bytes
 *
 * This macro copies at most @len bytes from the @from into the
 * fifo, depending of the available space and returns -EFAULT/0.
 *
 * Note that with only one concurrent reader and one concurrent
 * writer, you don't need extra locking to use these macro.
 */
extern int rfifo_from_user(rfifo fifo, 
	const void __user *from, unsigned long len, unsigned int *copied);

/**
 * rfifo_to_user - copies data from the fifo into user space
 * @fifo: address of the fifo to be used
 * @to: where the data must be copied
 * @len: the size of the destination buffer
 * @copied: pointer to output variable to store the number of copied bytes
 *
 * This macro copies at most @len bytes from the fifo into the
 * @to buffer and returns -EFAULT/0.
 *
 * Note that with only one concurrent reader and one concurrent
 * writer, you don't need extra locking to use these macro.
 */
extern int rfifo_to_user(rfifo fifo, 
	void __user *to, unsigned long len, unsigned int *copied);

#endif
