/*******************************************************************************
CallLogger.c

Logging functions for instrument drivers
Wilhelm Kurz, 2020

Create an xml log from instrument driver calls

copyright    (c) Rohde & Schwarz GmbH & Co. KG, Munich
******************************************************************************/


#ifdef _WIN32
  #include <windows.h>
#endif
#include "assert.h"
#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "ctype.h"
#include "CallLogger.h"
#include "printmacros.h"

#ifdef __linux__
#include <unistd.h>
#include <pthread.h>
#include <sys/syscall.h>
#include <errno.h>
#endif


#define MAXCALLDEPTH 10
#define DEFAULTSIZE  1000

#define CURRENTCALL g_pLogger->pCurrentCall

#define LF "\n"
#define TB "  "
#define VERSION "1.0"

enum INOUT { INOUT_IN, INOUT_OUT };

typedef enum  {
  attrDataType_Bool,
  attrDataType_Uint8,
  attrDataType_Uint16,
  attrDataType_Uint32,
  attrDataType_Int8,
  attrDataType_Int16,
  attrDataType_Int32,
  attrDataType_String,
  attrDataType_Int64
} attrDataType;

typedef struct
{
  size_t size;
  char* data;
} tDynString;

typedef struct {
  const char* replace;
  const char* with;
} tFindReplace;

static tFindReplace Pairs[] =
{
  {"&", "&amp;"},
  {"\"", "&quot;"},
  {"<", "&lt;"},
  {">", "&gt;"},
};

typedef enum {
  DTYPE_ViBoolean, DTYPE_INT32, DTYPE_ATTRSTATE, DTYPE_REAL64, DTYPE_SESSION, DTYPE_STRING, DTYPE_REAL64ARR, DTYPE_INT32ARR, DTYPE_CHARARR, DTYPE_STRINGARR, DTYPE_INT64, DTYPE_UINT32,
} DTYPE;

typedef struct
{
  DTYPE       type;
  const char* name;
  void*       pValue;
  ViInt32*    pCount;
  ViSession   attr;
}
tOutPara;


typedef struct tCall
{
  tDynString    text;
  long long     microsecondsstart;
  tDynString    sInputparameter;
  tDynString    sOutputparameter;
  tDynString    sChildrenTexts;

  tOutPara      outputparameter[20];
  size_t        numoutputparameter;
  struct tCall* parent;
  struct tCall* child;
} tCall;


typedef struct
{
  size_t      numNodes;
  size_t      next;
  size_t      count;
  size_t      maxcalldepth;
  size_t      wrapstartindex;
  ViBoolean   enableWrap;
  ViBoolean   stoponerror;
  ViBoolean   stopped;
  ViBoolean   wrap;
  int         calldepth;
  char        type[64];
  DecodeErrorCallback_f DecodeErrorCallback;

  tCall**     nodes;
  tCall*      pCurrentCall;
} tLogger;

static void InitCall(tCall* pCall);
static void ConstructCall(tCall* pCall);
static void DestoyCall(tCall* pCall);
static void DestoryCallstack();
static void ConstructCallstack();
static void PopCall();
static ViBoolean PushCall();

static void ClearString(tDynString* pNode);
static void AllocString(tDynString* pNode, size_t size);
static void AppendToString(tDynString* pNode, const char* szTxt);

static void Append(const char* szTxt);
static void CollectOutput(void);
static ViBoolean DetermineProcessorClockrate(void);
static double GetClockInMicroSeconds(ViBoolean bIsToCalculate);
static double g_ProcessorClockrate = -1;
static char *str_replace(char *orig, const char *rep, const char *with);
static void EncodeChar(const char c, char* out);
static char* ReplaceXmlEntities(char* str);
static const char* Indentation(int cnt);

static tLogger* g_pLogger = NULL;
static tCall g_CallStack[MAXCALLDEPTH];
static attrDataType lookupAttributeType(unsigned int id);


ViBoolean LogLoggingEnabled(void)
{
  return g_pLogger != NULL;
}

void LogSetMaxCallDepth(ViUInt32 maxcalldepth)
{
  if (g_pLogger)
    g_pLogger->maxcalldepth = maxcalldepth;
}

void LogSetWrapStartIndex(ViUInt32 wrapstartindex)
{
  if (g_pLogger && g_pLogger->enableWrap)
  {
    g_pLogger->wrapstartindex = wrapstartindex;
    if (g_pLogger->wrapstartindex > g_pLogger->numNodes-1)
      g_pLogger->wrapstartindex = g_pLogger->numNodes-1;
  }
}

void LogSetDecodeErrorCallback(DecodeErrorCallback_f f)
{
  g_pLogger->DecodeErrorCallback = f;
}

static void LogDecodeErrorStatus(ViStatus status, const char** Name, const char** Txt)
{
  static char szName[256];
  static char szText[256];

  if (g_pLogger->DecodeErrorCallback)
  {
    if (g_pLogger->DecodeErrorCallback(status, szName, szText) == VI_SUCCESS)
    {
      *Name = szName;
      if (status != VI_SUCCESS)
        *Txt = szText;
      else
        *Txt = "";        
    }
    else
    {
      *Name = *Txt = "Unknown error code";
    }
  }
}


void LogOpenLog(const char* type, ViInt32 historydepth, ViBoolean enablewrap, ViBoolean stoponerror)
{
  LogDiscardLog();

  DetermineProcessorClockrate();

  if (historydepth < 50)
    historydepth = 50;

  g_pLogger = (tLogger*)calloc(1, sizeof(tLogger));
  strcpy(g_pLogger->type, type);
  g_pLogger->numNodes = historydepth;
  g_pLogger->enableWrap = enablewrap;
  g_pLogger->stoponerror = stoponerror;
  g_pLogger->maxcalldepth = 1;
  g_pLogger->wrapstartindex = 20;


  ConstructCallstack();

  g_pLogger->nodes = (tCall**) calloc(g_pLogger->numNodes, sizeof(tCall*));
  for (size_t i = 0; i < g_pLogger->numNodes; i++)
  {
    g_pLogger->nodes[i] = (tCall*)calloc(1, sizeof(tCall));
    if (g_pLogger->nodes[i])
      ConstructCall(g_pLogger->nodes[i]);
    else
    {
      printf("Out of memory !!!\n");
      break;
    }
  }
  g_pLogger->pCurrentCall = g_pLogger->nodes[0];
}

void LogCloseLog(void)
{
  assert(g_pLogger->calldepth == 0);
  g_pLogger->stopped = VI_TRUE;
}

void LogDiscardLog(void)
{
  if (g_pLogger)
  {
    for (size_t i = 0; i < g_pLogger->numNodes; i++)
    {
      DestoyCall(g_pLogger->nodes[i]);
      free(g_pLogger->nodes[i]);
    }
    free(g_pLogger->nodes);
    free(g_pLogger);
    g_pLogger = NULL;

    DestoryCallstack();
  }
}

ViStatus LogSaveLog(const ViString filename)
{
  ViStatus rc = VI_SUCCESS;
  if (g_pLogger)
  {
    assert(g_pLogger->calldepth == 0);
    FILE* f = fopen(filename,"w");
    if (f)
    {
      size_t iStart = 0;
      size_t cnt = g_pLogger->next;

      fprintf(f, "%s","<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
      fprintf(f, "<root version=\"" VERSION "\" type=\"%s\">" LF, g_pLogger->type);

      if (g_pLogger->wrap)
      {
        cnt = g_pLogger->numNodes - g_pLogger->wrapstartindex;
        iStart = g_pLogger->next;
      
        if (g_pLogger->enableWrap)
        {
          for (size_t i = 0; i < g_pLogger->wrapstartindex; i++)
            fprintf(f, "%s", g_pLogger->nodes[i]->text.data);

          fprintf(f, "<wrap/>");
        }
      }

      size_t idx = iStart; 
      for (size_t i = 0; i < cnt; i++)
      {
        fprintf(f,"%s",g_pLogger->nodes[idx]->text.data);
        idx++;
        if (idx >= g_pLogger->numNodes)
          idx = g_pLogger->wrapstartindex;
      }  
      fprintf(f, "%s","</root>" LF);

      fclose(f);
    }
    else
    {
      rc = VI_ERROR_FILE_ACCESS;
    }
    LogDiscardLog();
  }
  return rc;
}


static void Append(const char* szTxt)
{
  //tDynString* pNode = g_pLogger->nodes[g_pLogger->next];
  //AppendToString(pNode, szTxt);

  tDynString* pTxt = &(g_pLogger->pCurrentCall->text);
  AppendToString(pTxt, szTxt);
}


static void LogAddStatus(ViStatus status)
{
  if (g_pLogger)
  {
    char tmp[1024];
    const char* Name = NULL, *Txt = NULL;
    LogDecodeErrorStatus(status, &Name, &Txt);

    Append(Indentation(g_pLogger->calldepth +1));
    sprintf(tmp, "<status type=\"success\" code=\"0x%08x\" name=\"%s\">%s</status>" LF, (int)status, Name, Txt);
    Append(tmp);
  }
}

void LogAddCall(const char* szDescription)
{
  if (g_pLogger && !g_pLogger->stopped)
  {
    if(PushCall())
    {
      g_pLogger->count++;
      InitCall(g_pLogger->pCurrentCall);

  #ifdef __linux__
      unsigned int threadid = (unsigned)syscall(__NR_gettid);
  #else
      unsigned int threadid = GetCurrentThreadId();
  #endif

      char tmp[1024];
      sprintf(tmp,"<call cnt=\"%zd\" description=\"%s\" thread=\"0x%08x\">" LF, g_pLogger->count, szDescription, threadid);
      Append(Indentation(g_pLogger->calldepth));
      Append(tmp);
    }
  }
}

void LogCloseCall(ViStatus status)
{
  if (g_pLogger && !g_pLogger->stopped)
  {
    if (g_pLogger->calldepth <= g_pLogger->maxcalldepth) 
    {
      long long now_microseconds = (long long)GetClockInMicroSeconds(VI_FALSE);
      long duration = (long)(now_microseconds - g_pLogger->pCurrentCall->microsecondsstart);

      LogAddStatus(status);

      char tmp[1024];
      sprintf(tmp, "<input time=\"%lld\">", g_pLogger->pCurrentCall->microsecondsstart);
      Append(Indentation(g_pLogger->calldepth + 1));
      Append(tmp);

      if (g_pLogger->pCurrentCall->sInputparameter.data[0])
      {
        Append(LF);
        Append(g_pLogger->pCurrentCall->sInputparameter.data);
        Append(Indentation(g_pLogger->calldepth + 1));
      }
      Append("</input>" LF);

      Append(g_pLogger->pCurrentCall->sChildrenTexts.data);

      sprintf(tmp, "<output time=\"%lld\" duration=\"%ld\">", now_microseconds, duration);
      Append(Indentation(g_pLogger->calldepth + 1));
      Append(tmp);

      CollectOutput();

      if(g_pLogger->pCurrentCall->sOutputparameter.data[0])
      {
        Append(LF);
        Append(g_pLogger->pCurrentCall->sOutputparameter.data);
        Append(Indentation(g_pLogger->calldepth + 1));
      }
      Append("</output>" LF);
      Append(Indentation(g_pLogger->calldepth));
      Append("</call>" LF);

      if (g_pLogger->calldepth > 1)
      {
        tCall* parent = (tCall*)g_pLogger->pCurrentCall->parent;
        AppendToString(&(parent->sChildrenTexts), g_pLogger->pCurrentCall->text.data);
      }
    }


    PopCall();

    if (g_pLogger->calldepth == 0)
    {
      g_pLogger->next++;

      if (g_pLogger->next >= g_pLogger->numNodes)
      {
        g_pLogger->wrap = VI_TRUE;
        g_pLogger->next = g_pLogger->wrapstartindex;

        if (!g_pLogger->enableWrap)
          g_pLogger->stopped = VI_TRUE;
      }
      g_pLogger->pCurrentCall = g_pLogger->nodes[g_pLogger->next];

      if (g_pLogger && g_pLogger->stoponerror && (status & _VI_ERROR))
        g_pLogger->stopped = VI_TRUE;
    }
  }
}


// ############# Bool ######################
static void LogAddBoolParameter(enum INOUT dir, const char* szName, ViBoolean value)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViBoolean\">%s</parameter>" LF, szName, value ? "VI_TRUE" : "VI_FALSE");
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;

    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddBoolArg2( const char* szName, ViBoolean value)
{
  LogAddBoolParameter(INOUT_IN, szName, value);
}

void LogAddBoolOut2(const char* szName, ViBoolean* pValue)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_ViBoolean;
    para->pValue = pValue;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

// ############# Session ######################
static void LogAddSessionParameter(enum INOUT dir, const char* szName, ViSession session)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViSession\">" _PERCD "</parameter>" LF, szName, session);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddSessionArg2(const char* szName, ViSession session)
{
   LogAddSessionParameter(INOUT_IN, szName, session);
}

void LogAddSessionOut2(const char* szName, ViSession* pValue)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
 
    para->name = szName;
    para->type = DTYPE_SESSION;
    para->pValue = pValue;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

// ############# Int32 ######################
static void LogAddInt32Parameter(enum INOUT dir, const char* szName, ViInt32 value)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViInt32\">" _PERCD "</parameter>" LF, szName, value);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddInt32Arg2(const char* szName, ViInt32 value)
{
  LogAddInt32Parameter(INOUT_IN, szName, value);
}

void LogAddInt32Out2(const char* szName, ViInt32* pValue)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_INT32;
    para->pValue = pValue;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

// ############# UInt32 ######################
static void LogAddUInt32Parameter(enum INOUT dir, const char* szName, ViUInt32 value)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViUInt32\">" _PERCU "</parameter>" LF, szName, value);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddUInt32Arg2(const char* szName, ViUInt32 value)
{
  LogAddUInt32Parameter(INOUT_IN, szName, value);
}

void LogAddUInt32Out2(const char* szName, ViUInt32* pValue)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_UINT32;
    para->pValue = pValue;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

// ############# Int64 ######################
static void LogAddInt64Parameter(enum INOUT dir, const char* szName, ViInt64 value)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViInt64\">" _PERCLD "</parameter>" LF, szName, value);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddInt64Arg2(const char* szName, ViInt64 value)
{
  LogAddInt64Parameter(INOUT_IN, szName, value);
}

void LogAddInt64Out2(const char* szName, ViInt64* pValue)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_INT64;
    para->pValue = pValue;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

// ############# Select ######################
static void LogAddSelectParameter(enum INOUT dir, const char* szName, ViInt32 value, const char* selectlist)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViInt32\" selectlist=\"%s\">" _PERCD "</parameter>" LF, szName, selectlist, value);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddSelectArg2(const char* szName, ViInt32 value, const char* selectlist)
{
  LogAddSelectParameter(INOUT_IN, szName, value, selectlist);
}


// ############# AttrState ######################

static void LogAddAttrStateParameterOut(const char* szName, void* pValue, ViAttr attr)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    attrDataType attrType = lookupAttributeType(attr);
    switch(attrType)
    {
    case attrDataType_String:
      sprintf(tmp, "<parameter name=\"%s\" type=\"ViAttrState\" attr=\"%d\">%s</parameter>" LF, szName, attrType, (char*)pValue);
      break;

    case attrDataType_Uint16:
    default:
      sprintf(tmp, "<parameter name=\"%s\" type=\"ViAttrState\" attr=\"%d\">" _PERCD "</parameter>" LF, szName, attrType, *(ViInt32*)pValue);
      break;

      
    }
    tDynString* pString = &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddAttrStateArg2(const char* szName, ViUInt32 value)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViAttrState\">" _PERCU "</parameter>" LF, szName, value);
    tDynString* pString = &g_pLogger->pCurrentCall->sInputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddAttrStateOut2(const char* szName, ViUInt32* pValue, ViAttr attr)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_ATTRSTATE;
    para->pValue = pValue;
    para->attr = attr;

    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}


// ############# Real64 ######################
static void LogAddReal64Parameter(enum INOUT dir, const char* szName, ViReal64 value)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViReal64\">%f</parameter>" LF, szName, value);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddReal64Arg2(const char* szName, ViReal64 value)
{
  LogAddReal64Parameter(INOUT_IN, szName, value);
}

void LogAddReal64Out2(const char* szName, ViReal64* pValue)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_REAL64;
    para->pValue = pValue;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

// ############# Real64Arr ######################
static void LogAddReal64ArrParameter(enum INOUT dir, const char* szName, ViReal64* value, ViInt32* count)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViReal64Array\" cnt=\"" _PERCD "\">", szName, *count);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);

    if (*count)
    {
      AppendToString(pString, LF);
      AppendToString(pString, Indentation(g_pLogger->calldepth + 3));
      AppendToString(pString, "<block>");

      tDynString dtmp;
      AllocString(&dtmp, 1000);

      for (int i = 0; i < *count; i++)
      {
        if ((i % 8) == 0)
        {
          AppendToString(&dtmp, LF);
          AppendToString(&dtmp, Indentation(g_pLogger->calldepth + 3));
        }
        sprintf(tmp,"%f  ", value[i]);
        AppendToString(&dtmp, tmp);
      }
      AppendToString(pString, dtmp.data);
      ClearString(&dtmp);
      AppendToString(pString, LF);
      AppendToString(pString, Indentation(g_pLogger->calldepth + 3));
      AppendToString(pString,"</block>" LF);
    }    

    sprintf(tmp, "</parameter>" LF);
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddReal64ArrArg2(const char* szName, ViReal64* value, ViInt32* count)
{
  LogAddReal64ArrParameter(INOUT_IN, szName, value, count);
}

void LogAddReal64ArrOut2(const char* szName, ViReal64* pValue, ViInt32* count)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_REAL64ARR;
    para->pValue = pValue;
    para->pCount = count;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

// ############# Int32Arr ######################
static void LogAddInt32ArrParameter(enum INOUT dir, const char* szName, ViInt32* value, ViInt32* count)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViInt32Array\" cnt=\"" _PERCD "\">", szName, *count);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);

    if (*count)
    {
      AppendToString(pString, LF);
      AppendToString(pString, Indentation(g_pLogger->calldepth + 3));
      AppendToString(pString, "<block>");

      tDynString dtmp;
      AllocString(&dtmp, 1000);

      for (int i = 0; i < *count; i++)
      {
        if ((i % 8) == 0)
        {
          AppendToString(&dtmp, LF);
          AppendToString(&dtmp, Indentation(g_pLogger->calldepth + 3));
        }

        sprintf(tmp, _PERCD " ", value[i]);
        AppendToString(&dtmp, tmp);
      }
      AppendToString(pString, dtmp.data);
      ClearString(&dtmp);
      AppendToString(pString, LF);
      AppendToString(pString, Indentation(g_pLogger->calldepth + 3));
      AppendToString(pString,"</block>" LF);
    }
    sprintf(tmp, "</parameter>" LF);
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddInt32ArrArg2(const char* szName, ViInt32* value, ViInt32* count)
{
  LogAddInt32ArrParameter(INOUT_IN, szName, value, count);
}

void LogAddInt32ArrOut2(const char* szName, ViInt32* pValue, ViInt32* count)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_INT32ARR;
    para->pValue = pValue;
    para->pCount = count;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}


// ############# CharArr ######################
static void LogAddCharArrParameter(enum INOUT dir, const char* szName, ViChar* value, ViInt32* count)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    char tmpLineLeft[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViCharArray\" cnt=\"" _PERCD "\">", szName, *count);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);

    if (*count)
    {
      AppendToString(pString, LF );
      AppendToString(pString, Indentation(g_pLogger->calldepth + 3));
      AppendToString(pString, "<block>");

      tDynString dtmp;
      AllocString(&dtmp, 1000);

      *tmpLineLeft = 0;
      for (int i = 0; i < *count; i++)
      {
      
        sprintf(tmp, "%02x ", (value[i] & 0xff));
        strcat(tmpLineLeft, tmp);

        if (i && (((i+1) % 32) == 0) || i == *count-1)
        {
          AppendToString(&dtmp, LF);
          AppendToString(&dtmp, Indentation(g_pLogger->calldepth + 3));
          AppendToString(&dtmp, tmpLineLeft);
          *tmpLineLeft = 0;
        }
      }

      AppendToString(pString, dtmp.data);
      AppendToString(pString, LF);
      AppendToString(pString, Indentation(g_pLogger->calldepth + 3));
      AppendToString(pString, "</block>" LF);
      ClearString(&dtmp);

  // ------------
      AllocString(&dtmp, 1000);
      AppendToString(pString, Indentation(g_pLogger->calldepth + 3));
      AppendToString(pString, "<readable>");

      *tmpLineLeft = 0;
      for (int i = 0; i < *count; i++)
      {
        if (value[i])
        {
          EncodeChar(value[i], tmp);
          strcat(tmpLineLeft, tmp);
        }
        if (i && (((i+1) % 80) == 0) || i == *count-1 || value[i] == 0)
        {
          AppendToString(&dtmp, LF);
          AppendToString(&dtmp, Indentation(g_pLogger->calldepth + 3));
          AppendToString(&dtmp, tmpLineLeft);
          *tmpLineLeft = 0;
        }
        if (value[i] == 0)
          break;

      }
      AppendToString(pString, dtmp.data);
      ClearString(&dtmp);
    }

    AppendToString(pString, LF);
    AppendToString(pString, Indentation(g_pLogger->calldepth + 3));
    AppendToString(pString, "</readable>" LF);

    sprintf(tmp, "</parameter>" LF);
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddCharArrArg2(const char* szName, ViChar* value, ViInt32* count)
{
  LogAddCharArrParameter(INOUT_IN, szName, value, count);
}

void LogAddCharArrOut2(const char* szName, ViChar* pValue, ViInt32* count)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_CHARARR;
    para->pValue = pValue;
    para->pCount = count;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

// ############# StringArr ######################
static void LogAddStringArrParameter(enum INOUT dir, const char* szName, ViString* value, ViInt32* count)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    sprintf(tmp, "<parameter name=\"%s\" type=\"ViStringArray\" cnt=\"" _PERCD "\">", szName, *count);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);

    if (value && *count)
    {
      AppendToString(pString, LF);
      for (int i = 0; i < *count; i++)
      {
        AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
        AppendToString(pString, "<block>" LF);
        AppendToString(pString, Indentation(g_pLogger->calldepth + 3));
        AppendToString(pString, value[i]);
        AppendToString(pString, LF);
        AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
        AppendToString(pString, "</block>");
      }
    } 
    sprintf(tmp, "</parameter>" LF);
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
  }
}

void LogAddStringArrArg2(const char* szName, ViString* value, ViInt32* count)
{
  LogAddStringArrParameter(INOUT_IN, szName, value, count);
}

void LogAddStringArrOut2(const char* szName, ViString* pValue, ViInt32* count)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_STRINGARR;
    para->pValue = pValue;
    para->pCount = count;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

// ############# String ######################
static void LogAddStringParameter(enum INOUT dir, const char* szName, const ViChar* value)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    char tmp[1024];
    char* Replaced = ReplaceXmlEntities((char*)value);
    const ViChar* Arg =  Replaced ? Replaced : value;

    sprintf(tmp, "<parameter name=\"%s\" type=\"ViString\">%s</parameter>" LF, szName, Arg);
    tDynString* pString = dir == INOUT_IN ? &g_pLogger->pCurrentCall->sInputparameter : &g_pLogger->pCurrentCall->sOutputparameter;
    AppendToString(pString, Indentation(g_pLogger->calldepth + 2));
    AppendToString(pString, tmp);
    free(Replaced);
  }
}

void LogAddStringArg2(const char* szName, const ViChar* value)
{
  LogAddStringParameter(INOUT_IN, szName, value);
}

void LogAddStringOut2(const char* szName, ViChar* pValue)
{
  if (g_pLogger && !g_pLogger->stopped && g_pLogger->calldepth == 1)
  {
    tOutPara* para = &(g_pLogger->pCurrentCall->outputparameter[g_pLogger->pCurrentCall->numoutputparameter]);
    para->name = szName;
    para->type = DTYPE_STRING;
    para->pValue = pValue;
    g_pLogger->pCurrentCall->numoutputparameter++;
  }
}

void CollectOutput(void)
{
  for (size_t i = 0; i < g_pLogger->pCurrentCall->numoutputparameter; i++)
  {
    switch(g_pLogger->pCurrentCall->outputparameter[i].type)
    {
      case DTYPE_ViBoolean:
        LogAddBoolParameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, *((ViBoolean*)g_pLogger->pCurrentCall->outputparameter[i].pValue));
        break;

      case DTYPE_INT32:
        LogAddInt32Parameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, *((ViInt32*)g_pLogger->pCurrentCall->outputparameter[i].pValue));
        break;

      case DTYPE_UINT32:
        LogAddUInt32Parameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, *((ViUInt32*)g_pLogger->pCurrentCall->outputparameter[i].pValue));
        break;

      case DTYPE_INT64:
        LogAddInt64Parameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, *((ViInt64*)g_pLogger->pCurrentCall->outputparameter[i].pValue));
        break;

      case DTYPE_ATTRSTATE:
        LogAddAttrStateParameterOut(g_pLogger->pCurrentCall->outputparameter[i].name, g_pLogger->pCurrentCall->outputparameter[i].pValue, g_pLogger->pCurrentCall->outputparameter[i].attr);
        break;

      case DTYPE_REAL64:
        LogAddReal64Parameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, *((ViReal64*)g_pLogger->pCurrentCall->outputparameter[i].pValue));
        break;

      case DTYPE_SESSION:
        LogAddSessionParameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, *((ViSession*)g_pLogger->pCurrentCall->outputparameter[i].pValue));
        break;

      case DTYPE_STRING:
        LogAddStringParameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, ((ViChar*)g_pLogger->pCurrentCall->outputparameter[i].pValue));
        break;

      case DTYPE_REAL64ARR:
        LogAddReal64ArrParameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, ((ViReal64*)g_pLogger->pCurrentCall->outputparameter[i].pValue), g_pLogger->pCurrentCall->outputparameter[i].pCount);
        break;

      case DTYPE_INT32ARR:
        LogAddInt32ArrParameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, ((ViInt32*)g_pLogger->pCurrentCall->outputparameter[i].pValue), g_pLogger->pCurrentCall->outputparameter[i].pCount);
        break;

      case DTYPE_CHARARR:
        LogAddCharArrParameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, ((ViChar*)g_pLogger->pCurrentCall->outputparameter[i].pValue), g_pLogger->pCurrentCall->outputparameter[i].pCount);
        break;

      case DTYPE_STRINGARR:
        LogAddStringArrParameter(INOUT_OUT, g_pLogger->pCurrentCall->outputparameter[i].name, ((ViString*)g_pLogger->pCurrentCall->outputparameter[i].pValue), g_pLogger->pCurrentCall->outputparameter[i].pCount);
        break;

      default:
        assert(0);
        break;
    }
  }
}



ViBoolean DetermineProcessorClockrate(void)
{
  if (g_ProcessorClockrate < 0)
  {
#ifdef __linux__
    // Read clockrate from /proc
# ifdef __powerpc__
    const char* tag[]{ "timebase" };
    std::ifstream f;
    f.open("/proc/cpuinfo");
    if (f.is_open()) for (std::string str; !f.eof();)
    {
      f >> str;
      if (str == tag[0])
      {
        f >> str;
        f >> str;
        g_ProcessorClockrate = atof(str.c_str());
        break;
      }
    }
#else
    const char* tag[]={ "cpu", "MHz" };
    FILE* f = fopen("/proc/cpuinfo","r");
    if (f) 
    {
      int rc = 0;  
      char str[1024];  
      do
      {
        rc = fscanf(f,"%s", str);
        if (!strcmp(str,tag[0]))
        {
          rc = fscanf(f,"%s", str);
          if (rc && !strcmp(str,tag[1]))
          {
            rc = fscanf(f,"%s", str);
            if (rc)
            {
              rc = fscanf(f,"%s", str);
              g_ProcessorClockrate = 1E6 * atof(str);
            }
            break;
          }
        }
      } while(rc);
    }
#endif
#else
    // Read clockrate from registry
    HKEY hKey;
    if (ERROR_SUCCESS == RegOpenKeyExA(HKEY_LOCAL_MACHINE, "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0", 0, KEY_READ, &hKey))
    {
      DWORD valType;
      DWORD valSize = sizeof(DWORD);
      DWORD val;
      const ViBoolean ok = (ERROR_SUCCESS == RegQueryValueExA(hKey, "~MHz", 0, &valType, (BYTE*)&val, &valSize));
      RegCloseKey(hKey);
      if (ok && REG_DWORD == valType)
      {
        g_ProcessorClockrate = 1E6 * val;
      }
    }
#endif
  }
  // Plausibility test
  return g_ProcessorClockrate > 1E7;
}


/* METHOD ********************************************************************/
/**
Reads the processor clock, or a fraction of it in the powerpc case.
*
@return       Clock
******************************************************************************/
#ifndef __linux__
#  pragma warning(disable:4035)
#endif


static long long LogReadProcessorClock(void)
{
#ifdef __linux__
# ifdef __powerpc__
  union { unsigned long long ll; unsigned long l[2]; } var;
  unsigned long tmp;

  __asm__ __volatile__(
    "mfspr  %0,269;"
    "mfspr  %1,268;"
    "mfspr  %2,269;"
    : "=r" (tmp), "=r" (var.l[1]), "=r" (var.l[0])
  );
  if (var.l[0] != tmp)
    var.l[1] = 0;
  return var.ll;
# else
#  if defined(__i386__) || defined(__x86_64__)
#    if defined(__x86_64__)
  unsigned hi, lo;
  __asm__ __volatile__("rdtsc" : "=a"(lo), "=d"(hi));
  unsigned long long val = ((unsigned long long)lo) | (((unsigned long long)hi) << 32);
  return val;
#    else
  int64_t val; __asm__ __volatile__("rdtsc" : "=&A" (val)); return val;
#    endif
#  else
  int64_t val;
  struct timespec ts;
  clock_gettime(CLOCK_MONOTONIC, &ts);
  val = 1000000000LL * ts.tv_sec + ts.tv_nsec;
  return val;
#  endif
# endif
#else
#if defined _AMD64_
  return (__rdtsc());
#else
  __asm { RDTSC }
#endif
#endif
}


/* METHOD ********************************************************************/
/**
  The static method calculates the time in micro seconds between the last
call of the method and the actual call. If the i/p param is false, no
time difference value is calculated.
Intended for debugging purposes only!
*
@param            ViBoolean bIsToCalculate = true  ... calculate the time difference
                                                false ... don't calculate any time, the
                                                          default value.
*
@return           double = time in [micro-seconds].
******************************************************************************/
double GetClockInMicroSeconds(ViBoolean bIsToCalculate)
{
  static long long sliClockTicksActual = 0;
  if (bIsToCalculate)
  {
    long long liClockTicksPrev = sliClockTicksActual;
    sliClockTicksActual = LogReadProcessorClock();
    if (sliClockTicksActual < liClockTicksPrev)
    {
      //case with type long long overflow
      return (double)((LLONG_MAX - liClockTicksPrev + 1 + sliClockTicksActual)) /
        (g_ProcessorClockrate /*[Hz]*/ * 1.e-6 /*[micro-seconds]*/);
    }
    else
    {
      //normal case
      return (double)((sliClockTicksActual - liClockTicksPrev)) /
        (g_ProcessorClockrate /*[Hz]*/ * 1.e-6) /*[micro-seconds]*/;
    }
  }
  else
  {
    //initial case
    sliClockTicksActual = LogReadProcessorClock();
    return sliClockTicksActual / (g_ProcessorClockrate * 1.e-6);
  }
}




// ###########################

void InitCall(tCall* pCall)
{
  pCall->text.data[0] = 0;
  pCall->sInputparameter.data[0] = 0;
  pCall->sOutputparameter.data[0] = 0;
  pCall->sChildrenTexts.data[0] = 0;
  pCall->microsecondsstart = (long long)GetClockInMicroSeconds(VI_FALSE);
  pCall->numoutputparameter = 0;
}

void ConstructCall(tCall* pCall)
{
  memset(pCall, 0, sizeof(tCall));
  AllocString(&(pCall->text), DEFAULTSIZE);
  AllocString(&(pCall->sInputparameter), 1000);
  AllocString(&(pCall->sOutputparameter), 1000);
  AllocString(&(pCall->sChildrenTexts), 1000);
  InitCall(pCall);
}

void DestoyCall(tCall* pCall)
{
  ClearString(&pCall->sInputparameter);
  ClearString(&pCall->sOutputparameter);
}


void ConstructCallstack()
{
  for (int i = 0; i < MAXCALLDEPTH; i++)
  {
    ConstructCall(g_CallStack + i);
    if (i)
    {
      g_CallStack[i].parent = (struct tCall*) &(g_CallStack[i - 1]);
      g_CallStack[i - 1].child = (struct tCall*) &(g_CallStack[i]);
    }
  }
}

void DestoryCallstack()
{
  for (int i = 0; i < MAXCALLDEPTH; i++)
  {
    DestoyCall(g_CallStack + i);
  }
}


ViBoolean PushCall()
{
  if (g_pLogger->calldepth < g_pLogger->maxcalldepth)
  {
    if (g_pLogger->calldepth == 0)
    {
      g_pLogger->pCurrentCall = g_pLogger->nodes[g_pLogger->next];
    }
    else
    {
      tCall* pCurrentCall = g_pLogger->pCurrentCall;
      if (g_pLogger->calldepth == 1)
      {
        g_CallStack[0].parent = (struct tCall*) pCurrentCall;
        pCurrentCall->child = (struct tCall*) &(g_CallStack[0]);
      }
      g_pLogger->pCurrentCall = (tCall*)pCurrentCall->child;
    }

    g_pLogger->calldepth++;
    return VI_TRUE;
  }
  else
  {
    g_pLogger->calldepth++;
    return VI_FALSE;
  }
}

void PopCall()
{
  assert(g_pLogger->calldepth > 0);

  if (g_pLogger->calldepth > 1 && g_pLogger->calldepth <= g_pLogger->maxcalldepth)
    g_pLogger->pCurrentCall = (tCall*)g_pLogger->pCurrentCall->parent;
  g_pLogger->calldepth--;
}



void ClearString(tDynString* pNode)
{
  free(pNode->data);
  pNode->data = NULL;
  pNode->size = 0;
}

void AllocString(tDynString* pNode, size_t size)
{
  pNode->data = (char*) calloc(size, sizeof(char));
  pNode->size = size;
}

void AppendToString(tDynString* pNode, const char* szTxt)
{
  if (*szTxt)
  {
    size_t newsize = pNode->size;
    size_t required = strlen(pNode->data) + strlen(szTxt)+1;
    while (newsize < required)
      newsize *= 2;

    if (pNode->size < newsize)
    {
      pNode->data = (char*)realloc(pNode->data, (newsize) * sizeof(char));
      pNode->size = newsize;
    }
    strcat(pNode->data, szTxt);
  }
}


// You must free the result if result is non-NULL.
char *str_replace(char *orig, const char *rep, const char *with)
{
  char *result; // the return string
  char *ins;    // the next insert point
  char *tmp;    // varies
  size_t len_rep;  // length of rep (the string to remove)
  size_t len_with; // length of with (the string to replace rep with)
  size_t len_front; // distance between rep and end of last rep
  int count;    // number of replacements

  // sanity checks and initialization
  if (!orig || !rep)
    return NULL;
  len_rep = strlen(rep);
  if (len_rep == 0)
    return NULL; // empty rep causes infinite loop during count
  if (!with)
    with = "";
  len_with = strlen(with);

  // count the number of replacements needed
  ins = orig;
  for (count = 0; ; ++count)
  {
    tmp = strstr(ins, rep);
    if (!tmp)
      break;
    ins = tmp + len_rep;
  }
  if (count == 0)
    return NULL;

  tmp = result = (char*) malloc(strlen(orig) + (len_with - len_rep) * count + 1);

  if (!result)
    return NULL;

  // first time through the loop, all the variable are set correctly
  // from here on,
  //    tmp points to the end of the result string
  //    ins points to the next occurrence of rep in orig
  //    orig points to the remainder of orig after "end of rep"
  while (count--) {
    ins = strstr(orig, rep);
    len_front = ins - orig;
    tmp = strncpy(tmp, orig, len_front) + len_front;
    tmp = strcpy(tmp, with) + len_with;
    orig += len_front + len_rep; // move to next "end of rep"
  }
  strcpy(tmp, orig);
  return result;
}

void EncodeChar(const char c, char* out)
{
  for (int i = 0; i < sizeof(Pairs) / sizeof(tFindReplace); i++)
  {
    if (c == Pairs[i].replace[0])
    {
      strcpy(out, Pairs[i].with);
      return;
    }
  }
  if (c > 0 && isprint(c))
    sprintf(out, "%c", c);
  else
    sprintf(out, ".");
}

char* ReplaceXmlEntities(char* str)
{

  char* pSearch = str;
  for (int i = 0; i < sizeof(Pairs) / sizeof(tFindReplace); i++)
  {
    char* pReplaced = str_replace(pSearch, Pairs[i].replace, Pairs[i].with);
    if (pReplaced)
    {
      if (pSearch != str)
        free(pSearch);
      pSearch = pReplaced;
    }
  }
  return pSearch != str ? pSearch : NULL;
}

const char* Indentation(int cnt)
{
  static const char Tabs[] = { TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB TB };
  #define NUMPTRS  (sizeof(Tabs) / sizeof(TB))  
  static const char* Pointers[NUMPTRS];
  static int ptrsinitialized = 0;
  if (!ptrsinitialized)
  {
    ptrsinitialized = 1;
    const char* p = Tabs + strlen(Tabs);
    for (int i = 0; i < NUMPTRS; i++)
    {
      Pointers[i] = p;
      p -= strlen(TB);
    }
  }
  assert(cnt < NUMPTRS);
  return Pointers[cnt];
}


typedef struct 
{
  unsigned long id;
  attrDataType type;
}attrId;

const attrId g_AttrIds[] =
{
{0XBFFF0001UL, attrDataType_String},
{0XBFFF0002UL, attrDataType_String},
{0X3FFF0003UL, attrDataType_Uint16},
{0X3FFF0004UL, attrDataType_Uint16},
{0X3FFF0005UL, attrDataType_Uint16},
{0X3FFF0007UL, attrDataType_Uint16},
{0X3FFF000DUL, attrDataType_Uint16},
{0X3FFF000FUL, attrDataType_Uint16},
{0X3FFF0011UL, attrDataType_Uint16},
{0X3FFF0013UL, attrDataType_Uint16},
{0X3FFF0016UL, attrDataType_Uint16},
{0X3FFF0018UL, attrDataType_Uint16},
{0X3FFF001AUL, attrDataType_Uint16},
{0X3FFF001BUL, attrDataType_Uint16},
{0X3FFF001CUL, attrDataType_Uint16},
{0X3FFF001EUL, attrDataType_Uint16},
{0X3FFF0021UL, attrDataType_Uint16},
{0X3FFF0022UL, attrDataType_Uint16},
{0X3FFF0023UL, attrDataType_Uint16},
{0X3FFF0024UL, attrDataType_Uint16},
{0X3FFF0025UL, attrDataType_Uint16},
{0X3FFF002AUL, attrDataType_Uint16},
{0X3FFF002BUL, attrDataType_Uint16},
{0X3FFF002DUL, attrDataType_Uint16},
{0X3FFF002EUL, attrDataType_Uint16},
{0X3FFF0036UL, attrDataType_Uint16},
{0X3FFF0038UL, attrDataType_Uint16},
{0X3FFF0039UL, attrDataType_Uint16},
{0X3FFF003AUL, attrDataType_Uint16},
{0X3FFF003CUL, attrDataType_Uint16},
{0X3FFF003DUL, attrDataType_Uint16},
{0X3FFF0040UL, attrDataType_Uint16},
{0X3FFF0041UL, attrDataType_Uint16},
{0X3FFF0045UL, attrDataType_Uint16},
{0X3FFF0047UL, attrDataType_Uint16},
{0X3FFF0057UL, attrDataType_Uint16},
{0X3FFF005CUL, attrDataType_Uint16},
{0X3FFF005EUL, attrDataType_Uint16},
{0X3FFF0062UL, attrDataType_Uint16},
{0X3FFF0067UL, attrDataType_Uint16},
{0X3FFF0068UL, attrDataType_Uint16},
{0X3FFF0069UL, attrDataType_Uint16},
{0X3FFF006BUL, attrDataType_Uint16},
{0X3FFF006CUL, attrDataType_Uint16},
{0X3FFF0070UL, attrDataType_Uint16},
{0XBFFF0072UL, attrDataType_String},
{0XBFFF0077UL, attrDataType_String},
{0X3FFF008BUL, attrDataType_Uint16},
{0X3FFF008DUL, attrDataType_Uint16},
{0X3FFF0094UL, attrDataType_Uint16},
{0X3FFF0098UL, attrDataType_Uint16},
{0X3FFF009AUL, attrDataType_Uint16},
{0X3FFF00ACUL, attrDataType_Uint16},
{0X3FFF00ADUL, attrDataType_Uint16},
{0X3FFF00AEUL, attrDataType_Uint16},
{0X3FFF00AFUL, attrDataType_Uint16},
{0X3FFF00B1UL, attrDataType_Uint16},
{0X3FFF00B2UL, attrDataType_Uint16},
{0X3FFF00B3UL, attrDataType_Uint16},
{0X3FFF00B4UL, attrDataType_Uint16},
{0X3FFF00BEUL, attrDataType_Uint16},
{0X3FFF00BFUL, attrDataType_Uint16},
{0X3FFF00C0UL, attrDataType_Uint16},
{0X3FFF00C1UL, attrDataType_Uint16},
{0X3FFF00C2UL, attrDataType_Uint16},
{0X3FFF00C3UL, attrDataType_Uint16},
{0X3FFF00C4UL, attrDataType_Uint16},
{0X3FFF00D5UL, attrDataType_Uint16},
{0X3FFF00D9UL, attrDataType_Uint16},
{0X3FFF00DDUL, attrDataType_Uint16},
{0X3FFF00DEUL, attrDataType_Uint16},
{0X3FFF00DFUL, attrDataType_Uint16},
{0X3FFF00E8UL, attrDataType_Uint16},
{0XBFFF00E9UL, attrDataType_String},
{0X3FFF0100UL, attrDataType_Uint16},
{0X3FFF0101UL, attrDataType_Uint16},
{0X3FFF0170UL, attrDataType_Uint16},
{0X3FFF0171UL, attrDataType_Uint16},
{0X3FFF0172UL, attrDataType_Uint16},
{0X3FFF0173UL, attrDataType_Uint16},
{0XBFFF0174UL, attrDataType_String},
{0X3FFF0175UL, attrDataType_Uint16},
{0X3FFF0176UL, attrDataType_Uint16},
{0X3FFF0177UL, attrDataType_Uint16},
{0X3FFF0181UL, attrDataType_Uint16},
{0X3FFF0184UL, attrDataType_Uint16},
{0X3FFF0189UL, attrDataType_Uint16},
{0X3FFF0192UL, attrDataType_Uint16},
{0X3FFF0194UL, attrDataType_Uint16},
{0XBFFF0195UL, attrDataType_String},
{0XBFFF0196UL, attrDataType_String},
{0X3FFF0197UL, attrDataType_Uint16},
{0XBFFF0199UL, attrDataType_String},
{0X3FFF019AUL, attrDataType_Uint16},
{0X3FFF019BUL, attrDataType_Uint16},
{0X3FFF019FUL, attrDataType_Uint16},
{0XBFFF01A0UL, attrDataType_String},
{0X3FFF01A1UL, attrDataType_Uint16},
{0X3FFF01A7UL, attrDataType_Uint16},
{0X3FFF01AFUL, attrDataType_Uint16},
{0X3FFF0201UL, attrDataType_Uint16},
{0X3FFF0202UL, attrDataType_Uint16},
{0X3FFF0205UL, attrDataType_Uint16},
{0X3FFF0206UL, attrDataType_Uint16},
{0XBFFF0207UL, attrDataType_Uint16},
{0X3FFF0208UL, attrDataType_Uint16},
{0X3FFF0209UL, attrDataType_Uint16},
{0X3FFF020AUL, attrDataType_Uint16},
{0X3FFF020BUL, attrDataType_Uint16},
{0X3FFF020CUL, attrDataType_Uint16},
{0X3FFF020DUL, attrDataType_Uint16},
{0X3FFF020EUL, attrDataType_Uint16},
{0X3FFF0211UL, attrDataType_Uint16},
{0X3FFF0212UL, attrDataType_Uint16},
{0X3FFF0213UL, attrDataType_Uint16},
{0X3FFF0214UL, attrDataType_Uint16},
{0X3FFF0215UL, attrDataType_Uint16},
{0X3FFF0216UL, attrDataType_Uint16},
{0X3FFF0221UL, attrDataType_Uint16},
{0X3FFF0222UL, attrDataType_Uint16},
{0X3FFF0223UL, attrDataType_Uint16},
{0X3FFF0224UL, attrDataType_Uint16},
{0X3FFF0225UL, attrDataType_Uint16},
{0X3FFF0226UL, attrDataType_Uint16},
{0X3FFF0228UL, attrDataType_Uint16},
{0X3FFF0229UL, attrDataType_Uint16},
{0X3FFF022AUL, attrDataType_Uint16},
{0X3FFF022BUL, attrDataType_Uint16},
{0X3FFF022CUL, attrDataType_Uint16},
{0X3FFF022DUL, attrDataType_Uint16},
{0X3FFF0231UL, attrDataType_Uint16},
{0X3FFF0232UL, attrDataType_Uint16},
{0X3FFF0233UL, attrDataType_Uint16},
{0X3FFF0234UL, attrDataType_Uint16},
{0X3FFF0235UL, attrDataType_Uint16},
{0X3FFF0236UL, attrDataType_Uint16},
{0X3FFF0238UL, attrDataType_Uint16},
{0X3FFF0239UL, attrDataType_Uint16},
{0X3FFF023AUL, attrDataType_Uint16},
{0X3FFF023BUL, attrDataType_Uint16},
{0X3FFF023CUL, attrDataType_Uint16},
{0X3FFF023DUL, attrDataType_Uint16},
{0X3FFF0240UL, attrDataType_Uint16},
{0X3FFF0241UL, attrDataType_Uint16},
{0X3FFF0242UL, attrDataType_Uint16},
{0X3FFF0243UL, attrDataType_Uint16},
{0X3FFF0244UL, attrDataType_Uint16},
{0X3FFF0245UL, attrDataType_Uint16},
{0X3FFF0246UL, attrDataType_Uint16},
{0X3FFF0300UL, attrDataType_Uint16},
{0X3FFF0301UL, attrDataType_Uint16},
{0X3FFF0302UL, attrDataType_Uint16},
{0X3FFF0303UL, attrDataType_Uint16},
{0X3FFF4006UL, attrDataType_Uint16},
{0X3FFF4010UL, attrDataType_Uint16},
{0X3FFF4011UL, attrDataType_Uint16},
{0X3FFF4012UL, attrDataType_Uint16},
{0X3FFF4023UL, attrDataType_Uint16},
{0X3FFF4025UL, attrDataType_Uint16},
{0X3FFF4026UL, attrDataType_Uint16},
{0X3FFF4027UL, attrDataType_Uint16},
{0X3FFF4041UL, attrDataType_Uint16},
{0XBFFF4042UL, attrDataType_String},
{0X3FFF4193UL, attrDataType_Uint16},
{0XBFFF4198UL, attrDataType_String},
{0X3FFF41B0UL, attrDataType_Uint16},
{0XBFFF41B1UL, attrDataType_Uint16},
{0X3FFF4240UL, attrDataType_Uint16},
{0X3FFF4241UL, attrDataType_Uint16},

/*- Attributes (platform dependent size) ------------------------------------*/

#if defined(_VI_INT64_UINT64_DEFINED) && defined(_VISA_ENV_IS_64_BIT)
{0X3FFF000AUL, attrDataType_Uint16},
{0X3FFF4028UL, attrDataType_Uint16},
{VI_ATTR_USER_DATA_64, attrDataType_Uint16},
{VI_ATTR_RET_COUNT_64, attrDataType_Uint16},
#else
{VI_ATTR_USER_DATA_32, attrDataType_Uint16},
{VI_ATTR_RET_COUNT_32, attrDataType_Uint16},
#endif
#if defined(_VI_INT64_UINT64_DEFINED)
{0X3FFF009BUL, attrDataType_Uint16},
{0X3FFF009CUL, attrDataType_Uint16},
{0X3FFF00D0UL, attrDataType_Uint16},
{0X3FFF00D1UL, attrDataType_Uint16},
#endif
#if defined(_VI_INT64_UINT64_DEFINED) && defined(_VISA_ENV_IS_64_BIT)
{VI_ATTR_WIN_BASE_ADDR_64, attrDataType_Uint16},
{VI_ATTR_WIN_SIZE_64, attrDataType_Uint16},
{VI_ATTR_MEM_BASE_64, attrDataType_Uint16},
{VI_ATTR_MEM_SIZE_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR0_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR1_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR2_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR3_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR4_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR5_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR0_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR1_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR2_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR3_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR4_64, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR5_64, attrDataType_Uint16},
#else
{VI_ATTR_WIN_BASE_ADDR_32, attrDataType_Uint16},
{VI_ATTR_WIN_SIZE_32, attrDataType_Uint16},
{VI_ATTR_MEM_BASE_32, attrDataType_Uint16},
{VI_ATTR_MEM_SIZE_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR0_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR1_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR2_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR3_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR4_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_BASE_BAR5_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR0_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR1_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR2_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR3_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR4_32, attrDataType_Uint16},
{VI_ATTR_PXI_MEM_SIZE_BAR5_32, attrDataType_Uint16},
#endif
};

attrDataType lookupAttributeType(unsigned int id)
{
  for (unsigned int i = 0; i < sizeof(g_AttrIds) / sizeof(attrId); i++)
  {
    if (g_AttrIds[i].id == id)
    {
      return g_AttrIds[i].type;
    }
  }
  return attrDataType_Uint16;
}
