/*****************************************************************************
 *  Rohde&Schwarz SGT signal generator instrument driver
 *  Callbacks
 *
 *  Original Release: January 2014
 *  By: Jiri Kominek
 *
 *  Should you have any technical questions please contact the hotline of
 *  Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
 *
 *  e-mail: CustomerSupport@rohde-schwarz.com
 *

 *****************************************************************************/

#include "rssgt.h"
#include <float.h>
#include <math.h>
#include "SgmaCommandIds.h"
#include "rssgt_utility.h"

 //#pragma warning (disable:4100) // unreferenced formal parameter

static ViString gs_path[] = { "", "A", "B", NULL };
/****/

/*****************************************************************************
 * Function: rssgt_binPattern_ReadCallback
 * Purpose:  This function overrides standard calback function.
 *****************************************************************************/
ViStatus rssgt_binPattern_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViStatus    error = VI_SUCCESS;
  ViInt32     idx = 0;
  ViInt32		j, i = 0;
  ViUInt64     hexValue = 0, b[1024];
  ViChar      *cmd = NULL,
    buffer[RS_MAX_MESSAGE_BUF_SIZE], bin[RS_MAX_MESSAGE_BUF_SIZE] = "";
  ViInt32     path = 1;
  ViChar      repcap[1024] = "";

  memset(b, 0, 1024 * sizeof(b[0]));
  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  if (repCapName && strlen(repCapName) > 0)
  {
    sprintf(repcap, "%s,%s", gs_path[path], repCapName);
  }
  else
  {
    sprintf(repcap, "%s", gs_path[path]);
  }

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repcap, &cmd));

  /* --- Attribute data type is used for VISA I/O --- */
  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_STRING:
    if (sessionProperties->attr[idx].access == RS_VAL_READ_ONLY)
    {
      sprintf(buffer, "%s", cmd);
    }
    else
    {
      sprintf(buffer, "%s?", cmd);
    }

    strcat(buffer, "\n");


    viCheckErr(rsPrintf(instrSession, "%s", buffer));

    viCheckErr(rsScanf(instrSession, "#H%[^,]", buffer));
    sscanf(buffer, _PERCLX, &hexValue);

    do
    {
      b[i] = hexValue % 2;
      hexValue >>= 1;
      i++;
    } while (hexValue);

    for (j = i - 1; j >= 0; j--)
    {
      sprintf(buffer, _PERCLD, b[j]);
      strcat(bin, buffer);
    }
    sscanf(bin, "%s", (char*)value);
    break;
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }


Error:

  if (cmd)
    free(cmd);

  return error;
};


/*****************************************************************************
 * Function: rssgt_binPattern_WriteCallback
 * Purpose:  This function overrides standard calback function.
 *****************************************************************************/
ViStatus rssgt_binPattern_WriteCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViStatus    error = VI_SUCCESS;
  ViInt32     idx = 0;
  ViChar      *cmd = NULL;
  ViInt32     patternLength;
  ViInt32     path = 1;
  ViChar      repcap[1024] = "";

  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  if (repCapName && strlen(repCapName) > 0)
  {
    sprintf(repcap, "%s,%s", gs_path[path], repCapName);
  }
  else
  {
    sprintf(repcap, "%s", gs_path[path]);
  }

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repcap, &cmd));

  patternLength = (ViInt32)strlen((ViString)value);

  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_STRING:
    viCheckErr(rsPrintf(instrSession, "%s #B%s,%ld\n", cmd, (ViString)value, patternLength));
    break;
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }
Error:

  if (cmd)
    free(cmd);

  return error;

}

/*****************************************************************************
 * Function: rssgt_ReadOnlyCollection_ReadCallback
 * Purpose:  This function overrides standard calback function.
 *****************************************************************************/
ViStatus rssgt_ReadOnlyCollection_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViStatus    error = VI_SUCCESS;
  ViInt32     idx = 0;
  ViUInt32    ret_cnt = 0;
  ViChar      *cmd = NULL,
    buffer[RS_MAX_MESSAGE_BUF_SIZE];
  ViChar*  	ext_buf = NULL;

  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  /* Check if attribute uses repeated capabilities */
  if (sessionProperties->attr[idx].usercaps == VI_TRUE)
  {
    if (!repCapName)
      viCheckErr(RS_ERROR_INVALID_REPEATED_CAPABILITY_NAME);

    if (strlen(repCapName) == 0)
      viCheckErr(RS_ERROR_INVALID_REPEATED_CAPABILITY_NAME);
  }

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repCapName, &cmd));

  /* --- Attribute data type is used for VISA I/O --- */
  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_STRING:
    strcpy(buffer, cmd);
    if (sessionProperties->attr[idx].access == RS_VAL_READ_ONLY)
      strcat(buffer, "\n");
    else
      strcat(buffer, "?\n");

    viCheckErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));

    viCheckErr(Rs_ReadDataUnknownLength(instrSession, &ext_buf, &ret_cnt));
    if ((ViInt32)ret_cnt > bufSize)
      error = VI_SUCCESS_MAX_CNT;
    ext_buf[strcspn(ext_buf, "\r\n")] = '\0';
    if ((ViInt32)ret_cnt > bufSize) ret_cnt = bufSize - 1;
    memset(value, 0, bufSize);
    strncpy(value, ext_buf, ret_cnt);


    break;
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }

Error:
  if (cmd)
    free(cmd);

  if (ext_buf)
    free(ext_buf);

  return error;
}

/*****************************************************************************
 * Function: rssgt_WriteReal64_Callback
 * Purpose:  writes a double using either the fast method or via the textual interface
 *****************************************************************************/
static ViStatus rssgt_WriteReal64_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, int FastParameterId, void *value)
{
  ViUInt32 supported = 0;
  if (repCapName && (*repCapName == 'B'))
    FastParameterId++;
  viFastSettingSupported(io, FastParameterId, &supported);
  if (supported)
  {
    return viFastSetReal64(io, FastParameterId, *(ViReal64*)value);
  }
  return Rs_WriteCallback(io, repCapName, attributeId, value);
}

/*****************************************************************************
 * Function: rssgt_ReadReal64_Callback
 * Purpose:  reads a double using either the fast method or via the textual interface
 *****************************************************************************/
static ViStatus rssgt_ReadReal64_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, int FastParameterId, ViInt32 bufSize,
  void *value)
{
  ViUInt32 supported = 0;
  if (repCapName && (*repCapName == 'B'))
    FastParameterId++;
  viFastQuerySupported(io, FastParameterId, &supported);
  if (supported)
  {
    return viFastQueryReal64(io, FastParameterId, (ViReal64*)value);
  }
  return Rs_ReadCallback(io, repCapName, attributeId, bufSize, value);
}

#if 0
/*****************************************************************************
 * Function: rssgt_WriteInt32_Callback
 * Purpose:  writes an int using either the fast method or via the textual interface
 *****************************************************************************/
static ViStatus rssgt_WriteInt32_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, int FastParameterId, void *value)
{
  ViUInt32 supported = 0;
  if (repCapName && (*repCapName == 'B'))
    FastParameterId++;
  viFastSettingSupported(io, FastParameterId, &supported);
  if (supported)
  {
    return viFastSetInt32(io, FastParameterId, *(ViInt32*)value);
  }
  return Rs_WriteCallback(io, repCapName, attributeId, value);
}

/*****************************************************************************
 * Function: rssgt_ReadInt32_Callback
 * Purpose:  reads an integer using either the fast method or via the textual interface
 *****************************************************************************/

static ViStatus rssgt_ReadInt32_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, int FastParameterId, ViInt32 bufSize,
  void *value)
{
  ViUInt32 supported = 0;
  if (repCapName && (*repCapName == 'B'))
    FastParameterId++;
  viFastQuerySupported(io, FastParameterId, &supported);
  if (supported)
  {
    return viFastQueryInt32(io, FastParameterId, (ViInt32*)value);
  }
  return Rs_ReadCallback(io, repCapName, attributeId, bufSize, value);
}
#endif

/*****************************************************************************
 * Function: rssgt_WriteBoolean_Callback
 * Purpose:  writes an Boolean using either the fast method or via the textual interface
 *****************************************************************************/
static ViStatus rssgt_WriteBoolean_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, int FastParameterId, void *value)
{
  ViUInt32 supported = 0;
  if (repCapName && (*repCapName == 'B'))
    FastParameterId++;
  viFastSettingSupported(io, FastParameterId, &supported);
  if (supported)
  {
    ViInt32 IntVal = 0;
    if (*((ViBoolean*)value) == VI_TRUE)
      IntVal = 1;

    return viFastSetInt32(io, FastParameterId, IntVal);
  }

  return Rs_WriteCallback(io, repCapName, attributeId, value);
}

/*****************************************************************************
 * Function: rssgt_ReadBoolean_Callback
 * Purpose:  reads an Boolean using either the fast method or via the textual interface
 *****************************************************************************/

static ViStatus rssgt_ReadBoolean_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, int FastParameterId, ViInt32 bufSize,
  void *value)
{
  ViUInt32 supported = 0;
  if (repCapName && (*repCapName == 'B'))
    FastParameterId++;
  viFastQuerySupported(io, FastParameterId, &supported);
  if (supported)
  {
    ViInt32* IntVal = (ViInt32*)malloc(RS_VAL_INT32_SIZE);
    ViStatus rc;

    if (IntVal != NULL)
    {
      *IntVal = 0;
    }
    else
      return RS_ERROR_OUT_OF_MEMORY;

    rc = viFastQueryInt32(io, FastParameterId, IntVal);
    *((ViBoolean*)value) = (*IntVal != 0 ? VI_TRUE : VI_FALSE);

    free(IntVal);

    return rc;
  }

  return Rs_ReadCallback(io, repCapName, attributeId, bufSize, value);
}

/*****************************************************************************
 * Function: rssgt_WriteEvent_Callback
 * Purpose:  writes an int using either the fast method or via the textual interface
 *****************************************************************************/
static ViStatus rssgt_WriteEvent_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, int FastParameterId)
{
  ViUInt32 supported = 0;
  if (repCapName && (*repCapName == 'B'))
    FastParameterId++;
  viFastSettingSupported(io, FastParameterId, &supported);
  if (supported)
  {
    return viFastSetInt32(io, FastParameterId, 1);
  }
  return Rs_WriteCallback(io, repCapName, attributeId, NULL);
}

/*****************************************************************************
 * Function: rssgt_WriteFrequency_Callback
 * Purpose:  This function overrides standard callback function.
 * writes a frequency value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteFrequency_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  return rssgt_WriteReal64_Callback(io, repCapName, attributeId, IdSgtFreqFreq, value);
}

/*****************************************************************************
 * Function: rssgt_ReadFrequency_Callback
 * Purpose:  This function overrides standard callback function.
 * reads a frequency value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_ReadFrequency_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize, void *value)
{
  return rssgt_ReadReal64_Callback(io, repCapName, attributeId, IdSgtFreqFreq, bufSize, value);
}

/*****************************************************************************
 * Function: rssgt_WriteLevel_Callback
 * Purpose:  This function overrides standard callback function.
 * writes a level value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteLevel_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  return rssgt_WriteReal64_Callback(io, repCapName, attributeId, IdSgtPowPow, value);
}

/*****************************************************************************
 * Function: rssgt_ReadLevel_Callback
 * Purpose:  This function overrides standard callback function.
 * reads a level value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_ReadLevel_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize,
  void *value)
{
  return rssgt_ReadReal64_Callback(io, repCapName, attributeId, IdSgtPowPow, bufSize, value);
}

/*****************************************************************************
 * Function: rssgt_WriteRfState_Callback
 * Purpose:  This function overrides standard callback function.
 * writes a RfState value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteRfState_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  return rssgt_WriteBoolean_Callback(io, repCapName, attributeId, IdSgtRfState, value);
}

/*****************************************************************************
 * Function: rssgt_ReadRfState_Callback
 * Purpose:  This function overrides standard callback function.
 * reads a RfState value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_ReadRfState_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize,
  void *value)
{
  return rssgt_ReadBoolean_Callback(io, repCapName, attributeId, IdSgtRfState, bufSize, value);
}
/*****************************************************************************
 * Function: rssgt_WriteModState_Callback
 * Purpose:  This function overrides standard callback function.
 * writes a ModState value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteModState_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  return rssgt_WriteBoolean_Callback(io, repCapName, attributeId, IdSgtModState, value);
}

/*****************************************************************************
 * Function: rssgt_ReadModState_Callback
 * Purpose:  This function overrides standard callback function.
 * reads a ModState value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_ReadModState_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize,
  void *value)
{
  return rssgt_ReadBoolean_Callback(io, repCapName, attributeId, IdSgtModState, bufSize, value);
}

/*****************************************************************************
 * Function: rssgt_WriteIqWidebandState_Callback
 * Purpose:  This function overrides standard callback function.
 * writes a IqWidebandState value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteIqWidebandState_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  return rssgt_WriteBoolean_Callback(io, repCapName, attributeId, IdSgtIqWidebandState, value);
}

/*****************************************************************************
 * Function: rssgt_ReadIqWidebandState_Callback
 * Purpose:  This function overrides standard callback function.
 * reads a IqWidebandState value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_ReadIqWidebandState_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize,
  void *value)
{
  return rssgt_ReadBoolean_Callback(io, repCapName, attributeId, IdSgtIqWidebandState, bufSize, value);
}

/*****************************************************************************
 * Function: rssgt_WriteDigitalAttenuation_Callback
 * Purpose:  This function overrides standard callback function.
 * writes a DigitalAttenuation value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteDigitalAttenuation_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  return rssgt_WriteReal64_Callback(io, repCapName, attributeId, IdSgtPowAttDig, value);
}

/*****************************************************************************
 * Function: rssgt_ReadDigitalAttenuation_Callback
 * Purpose:  This function overrides standard callback function.
 * reads a DigitalAttenuation value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_ReadDigitalAttenuation_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize, void *value)
{
  return rssgt_ReadReal64_Callback(io, repCapName, attributeId, IdSgtPowAttDig, bufSize, value);
}

/*****************************************************************************
 * Function: rssgt_WriteFrequencyOffset_Callback
 * Purpose:  This function overrides standard callback function.
 * writes a FrequencyOffset value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteFrequencyOffset_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  return rssgt_WriteReal64_Callback(io, repCapName, attributeId, IdSgtFreqOffsUnicod, value);
}

/*****************************************************************************
 * Function: rssgt_ReadFrequencyOffset_Callback
 * Purpose:  This function overrides standard callback function.
 * reads a FrequencyOffset value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_ReadFrequencyOffset_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize, void *value)
{
  return rssgt_ReadReal64_Callback(io, repCapName, attributeId, IdSgtFreqOffsUnicod, bufSize, value);
}

/*****************************************************************************
 * Function: rssgt_WritePhaseOffset_Callback
 * Purpose:  This function overrides standard callback function.
 * writes a PhaseOffset value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WritePhaseOffset_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  return rssgt_WriteReal64_Callback(io, repCapName, attributeId, IdSgtPhasOffsUnicod, value);
}

/*****************************************************************************
 * Function: rssgt_ReadPhaseOffset_Callback
 * Purpose:  This function overrides standard callback function.
 * reads a PhaseOffset value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_ReadPhaseOffset_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize, void *value)
{
  return rssgt_ReadReal64_Callback(io, repCapName, attributeId, IdSgtPhasOffsUnicod, bufSize, value);
}
/*****************************************************************************
 * Function: rssgt_WriteArbNextSegmentExecute_Callback
 * Purpose:  This function overrides standard callback function.
 * triggers the next ARB segment using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteArbNextSegmentExecute_Callback(ViSession instrSession, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  ViStatus    error = VI_SUCCESS;
  ViInt32     path = 1;

  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));
  viCheckErr(rssgt_WriteEvent_Callback(instrSession, gs_path[path], attributeId, IdSgtArbSegmNextExec));
Error:
  return error;
}

/*****************************************************************************
 * Function: rssgt_WriteArbExecuteTrigger_Callback
 * Purpose:  This function overrides standard callback function.
 * triggers the ARB using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteArbTriggerExecute_Callback(ViSession instrSession, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  ViStatus    error = VI_SUCCESS;
  ViInt32     path = 1;

  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));
  viCheckErr(rssgt_WriteEvent_Callback(instrSession, gs_path[path], attributeId, IdSgtArbTrigExec));
Error:
  return error;
}

/*****************************************************************************
 * Function: rssgt_WriteFrequency_Callback
 * Purpose:  This function overrides standard callback function.
 * writes a frequency value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_WriteArbNextSegment_Callback(ViSession instrSession, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  ViStatus    error = VI_SUCCESS;
  ViInt32     path = 1;

  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));
  viCheckErr(rssgt_WriteReal64_Callback(instrSession, gs_path[path], attributeId, IdSgtArbSegmNext, value));
Error:
  return error;
}

/*****************************************************************************
 * Function: rssgt_ReadFrequency_Callback
 * Purpose:  This function overrides standard callback function.
 * reads a frequency value using either the fast method or via the textual interface
 *****************************************************************************/
ViStatus rssgt_ReadArbNextSegment_Callback(ViSession instrSession, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize, void *value)
{
  ViStatus    error = VI_SUCCESS;
  ViInt32     path = 1;

  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));
  viCheckErr(rssgt_ReadReal64_Callback(instrSession, gs_path[path], attributeId, IdSgtArbSegmNext, bufSize, value));
Error:
  return error;
}


/*****************************************************************************
 * Function: rssgt_WriteOpcTimeout_Callback
 * Purpose:  This function overrides standard callback function.
 * writes the standard timeout value
 *****************************************************************************/
ViStatus rssgt_WriteOpcTimeout_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, void *value)
{
  return  rsSetAttribute(io, VI_ATTR_TMO_VALUE, *(ViInt32*)value);
}

/*****************************************************************************
 * Function: rssgt_ReadOpcTimeout_Callback
 * Purpose:  This function overrides standard callback function.
 * reads the standard timeout value
 *****************************************************************************/
ViStatus rssgt_ReadOpcTimeout_Callback(ViSession io, ViConstString repCapName,
  ViAttr attributeId, ViInt32 bufSize, void *value)
{
  return rsGetAttribute(io, VI_ATTR_TMO_VALUE, value);
}

/*****************************************************************************
 * Function: rssgt_doubleQuotedString_ReadCallback
 * Purpose:  This function overrides standard calback funtion. Reads string
 *           returned with apostrophes
 *****************************************************************************/
ViStatus rssgt_PathDoubleQuotedString_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViStatus    error = VI_SUCCESS;
  ViUInt32    count = bufSize;
  ViInt32     idx = 0;
  ViChar      *cmd = NULL,
    *buffer = NULL;
  ViChar      format_string[1024] = "";
  ViInt32     path = 1;
  ViChar      repcap[1024] = "";

  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  if (repCapName && strlen(repCapName) > 0)
  {
    sprintf(repcap, "%s,%s", gs_path[path], repCapName);
  }
  else
  {
    sprintf(repcap, "%s", gs_path[path]);
  }

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repcap, &cmd));

  /* --- If the range table for the attribute exists, revert 'command string'
     back to the 'discreteOrMinValue' and return it as attribute value. --- */

     /* Range Table Callback */
  if (sessionProperties->attr[idx].rangeTableCallback)
  {
    viCheckErr((sessionProperties->attr[idx].rangeTableCallback)(instrSession, repcap, attributeId, &(sessionProperties->attr[idx].rangeTable)));
  }

  /* --- Attribute data type is used for VISA I/O --- */
  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_STRING:
    if (sessionProperties->attr[idx].access == RS_VAL_READ_ONLY)
    {
      viCheckErr(rsPrintf(instrSession, "%s\n", cmd));
    }
    else
    {
      viCheckErr(rsPrintf(instrSession, "%s?\n", cmd));
    }
    viCheckErr(Rs_ReadDataUnknownLength(instrSession, &buffer, &count));
    sprintf(format_string, "\"%% "_PERCD "[^\"]\"", bufSize - 1);
    memset(value, 0, bufSize);
    sscanf(buffer, format_string, value);
    break;
  case RS_VAL_INT32:
  case RS_VAL_REAL64:
  case RS_VAL_BOOLEAN:
  case RS_VAL_ADDR:
  case RS_VAL_SESSION:
  case RS_VAL_EVENT:
  case RS_VAL_UNKNOWN_TYPE:
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }

Error:

  if (buffer)
    free(buffer);

  if (cmd)
    free(cmd);

  return error;
}

/*****************************************************************************
 * Function: rssgt_doubleQuotedString_ReadCallback
 * Purpose:  This function overrides standard calback funtion. Reads string
 *           returned with apostrophes
 *****************************************************************************/
ViStatus rssgt_doubleQuotedString_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViStatus    error = VI_SUCCESS;
  ViUInt32    count = bufSize;
  ViInt32     idx = 0;
  ViChar      *cmd = NULL,
    *buffer = NULL;
  ViChar      format_string[1024] = "";

  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  /* Check if attribute uses repeated capabilities */
  if (sessionProperties->attr[idx].usercaps == VI_TRUE)
  {
    if (!repCapName)
      viCheckErr(RS_ERROR_INVALID_REPEATED_CAPABILITY_NAME);

    if (strlen(repCapName) == 0)
      viCheckErr(RS_ERROR_INVALID_REPEATED_CAPABILITY_NAME);
  }

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repCapName, &cmd));

  /* --- If the range table for the attribute exists, revert 'command string'
     back to the 'discreteOrMinValue' and return it as attribute value. --- */

     /* Range Table Callback */
  if (sessionProperties->attr[idx].rangeTableCallback)
  {
    viCheckErr((sessionProperties->attr[idx].rangeTableCallback)(instrSession, repCapName, attributeId, &(sessionProperties->attr[idx].rangeTable)));
  }

  /* --- Attribute data type is used for VISA I/O --- */
  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_STRING:
    if (sessionProperties->attr[idx].access == RS_VAL_READ_ONLY)
    {
      viCheckErr(rsPrintf(instrSession, "%s\n", cmd));
    }
    else
    {
      viCheckErr(rsPrintf(instrSession, "%s?\n", cmd));
    }
    viCheckErr(Rs_ReadDataUnknownLength(instrSession, &buffer, &count));
    sprintf(format_string, "\"%%%d[^\"]\"", (int)bufSize - 1);
    memset(value, 0, bufSize);
    sscanf(buffer, format_string, value);
    break;
  case RS_VAL_INT32:
  case RS_VAL_REAL64:
  case RS_VAL_BOOLEAN:
  case RS_VAL_ADDR:
  case RS_VAL_SESSION:
  case RS_VAL_EVENT:
  case RS_VAL_UNKNOWN_TYPE:
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }

Error:

  if (buffer)
    free(buffer);

  if (cmd)
    free(cmd);

  return error;
}

/*****************************************************************************
 * Function: rssgt_PathDoubleQuotedString_WriteCallback
 * Purpose:  This function overrides standard calback funtion. Writes string
 *           with apostrophes
 *****************************************************************************/
ViStatus rssgt_PathDoubleQuotedString_WriteCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViStatus    error = VI_SUCCESS;
  ViInt32     idx = 0;
  ViChar      *cmd = NULL;
  ViInt32     path = 1;
  ViChar      repcap[1024] = "";

  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  if (repCapName && strlen(repCapName) > 0)
  {
    sprintf(repcap, "%s,%s", gs_path[path], repCapName);
  }
  else
  {
    sprintf(repcap, "%s", gs_path[path]);
  }

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repcap, &cmd));

  /* --- If the range table for the attribute exists, pass 'command string' or
     'discreteOrMinValue' instead of attribute value. --- */
     /* Range Table Callback */
  if (sessionProperties->attr[idx].rangeTableCallback)
  {
    viCheckErr((sessionProperties->attr[idx].rangeTableCallback)(instrSession, repcap, attributeId, &(sessionProperties->attr[idx].rangeTable)));
  }

  if (sessionProperties->attr[idx].rangeTable)
    if (sessionProperties->attr[idx].rangeTable->type == RS_VAL_DISCRETE)
    {
      ViUInt16    i = 0;
      ViChar      *p2string;
      ViReal64    *p2discrete;
      ViInt32     data_type = sessionProperties->attr[idx].dataType;
      ViReal64    precision = sessionProperties->attr[idx].comparePrecision;

      precision = (precision < DBL_MIN) ? DBL_MIN : precision;

      /* Allowed data types are only ViInt32 or ViReal64 */
      switch (data_type)
      {
      case RS_VAL_INT32:
      case RS_VAL_REAL64:
        /* Valid data types */
        break;
      case RS_VAL_STRING:
      case RS_VAL_BOOLEAN:
      case RS_VAL_EVENT:
      case RS_VAL_ADDR:
      case RS_VAL_SESSION:
      case RS_VAL_UNKNOWN_TYPE:
      default:
        viCheckErr(RS_ERROR_INVALID_TYPE);
      }

      /* Go over range table values and find appropriate command string for passed
         attribute value. For floating point values use predefined attribute precision. */
      do
      {
        p2discrete = &sessionProperties->attr[idx].rangeTable->rangeValues[i].discreteOrMinValue;
        p2string = sessionProperties->attr[idx].rangeTable->rangeValues[i].cmdString;

        if (sessionProperties->attr[idx].rangeTable->rangeValues[i].cmdString == RS_RANGE_TABLE_END_STRING)
          return RS_ERROR_INVALID_VALUE;

        if (data_type == RS_VAL_INT32)
          if (fabs(*p2discrete - *(ViInt32 *)value) <= DBL_MIN)
          {
            viCheckErr(rsPrintf(instrSession, "%s \"%s\"\n", cmd, p2string));
            break;
          }

        if (data_type == RS_VAL_REAL64)
          if (fabs(*p2discrete - *(ViReal64 *)value) <= precision)
          {
            viCheckErr(rsPrintf(instrSession, "%s \"%s\"\n", cmd, p2string));
            break;
          }

        i++;

      } while (i);

      goto Error; /* Exit the function */
    }

  /* --- Attribute data type is used for VISA I/O --- */
  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_INT32:
    viCheckErr(rsPrintf(instrSession, "%s %ld\n", cmd, *(ViInt32 *)value));
    break;
  case RS_VAL_REAL64:
    viCheckErr(rsPrintf(instrSession, "%s %.12f\n", cmd, *(ViReal64 *)value));
    break;
  case RS_VAL_STRING:
    viCheckErr(rsPrintf(instrSession, "%s \"%s\"\n", cmd, value));
    break;
  case RS_VAL_BOOLEAN:
    viCheckErr(rsPrintf(instrSession, "%s %s\n", cmd, (*(ViBoolean *)value == VI_FALSE) ? "OFF" : "ON"));
    break;
  case RS_VAL_EVENT:
    viCheckErr(rsPrintf(instrSession, "%s\n", cmd));
    break;
  case RS_VAL_ADDR:
  case RS_VAL_SESSION:
  case RS_VAL_UNKNOWN_TYPE:
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }

Error:

  if (cmd)
    free(cmd);

  return error;

}

/*****************************************************************************
 * Function: rssgt_doubleQuotedString_WriteCallback
 * Purpose:  This function overrides standard calback funtion. Writes string
 *           with apostrophes
 *****************************************************************************/
ViStatus rssgt_doubleQuotedString_WriteCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViStatus    error = VI_SUCCESS;
  ViInt32     idx = 0;
  ViChar      *cmd = NULL;

  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  /* Check if attribute uses repeated capabilities */
  if (sessionProperties->attr[idx].usercaps == VI_TRUE)
  {
    if (!repCapName)
      viCheckErr(RS_ERROR_INVALID_REPEATED_CAPABILITY_NAME);

    if (strlen(repCapName) == 0)
      viCheckErr(RS_ERROR_INVALID_REPEATED_CAPABILITY_NAME);
  }

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repCapName, &cmd));

  /* --- If the range table for the attribute exists, pass 'command string' or
     'discreteOrMinValue' instead of attribute value. --- */
     /* Range Table Callback */
  if (sessionProperties->attr[idx].rangeTableCallback)
  {
    viCheckErr((sessionProperties->attr[idx].rangeTableCallback)(instrSession, repCapName, attributeId, &(sessionProperties->attr[idx].rangeTable)));
  }

  if (sessionProperties->attr[idx].rangeTable)
    if (sessionProperties->attr[idx].rangeTable->type == RS_VAL_DISCRETE)
    {
      ViUInt16    i = 0;
      ViChar      *p2string;
      ViReal64    *p2discrete;
      ViInt32     data_type = sessionProperties->attr[idx].dataType;
      ViReal64    precision = sessionProperties->attr[idx].comparePrecision;

      precision = (precision < DBL_MIN) ? DBL_MIN : precision;

      /* Allowed data types are only ViInt32 or ViReal64 */
      switch (data_type)
      {
      case RS_VAL_INT32:
      case RS_VAL_REAL64:
        /* Valid data types */
        break;
      case RS_VAL_STRING:
      case RS_VAL_BOOLEAN:
      case RS_VAL_EVENT:
      case RS_VAL_ADDR:
      case RS_VAL_SESSION:
      case RS_VAL_UNKNOWN_TYPE:
      default:
        viCheckErr(RS_ERROR_INVALID_TYPE);
      }

      /* Go over range table values and find appropriate command string for passed
         attribute value. For floating point values use predefined attribute precision. */
      do
      {
        p2discrete = &sessionProperties->attr[idx].rangeTable->rangeValues[i].discreteOrMinValue;
        p2string = sessionProperties->attr[idx].rangeTable->rangeValues[i].cmdString;

        if (sessionProperties->attr[idx].rangeTable->rangeValues[i].cmdString == RS_RANGE_TABLE_END_STRING)
          return RS_ERROR_INVALID_VALUE;

        if (data_type == RS_VAL_INT32)
          if (fabs(*p2discrete - *(ViInt32 *)value) <= DBL_MIN)
          {
            viCheckErr(rsPrintf(instrSession, "%s \"%s\"\n", cmd, p2string));
            break;
          }

        if (data_type == RS_VAL_REAL64)
          if (fabs(*p2discrete - *(ViReal64 *)value) <= precision)
          {
            viCheckErr(rsPrintf(instrSession, "%s \"%s\"\n", cmd, p2string));
            break;
          }

        i++;

      } while (i);

      goto Error; /* Exit the function */
    }

  /* --- Attribute data type is used for VISA I/O --- */
  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_INT32:
    viCheckErr(rsPrintf(instrSession, "%s %ld\n", cmd, *(ViInt32 *)value));
    break;
  case RS_VAL_REAL64:
    viCheckErr(rsPrintf(instrSession, "%s %.12f\n", cmd, *(ViReal64 *)value));
    break;
  case RS_VAL_STRING:
    viCheckErr(rsPrintf(instrSession, "%s \"%s\"\n", cmd, value));
    break;
  case RS_VAL_BOOLEAN:
    viCheckErr(rsPrintf(instrSession, "%s %s\n", cmd, (*(ViBoolean *)value == VI_FALSE) ? "OFF" : "ON"));
    break;
  case RS_VAL_EVENT:
    viCheckErr(rsPrintf(instrSession, "%s\n", cmd));
    break;
  case RS_VAL_ADDR:
  case RS_VAL_SESSION:
  case RS_VAL_UNKNOWN_TYPE:
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }

Error:

  if (cmd)
    free(cmd);

  return error;

}

/*****************************************************************************
 * Function: rssgt_path_WriteCallback
 * Purpose:  This function overrides standard calback funtion. Formats path
 *           into command
 *****************************************************************************/
ViStatus rssgt_Path_WriteCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViChar      *buffer = NULL;
  ViStatus    error = VI_SUCCESS;
  ViInt32     idx = 0;
  ViChar      *cmd = NULL;
  ViChar      *p2buffer = NULL;
  ViInt32     path = 1;
  ViChar      repcap[1024] = "";

  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  if (repCapName && strlen(repCapName) > 0)
  {
    sprintf(repcap, "%s,%s", gs_path[path], repCapName);
  }
  else
  {
    sprintf(repcap, "%s", gs_path[path]);
  }

  if (sessionProperties->attr[idx].dataType == RS_VAL_STRING)
  {
    buffer = (ViChar*)malloc(strlen((ViChar*)value) + RS_MAX_MESSAGE_BUF_SIZE);
  }
  else
  {
    buffer = (ViChar*)malloc(RS_MAX_MESSAGE_BUF_SIZE);
  }
  viCheckAlloc(buffer);
  p2buffer = buffer;

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repcap, &cmd));

  if ((sessionProperties->attr[idx].flags & RS_VAL_WAIT_FOR_OPC_AFTER_WRITES) != 0)
  {
    checkErr(Rs_ClearBeforeRead(instrSession));
    /* to appease lint */
    p2buffer += snprintf(p2buffer, 10, "*CLS;");
  }
  /* --- If the range table for the attribute exists, pass 'command string' or
     'discreteOrMinValue' instead of attribute value. --- */

     /* Range Table Callback */
  if (sessionProperties->attr[idx].rangeTableCallback)
  {
    viCheckErr((sessionProperties->attr[idx].rangeTableCallback)(instrSession, repcap, attributeId, &(sessionProperties->attr[idx].rangeTable)));
  }

  if (sessionProperties->attr[idx].rangeTable)
  {
    if (sessionProperties->attr[idx].rangeTable->type == RS_VAL_DISCRETE)
    {
      ViUInt16    i = 0;
      ViChar      *p2string;
      ViReal64    *p2discrete;
      ViInt32     data_type = sessionProperties->attr[idx].dataType;
      ViReal64    precision = sessionProperties->attr[idx].comparePrecision;

      precision = (precision < DBL_MIN) ? DBL_MIN : precision;

      /* Allowed data types are only ViInt32 or ViReal64 */
      switch (data_type)
      {
      case RS_VAL_INT32:
      case RS_VAL_REAL64:
        /* Valid data types */
        break;
      case RS_VAL_STRING:
      case RS_VAL_BOOLEAN:
      case RS_VAL_EVENT:
      case RS_VAL_ADDR:
      case RS_VAL_SESSION:
      case RS_VAL_UNKNOWN_TYPE:
      default:
        viCheckErr(RS_ERROR_INVALID_TYPE);
      }

      /* Go over range table values and find appropriate command string for passed
         attribute value. For floating point values use predefined attribute precision. */
      do
      {
        p2discrete = &sessionProperties->attr[idx].rangeTable->rangeValues[i].discreteOrMinValue;
        p2string = sessionProperties->attr[idx].rangeTable->rangeValues[i].cmdString;

        if (sessionProperties->attr[idx].rangeTable->rangeValues[i].cmdValue == RS_RANGE_TABLE_END_VALUE)
          viCheckErr(RS_ERROR_INVALID_VALUE);

        if (data_type == RS_VAL_INT32)
          if (fabs(*p2discrete - *(ViInt32 *)value) <= DBL_MIN)
          {
            break;
          }

        if (data_type == RS_VAL_REAL64)
          if (fabs(*p2discrete - *(ViReal64 *)value) <= precision)
          {
            break;
          }

        i++;

      } while (i != 0);
      p2buffer += snprintf(p2buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s %s", cmd, p2string);
      strcat(buffer, "\n");

      viCheckErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));
      goto Error; /* Exit the function */
    }
  }
  /* --- Attribute data type is used for VISA I/O --- */
  /* Check Wait for OPC after writes flag */
  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_INT32:
    (void)snprintf(buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s " _PERCD, cmd, *(ViInt32 *)value);
    break;
  case RS_VAL_INT64:
    (void)snprintf(buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s " _PERCLD, cmd, *(ViInt64 *)value);
    break;
  case RS_VAL_REAL64:
    //sprintf(buffer, "%s %.12lf", cmd, *(ViReal64 *)value);
    //Modified to fix CVI bug formating 1.0e+11 to 0:00000.0000
    p2buffer += snprintf(p2buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s %.15lG", cmd, *(ViReal64 *)value);
    break;
  case RS_VAL_STRING:
    p2buffer += snprintf(p2buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s \'%s\'", cmd, (ViString)value);
    break;
  case RS_VAL_BOOLEAN:
    p2buffer += snprintf(p2buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s %s", cmd, (*(ViBoolean *)value == VI_FALSE) ? "OFF" : "ON");
    break;
  case RS_VAL_EVENT:
    p2buffer += snprintf(p2buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s", cmd);
    break;
  case RS_VAL_ADDR:
  case RS_VAL_SESSION:
  case RS_VAL_UNKNOWN_TYPE:
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }

  if (sessionProperties->attr[idx].flags & RS_VAL_WAIT_TO_CONTINUE)
    strcat(buffer, ";*WAI");

  //if (sessionProperties -> attr[idx].flags & RS_VAL_WAIT_FOR_OPC_AFTER_WRITES)
  //    strcat (buffer,";*OPC");

  strcat(buffer, "\n");

  viCheckErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));

Error:
  if (cmd)
    free(cmd);
  if (buffer)
    free(buffer);

  return error;
}

/*****************************************************************************
 * Function: rssgt_path_ReadCallback
 * Purpose:  This function overrides standard calback funtion. Formats path
 *           into command
 *****************************************************************************/
ViStatus rssgt_Path_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViStatus    error = VI_SUCCESS;
  ViInt32     idx = 0;
  ViChar      *cmd = NULL;
  ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE] = "";
  ViChar      *p2buf = NULL;
  ViChar      format_string[RS_MAX_MESSAGE_LEN] = "";
  ViUInt32    buf_size = 0;
  ViInt32     path = 1;
  ViChar      repcap[1024] = "";

  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  if (repCapName && strlen(repCapName) > 0)
  {
    sprintf(repcap, "%s,%s", gs_path[path], repCapName);
  }
  else
  {
    sprintf(repcap, "%s", gs_path[path]);
  }

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repcap, &cmd));

  /* --- If the range table for the attribute exists, revert 'command string'
     back to the 'discreteOrMinValue' and return it as attribute value. --- */

     /* Range Table Callback */
  if (sessionProperties->attr[idx].rangeTableCallback)
  {
    viCheckErr((sessionProperties->attr[idx].rangeTableCallback)(instrSession, repcap, attributeId, &(sessionProperties->attr[idx].rangeTable)));
  }

  if (sessionProperties->attr[idx].rangeTable)
    if (sessionProperties->attr[idx].rangeTable->type == RS_VAL_DISCRETE)
    {
      ViUInt16    i = 0;
      ViChar      *p2string;
      ViInt32     found;

      buffer[0] = '\0';

      if (sessionProperties->attr[idx].access == RS_VAL_READ_ONLY)
        (void)snprintf(buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s\n", cmd);
      else
        (void)snprintf(buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s?\n", cmd);
      viCheckErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));
      viCheckErr(rsRead(instrSession, (ViBuf)buffer, RS_MAX_MESSAGE_BUF_SIZE, NULL));

      if (sscanf(buffer, "%[^\r\n]", buffer) != 1)
        checkErr(RS_ERROR_INVALID_VALUE);

      /* Go over range table values and check if the command string is present */
      do
      {
        p2string = sessionProperties->attr[idx].rangeTable->rangeValues[i].cmdString;

        if (sessionProperties->attr[idx].rangeTable->rangeValues[i].cmdValue == RS_RANGE_TABLE_END_VALUE)
        {
          viCheckErr(RS_ERROR_INVALID_VALUE);
        }
#ifdef RS_LONG_RESPONSE            
        if ((strlen(p2string) <= 3) || (strlen(buffer) > 4))
          found = strcmp(buffer, p2string);
        else
          found = strncmp(buffer, p2string, strlen(p2string));
#else
        found = strcmp(buffer, p2string);
#endif
        if (found == 0)
          break;

        i++;

      } while (i != 0);

      /* Return value from range table */
      switch (sessionProperties->attr[idx].dataType)
      {
      case RS_VAL_INT32:
        *(ViInt32 *)value = (ViInt32)sessionProperties->attr[idx].rangeTable->rangeValues[i].discreteOrMinValue;
        break;
      case RS_VAL_REAL64:
        *(ViReal64 *)value = sessionProperties->attr[idx].rangeTable->rangeValues[i].discreteOrMinValue;
        break;
      case RS_VAL_STRING:
      case RS_VAL_BOOLEAN:
      case RS_VAL_ADDR:
      case RS_VAL_SESSION:
      case RS_VAL_EVENT:
      case RS_VAL_UNKNOWN_TYPE:
      default:
        viCheckErr(RS_ERROR_INVALID_TYPE);
      }

      goto Error;
    }
  if (sessionProperties->attr[idx].access == RS_VAL_READ_ONLY)
  {
    (void)snprintf(buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s\n", cmd);
    viCheckErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));
  }
  else
  {
    (void)snprintf(buffer, RS_MAX_MESSAGE_BUF_SIZE, "%s?\n", cmd);
    viCheckErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));
  }
  //viCheckErr (rsRead (instrSession, (ViBuf) buffer, RS_MAX_MESSAGE_BUF_SIZE, NULL));
  viCheckErr(Rs_ReadDataUnknownLength(instrSession, &p2buf, &buf_size));
  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_INT32:
    *((ViInt32*)value) = atol(p2buf);
    break;
  case RS_VAL_INT64:
    if (sscanf(p2buf, _PERCLD, (ViInt64*)value) != 1)
      checkErr(RS_ERROR_INVALID_VALUE);
    break;
  case RS_VAL_REAL64:
    *((ViReal64*)value) = atof(p2buf);
    break;
  case RS_VAL_STRING:
    if (strspn(p2buf, "\"") != 0)
      (void)snprintf(format_string, RS_MAX_MESSAGE_LEN, "%%*1[\"]%%" _PERCD "[^\"]", bufSize - 1);
    else if (strspn(p2buf, "'") != 0)
      (void)snprintf(format_string, RS_MAX_MESSAGE_LEN, "%%*1[']%%" _PERCD "[^']", bufSize - 1);
    else
      (void)snprintf(format_string, RS_MAX_MESSAGE_LEN, "%%" _PERCD "[^\r\n]", bufSize - 1);

    memset(value, 0, bufSize);
    (void)sscanf(p2buf, format_string, (char*)value);
    break;
  case RS_VAL_BOOLEAN:
    if (strncmp(p2buf, "ON", 2) == 0)
      *(ViBoolean*)value = 1;
    else
    {
      if (strncmp(p2buf, "OFF", 3) == 0)
        *(ViBoolean*)value = 0;
      else
        (void)sscanf(p2buf, "%hu", (ViBoolean*)value);
    }
    break;
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }

  if (buf_size > (ViUInt32)bufSize)
    error = VI_SUCCESS_MAX_CNT;

Error:

  if (cmd)
    free(cmd);
  if (p2buf)
    free(p2buf);

  return error;
}


/*****************************************************************************
* Function: rssgt_path_ReadCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathBool_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  ViStatus    error = VI_SUCCESS;

  viCheckErr(rssgt_Path_ReadCallback(instrSession, repCapName, attributeId, bufSize, value));
Error:
  return error;
}

/*****************************************************************************
* Function: rssgt_PathDbl_ReadCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathDbl_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  ViStatus    error = VI_SUCCESS;

  viCheckErr(rssgt_Path_ReadCallback(instrSession, repCapName, attributeId, bufSize, value));
Error:
  return error;
}

/*****************************************************************************
* Function: rssgt_PathReadOnlyCollection_ReadCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathReadOnlyCollection_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  RsSessionPropertiesPtr  sessionProperties = Rs_ViSession(instrSession);
  ViStatus    error = VI_SUCCESS;
  ViInt32     idx = 0;
  ViUInt32    ret_cnt = 0;
  ViChar      *cmd = NULL,
    buffer[RS_MAX_MESSAGE_BUF_SIZE];
  ViChar*  	ext_buf = NULL;
  ViInt32     path = 1;
  ViChar      repcap[1024] = "";

  viCheckErr(Rs_GetAttributeIndex(instrSession, attributeId, &idx));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  /* If no command available, exit the function */
  if (sessionProperties->attr[idx].command == NULL)
    return VI_SUCCESS;

  if (repCapName && strlen(repCapName) > 0)
  {
    sprintf(repcap, "%s,%s", gs_path[path], repCapName);
  }
  else
  {
    sprintf(repcap, "%s", gs_path[path]);
  }

  /* Build the command string */
  viCheckErr(Rs_BuildCommandString(instrSession, attributeId, idx, repcap, &cmd));

  /* --- Attribute data type is used for VISA I/O --- */
  switch (sessionProperties->attr[idx].dataType)
  {
  case RS_VAL_STRING:
    strcpy(buffer, cmd);
    if (sessionProperties->attr[idx].access == RS_VAL_READ_ONLY)
      strcat(buffer, "\n");
    else
      strcat(buffer, "?\n");

    viCheckErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));

    viCheckErr(Rs_ReadDataUnknownLength(instrSession, &ext_buf, &ret_cnt));
    if ((ViInt32)ret_cnt > bufSize)
      error = VI_SUCCESS_MAX_CNT;
    ext_buf[strcspn(ext_buf, "\r\n")] = '\0';
    if ((ViInt32)ret_cnt > bufSize) ret_cnt = bufSize - 1;
    memset(value, 0, bufSize);
    strncpy(value, ext_buf, ret_cnt);


    break;
  default:
    viCheckErr(RS_ERROR_INVALID_TYPE);
  }

Error:
  if (cmd)
    free(cmd);

  if (ext_buf)
    free(ext_buf);

  return error;
}

/*****************************************************************************
* Function: rssgt_PathInt32_ReadCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathInt32_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  ViStatus    error = VI_SUCCESS;

  viCheckErr(rssgt_Path_ReadCallback(instrSession, repCapName, attributeId, bufSize, value));
Error:
  return error;
}

/*****************************************************************************
* Function: rssgt_PathString_ReadCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathString_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
  ViStatus    error = VI_SUCCESS;

  viCheckErr(rssgt_Path_ReadCallback(instrSession, repCapName, attributeId, bufSize, value));
Error:
  return error;
}

/*****************************************************************************
* Function: rssgt_PathBool_WriteCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathBool_WriteCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, void *value)
{
  ViStatus    error = VI_SUCCESS;

  viCheckErr(rssgt_Path_WriteCallback(instrSession, repCapName, attributeId, value));
Error:
  return error;
}

/*****************************************************************************
* Function: rssgt_PathEvent_WriteCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathEvent_WriteCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, void *value)
{
  ViStatus    error = VI_SUCCESS;

  viCheckErr(rssgt_Path_WriteCallback(instrSession, repCapName, attributeId, value));
Error:
  return error;
}

/*****************************************************************************
* Function: rssgt_PathDbl_WriteCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathDbl_WriteCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, void *value)
{
  ViStatus    error = VI_SUCCESS;

  viCheckErr(rssgt_Path_WriteCallback(instrSession, repCapName, attributeId, value));
Error:
  return error;
}
/*****************************************************************************
* Function: rssgt_PathString_WriteCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathString_WriteCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, void *value)
{
  ViStatus    error = VI_SUCCESS;

  viCheckErr(rssgt_Path_WriteCallback(instrSession, repCapName, attributeId, value));
Error:
  return error;
}
/*****************************************************************************
* Function: rssgt_PathInt32_WriteCallback
* Purpose:  This function overrides standard calback funtion. Formats path
*           into command
*****************************************************************************/
ViStatus rssgt_PathInt32_WriteCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, void *value)
{
  ViStatus    error = VI_SUCCESS;

  viCheckErr(rssgt_Path_WriteCallback(instrSession, repCapName, attributeId, value));
Error:
  return error;
}
