
# rssgtpy

A Python wrapper for the rssgt instrument driver for vector signal generator R&S SGT100A.
This package encapsulates the rssgt CVI instrument driver and offers an object oriented interface which makes programming in Python more comfortable.
 

## Installation

* Install a python interpreter

    Go to [https://www.python.org/downloads/](https://www.python.org/downloads/ "Download python") and scroll down to section `Looking for a specific release`. Find the version you want (we recommend a 3.7.xx Version), download and follow instructions. Select feature `Add python to system search path` in the installation wizard.

* Either install a binary archive in folder ```dist``` 

    Linux
    ```batch
    cd dist
    pip install ./rssgtpy-x.y.z.linux-x86_64.tar.gz
    ```

    Windows
    ```batch
    python Install.py dist\rssgtpy-x.y.z.win-amd64.zip
    ```

* Or build from sources
    	
	```batch
	python setup.py install	
	```

* Check, if the installation was succesful

	```batch
	python 
    >>> import rssgtpy
    >>>
	```
    If module rssgtpy was found there will be no error message.


* Under Linux if you require remote control via PCIe or fast socket make shared libraries libsocketcontroller.so and libpciecontroller.so accessible to the program e.g. by command
    ```batch
    export LD_LIBRARY_PATH=PATHTOLIBRARIES
    ```
* Under Windows if you require remote control via PCIe or fast socket DLLs socketcontroller.dll and pciecontroller.dll come with SGMA-GUI so make sure to install SGMA-GUI. 

* For more information about high speed remote control see
 https://www.rohde-schwarz.com/applications/high-speed-remote-control-of-r-s-sgma-rf-sources-application-note_56280-241731.html

## Example



```python
import rssgtpy as rssgt

# replace by resource string of your device 
resourceString = "tcpip::172.29.0.47::instr"

sgt1 = rssgt.Instrument(resourceString)
if sgt1.status == 0:
    try:
        # print *IDN
        print(sgt1.IDQueryResponse())

        # set / get RF frequency
        f1 = 4.56e9
        print ("Setting frequency to %10.2f GHz" % (f1/1e9))
        sgt1.SetAttributeViReal64(rssgt.ATTR_RF_FREQUENCY, f1)
        f = sgt1.GetAttributeViReal64(rssgt.ATTR_RF_FREQUENCY)
        print ("Read: %10.2f Ghz" % (f/1e9))

        f2 = 1.23e9
        print ("Setting frequency to %10.2f GHz" % (f2/1e9))
        sgt1.SetAttributeViReal64(rssgt.ATTR_RF_FREQUENCY, f2)
        f = sgt1.GetAttributeViReal64(rssgt.ATTR_RF_FREQUENCY)
        print ("Read: %10.2f Ghz" % (f/1e9))

        # error handling
        print ("Error handling:")
        sgt1.SetAttributeViReal64(rssgt.ATTR_RF_FREQUENCY, 888e9)
        if sgt1.status:
            print("Status = %d, %s\n" % (sgt1.status, sgt1.statusMessage()))
        else:        
            print ("Error status expected!") 

        # provoke exception
        print ("Test exception:")
        sgt1.exceptionhandling = rssgt.EXCEPTION_HANDLING.WARNING
        sgt1.SetAttributeViReal64(rssgt.ATTR_RF_FREQUENCY, 999e9)
        print ("Exception expected!")
        
    except RuntimeError:
        print ("Exception occurred: %s" % sgt1.statusMessage())

    finally:
        sgt1.close()

else:
    print("cannot open %s" % resourceString)
```

## Documentation

See files ``rssgt_attr.chm`` and ``rssgt_vxi.chm`` in folder ``src/InstrumentDriver`` for a documentation of attributes and functions.

## Developer

### Windows

Execute
```batch
makeSolution.cmd
```

to create a Visual Studio solution and project in folder build.
Open rssgtpy.sln in Visual Studio and build project rssgtpy.


### Linux
```batch
python setup.py install	
```
or 
```batch
python setup.py develop
```

### Build binary archive
```batch
python setup.py bdist	
```

