/*****************************************************************************
 *  Rohde&Schwarz SGT Signal Generator
 *
 *  Original Release: January 2014
 *  By: Petr Kominek
 *
 *  Should you have any technical questions please contact the hotline of
 *  Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
 *
 *  e-mail: CustomerSupport@rohde-schwarz.com
 *
 *  Modification History:
 *     see ChangeLog
 *
 *****************************************************************************/

#include <string.h>  
#include <stdio.h>  
#include <stdlib.h>
#include <math.h>
#include "rssgt_utility.h"
#include "rssgt.h"
#include "SgmaCommandIds.h"
#include "CallLogger.h"

#include "time.h"
#if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__))
#define ANSI
#endif
#ifndef ANSI                /* UNIX compatible */
#include <sys/timeb.h>
#include <sys/types.h>
#include <unistd.h>
#define Sleep sleep
#else                       /* ANSI compatible */
#include <windows.h>
#endif

#define BUFFER_SIZE             2048L    /* File I/O buffer size */

#if defined(_MSC_VER)
#define snprintf _snprintf
#endif

 /*****************************************************************************
  *- Value Definition and Range Tables ---------------------------------------*
  *****************************************************************************/
static ViString enabledArr[] = { "OFF","ON", NULL };
static ViString regGroupArr[] = { "OPER","QUES", NULL };
static ViString regTypeArr[] = { "COND","EVEN","ENAB","NTR","PTR", NULL };
static ViString ARBNextAction[] = { "NEXT", "BLAN", "ENDL", "SEG0", "SEG1", "SEG2",
                   "SEG3", "SEG4", "SEG5", "SEG6", "SEG7", "SEG8",
                   "SEG9", "SEG10", "SEG11", "SEG12", "SEG13", "SEG14",
                   "SEG15", "SEG16", "SEG17", "SEG18", "SEG19", "SEG20",
                   "SEG21", "SEG22", "SEG23", "SEG24", "SEG25", "SEG26",
                   "SEG27", "SEG28", "SEG29", "SEG30", "SEG31", NULL };

/*****************************************************************************
 *- Callback Declarations ---------------------------------------------------*
 *****************************************************************************/
ViStatus _VI_FUNC rssgt_RsClose(ViSession instrSession);
ViStatus get_error_name_and_message(ViStatus statusCode, ViChar _VI_FAR errName[], ViChar _VI_FAR errMessage[]);

#define RSSGT_DEFAULT_TIMEOUT   10000   /* Default Timeout value in ms (10 sec) */


/*****************************************************************************
 *- User-Callable Functions (Exportable Functions) --------------------------*
 *****************************************************************************/


 /// HIFN  This function performs the following initialization actions:
 /// HIFN     
 /// HIFN     - Creates a new instrument driver session.
 /// HIFN     
 /// HIFN     - Opens a session to the specified device using the interface and 
 /// HIFN  address you specify for the Resource Name parameter.
 /// HIFN     
 /// HIFN     - If the ID Query parameter is set to VI_TRUE, this function queries 
 /// HIFN  the instrument ID and checks that it is valid for this instrument 
 /// HIFN  driver.
 /// HIFN     
 /// HIFN     - If the Reset parameter is set to VI_TRUE, this function resets the 
 /// HIFN  instrument to a known state.
 /// HIFN     
 /// HIFN     - Sends initialization commands to set the instrument to the state 
 /// HIFN  necessary for the operation of the instrument driver.
 /// HIFN     
 /// HIFN     - Returns a ViSession handle that you use to identify the instrument 
 /// HIFN  in all subsequent instrument driver function calls.
 /// HIFN     
 /// HIFN     Note:
 /// HIFN     
 /// HIFN     This function creates a new session each time you invoke it. Although 
 /// HIFN  you can open more than one session for the same resource, it is best 
 /// HIFN  not to do so. You can use the same session in multiple program threads. 
 /// HIFN  You can use the prefix_LockSession and prefix_UnlockSession functions 
 /// HIFN  to protect sections of code that require exclusive access to the 
 /// HIFN  resource.
 /// HIFN     
 /// HIRET Returns the status code of this operation.
 /// HIPAR resourceName/Pass the resource name of the device to initialize.
 /// HIPAR idQuery/Specify whether you want the instrument driver to perform an ID Query.
 /// HIPAR resetDevice/Specify whether you want the to reset the instrument during the 
 /// HIPAR resetDevice/initialization procedure.
 /// HIPAR instrSession/Returns a ViSession handle that you use to identify the instrument in 
 /// HIPAR instrSession/all subsequent instrument driver function calls.
ViStatus _VI_FUNC rssgt_init(
  ViRsrc resourceName,
  ViBoolean idQuery,
  ViBoolean resetDevice,
  ViSession* instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddStringArg(resourceName);
    LogAddBoolArg(idQuery);
    LogAddBoolArg(resetDevice);
    LogAddSessionOut(instrSession);
  }
  ViStatus    error = VI_SUCCESS;

  if (instrSession == 0) {
    (void)Rs_SetErrorInfo(0, VI_FALSE, RS_ERROR_INVALID_PARAMETER,
      VI_ERROR_PARAMETER4, "Null address for Instrument Handle");
    checkErr(RS_ERROR_INVALID_PARAMETER);
  }
  else {
    checkErr(rssgt_InitWithOptions(resourceName, idQuery, resetDevice,
      "QueryInstrStatus=1", instrSession));
  }

Error:
  LogCloseCall(error);
  return error;
}



/// HIFN  This function performs the following initialization actions:
/// HIFN     
/// HIFN     - Creates a new instrument driver session and optionally sets the 
/// HIFN  initial state of the following session attributes:
/// HIFN     
/// HIFN         prefix_ATTR_RANGE_CHECK
/// HIFN         prefix_ATTR_QUERY_INSTRUMENT_STATUS
/// HIFN         prefix_ATTR_CACHE (currently not used)
/// HIFN         prefix_ATTR_SIMULATE
/// HIFN         prefix_ATTR_RECORD_COERCIONS (currently not used)
/// HIFN     
/// HIFN     - Opens a session to the specified device using the interface and 
/// HIFN  address you specify for the Resource Name parameter.
/// HIFN     
/// HIFN     - If the ID Query parameter is set to VI_TRUE, this function queries 
/// HIFN  the instrument ID and checks that it is valid for this instrument 
/// HIFN  driver.
/// HIFN     
/// HIFN     - If the Reset parameter is set to VI_TRUE, this function resets the 
/// HIFN  instrument to a known state.
/// HIFN     
/// HIFN     - Sends initialization commands to set the instrument to the state 
/// HIFN  necessary for the operation of the instrument driver.
/// HIFN     
/// HIFN     - Returns a ViSession handle that you use to identify the instrument 
/// HIFN  in all subsequent instrument driver function calls.
/// HIFN     
/// HIFN     Note:
/// HIFN     
/// HIFN     This function creates a new session each time you invoke it. Although 
/// HIFN  you can open more than one session for the same resource, it is best 
/// HIFN  not to do so. You can use the same session in multiple program threads. 
/// HIFN  You can use the prefix_LockSession and prefix_UnlockSession functions 
/// HIFN  to protect sections of code that require exclusive access to the 
/// HIFN  resource.
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR resourceName/Pass the resource name of the device to initialize.
/// HIPAR idQuery/Specify whether you want the instrument driver to perform an ID Query.
/// HIPAR resetDevice/Specify whether you want the to reset the instrument during the 
/// HIPAR resetDevice/initialization procedure.
/// HIPAR optionString/You can use this control to set the initial value of certain 
/// HIPAR optionString/attributes for the session. The following table lists the attributes 
/// HIPAR optionString/and the name you use in this parameter to identify the attribute.
/// HIPAR newInstrSession/Returns a ViSession handle that you use to identify the instrument in 
/// HIPAR newInstrSession/all subsequent instrument driver function calls.
ViStatus _VI_FUNC rssgt_InitWithOptions(
  ViRsrc resourceName,
  ViBoolean idQuery,
  ViBoolean resetDevice,
  ViString optionString,
  ViSession* newInstrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddStringArg(resourceName);
    LogAddBoolArg(idQuery);
    LogAddBoolArg(resetDevice);
    LogAddStringArg(optionString);
    LogAddSessionOut(newInstrSession);
  }
  ViStatus                error = VI_SUCCESS;
  ViChar                  newResourceName[RS_MAX_MESSAGE_BUF_SIZE] = "";
  ViChar                  newOptionString[RS_MAX_MESSAGE_BUF_SIZE] = "";
  ViBoolean               isLogicalName = 0;
  ViSession               instrSession = 0, rmSession = 0;
  RsAttrPropertiesPtr     *attrList = NULL;
  RsRepCapPtr             repCapTable = NULL;

  if (newInstrSession == 0)
  {
    (void)Rs_SetErrorInfo(0, VI_FALSE, RS_ERROR_INVALID_PARAMETER,
      VI_ERROR_PARAMETER5, "Null address for Instrument Handle");
    checkErr(RS_ERROR_INVALID_PARAMETER);
  }

  checkErr(rsOpenDefaultRM(&rmSession));

  viCheckErr(rsOpen(rmSession, resourceName, 0, 0, &instrSession));


  viCheckErr(rsSetAttribute(instrSession, VI_ATTR_TMO_VALUE, RSSGT_DEFAULT_TIMEOUT));
  viCheckErr(rsSetAttribute(instrSession, VI_ATTR_TERMCHAR_EN, VI_TRUE));


  viCheckErr(rsSetBuf(instrSession, (ViUInt16)(VI_READ_BUF | VI_WRITE_BUF), 4000));
  viCheckErr(rsSetAttribute(instrSession, VI_ATTR_WR_BUF_OPER_MODE, VI_FLUSH_ON_ACCESS));
  viCheckErr(rsSetAttribute(instrSession, VI_ATTR_RD_BUF_OPER_MODE, VI_FLUSH_ON_ACCESS));

  checkErr(Rs_GetInfoFromResourceName(resourceName,
    (ViString)optionString,
    newResourceName,
    newOptionString,
    &isLogicalName));

  ATTRLIST(attrList, rssgt);
  checkErr(Rs_SpecificDriverNew(instrSession, "rssgt", newOptionString, attrList));

  REPCAPTABLE(repCapTable, rssgt);
  checkErr(Rs_BuildRepCapTable(instrSession, repCapTable));


  checkErr(Rs_SetAttribute(instrSession, "", RS_ATTR_VISA_RM_SESSION, 0, &rmSession));

  checkErr(rsWrite(instrSession, (ViBuf) "&GTR\n", 5, NULL));

  if (resetDevice == VI_TRUE)
  {
    checkErr(rssgt_reset(instrSession));
  }
  else
    checkErr(rssgt_DefaultInstrSetup(instrSession));

  if (idQuery == VI_TRUE)
  {
    ViChar rdBuffer[RSSGT_IO_BUFFER_SIZE] = "";

    checkErr(Rs_GetAttribute(instrSession, NULL, (ViAttr)RSSGT_ATTR_ID_QUERY_RESPONSE,
      0, RSSGT_IO_BUFFER_SIZE, rdBuffer));

    if (strncmp(rdBuffer, RSSGT_VALID_ID_RESPONSE_STRING, strlen(RSSGT_VALID_ID_RESPONSE_STRING)) != 0)
    {
      viCheckErr(VI_ERROR_FAIL_ID_QUERY);
    }
  }

  if (isLogicalName == VI_FALSE)
  {
    ViInt32 oldFlag = 0;

    checkErr(Rs_GetAttributeFlags(instrSession, RS_ATTR_IO_RESOURCE_DESCRIPTOR, &oldFlag));
    checkErr(Rs_SetAttributeFlags(instrSession, RS_ATTR_IO_RESOURCE_DESCRIPTOR, (oldFlag & 0xfffb) | 0x0010));
    checkErr(Rs_SetAttribute(instrSession, "", RS_ATTR_IO_RESOURCE_DESCRIPTOR, 0, newResourceName));
  }


  if (isLogicalName == VI_TRUE)
    checkErr(Rs_ApplyDefaultSetup(instrSession));

  checkErr(rssgt_CheckStatus(instrSession));

Error:

  if (error & _VI_ERROR)
  {
    ViSession   rmSess = 0;

    (void)Rs_GetAttribute(instrSession, NULL, RS_ATTR_VISA_RM_SESSION, 0, (unsigned int) sizeof(rmSess), &rmSess);

    if (instrSession != 0)
      (void)rsClose(instrSession);
    (void)rsClose(rmSess);
  }
  else
  {
    if (newInstrSession)
      *newInstrSession = instrSession;
  }

  if (error & _VI_ERROR)
    (void)Rs_Dispose(instrSession);

  LogCloseCall(error);
  return error;
}



/// HIFN  This function sets the operation mode.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_BB_OPERATION_MODE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce]:OPMode
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR operationMode/Sets the operation mode.
ViStatus _VI_FUNC rssgt_ConfigureOperationMode(
  ViSession instrSession,
  ViInt32 operationMode
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(operationMode, OperationMode);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_BB_OPERATION_MODE,
    operationMode), 2, "Operation Mode");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function selects the path.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/selects the path
ViStatus _VI_FUNC rssgt_SelectActivePath(ViSession instrSession, ViInt32 path)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(path);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_PATH,
    path), 2, "Path");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function configures the ARB generator by specifying the sample 
/// HIFN  frequency, selected waveform and state.
/// HIFN     
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_WAVEFORM_SELECT
/// HIFN     RSSGT_ATTR_ARB_CLOCK_FREQUENCY
/// HIFN     RSSGT_ATTR_ARB_ENABLED
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:WAVeform:SELect
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:CLOCk
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:STATe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR waveform/Selects the waveform file. If the file is not on the default path, the 
/// HIPAR waveform/path must be specified at the same time. If no file of the specified 
/// HIPAR waveform/name exists, it is created. The file extension may be omitted. Only 
/// HIPAR waveform/files with the file extension *.wv will be created or loaded.
/// HIPAR clockFrequency/Sets the clock rate in samples. Loading a waveform sets the clock rate 
/// HIPAR clockFrequency/that is defined in the waveform tag 'clock'. This function subsequently 
/// HIPAR clockFrequency/changes the clock rate; see data sheet for value range.
/// HIPAR arbEnabled/Switches the ARB generator on. Any other standards or digital 
/// HIPAR arbEnabled/modulation that may be in the ON state are automatically turned OFF. 
/// HIPAR arbEnabled/The selected waveform is generated straight away (ARB:SEQ  
/// HIPAR arbEnabled/AUTO|RETRigger) or after the first trigger event (ARB:SEQ  AAUT|ARET), 
/// HIPAR arbEnabled/depending on the trigger setting.
ViStatus _VI_FUNC rssgt_ConfigureARB(
  ViSession instrSession,
  ViString waveform,
  ViReal64 clockFrequency,
  ViBoolean arbEnabled
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(waveform);
    LogAddReal64Arg(clockFrequency);
    LogAddBoolArg(arbEnabled);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_WAVEFORM_SELECT,
    waveform), 2, "Waveform");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_CLOCK_FREQUENCY,
    clockFrequency), 3, "Clock Frequency");
  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_ARB_ENABLED,
    arbEnabled), 4, "ARB Enabled");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the ARB Offsets.
/// HIFN     
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_BB_FREQUENCY_OFFSET
/// HIFN     RSSGT_ATTR_BB_PHASE_OFFSET
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:FOFFset
/// HIFN     [SOURce<[1]|2>]:BB:POFFset
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR frequencyOffset/Sets the frequency offset for the baseband signal. The offset affects 
/// HIPAR frequencyOffset/the signal on the baseband block output. It shifts the useful baseband 
/// HIPAR frequencyOffset/signal in the center frequency.
/// HIPAR phaseOffset/Sets the relative phase offset for the selected baseband signal 
/// HIPAR phaseOffset/compared to the baseband signals of the other baseband sources (second 
/// HIPAR phaseOffset/path or external baseband). The gain affects the signal on the baseband 
/// HIPAR phaseOffset/block output.
ViStatus _VI_FUNC rssgt_ConfigureARBOffsets(
  ViSession instrSession,
  ViReal64 frequencyOffset,
  ViReal64 phaseOffset
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(frequencyOffset);
    LogAddReal64Arg(phaseOffset);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_BB_FREQUENCY_OFFSET,
    frequencyOffset), 2, "Frequency Offset");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_BB_PHASE_OFFSET,
    phaseOffset), 3, "Phase Offset");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN writes the block data to the file identified by waveformFilename
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR waveformFilename/File name.
/// HIPAR binaryBlockLength/Defines the number of bytes in the <Data> array.
/// HIPAR data/Binary block data array.
ViStatus _VI_FUNC rssgt_WriteARBData(ViSession instrSession, ViString waveformFilename,
  ViInt32 binaryBlockLength, ViChar data[])
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(waveformFilename);
    LogAddInt32Arg(binaryBlockLength);
    LogAddCharArrArg(data, &binaryBlockLength);
  }
  ViStatus    error = VI_SUCCESS;
  ViChar*     pbuffer;
  ViChar*     p2buf;
  size_t      size = strlen(waveformFilename) + binaryBlockLength + 1000;
  ViInt32  path = 1;

  p2buf = pbuffer = (ViChar*)malloc(size);

  viCheckAlloc(pbuffer);

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  p2buf += snprintf(pbuffer, size, "SOUR" _PERCD ":BB:ARB:WAV:DATA '%s',#" _PERCD _PERCD, path,
    waveformFilename,
    (ViInt32)(floor(log10((double)binaryBlockLength)) + 1),
    binaryBlockLength);

  memcpy(p2buf, data, (size_t)binaryBlockLength);
  p2buf += binaryBlockLength;
  *p2buf++ = '\n';

  checkErr(rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)(p2buf - pbuffer), NULL));
  checkErr(rssgt_CheckStatus(instrSession));
Error:
  if (pbuffer)
    free(pbuffer);

  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN Queries the content of the specified tag of the specified Waveform file (in <Waveform Filename>).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR waveformFilename/File name.
/// HIPAR tag/Specifies the tag to be queried.
/// HIPAR bufferSize/Length of the buffer.
/// HIPAR buffer/Buffer.
ViStatus _VI_FUNC rssgt_ReadARBFileTag(ViSession instrSession,
  ViString waveformFilename, ViString tag,
  ViInt32 bufferSize, ViChar buffer[])
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(waveformFilename);
    LogAddStringArg(tag);
    LogAddInt32Arg(bufferSize);
    LogAddCharArrOut(buffer, &bufferSize);
  }
  ViStatus  error = VI_SUCCESS;
  ViInt32  path = 1;
  ViChar    firstChar[20] = "";
  ViUInt32  cnt = 0;
  ViChar    format[32] = "",
    length_of_length[10] = "";
  ViInt64   length = 0;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  checkErr(rsPrintf(instrSession, "SOUR%ld:BB:ARB:WAV:DATA? '%s','%s'\n", path, waveformFilename, tag));

  checkErr(rsRead(instrSession, (ViPBuf)firstChar, 1, NULL));

  if (*firstChar != '#')
  {
    error = RS_ERROR_READING_FILE;
    goto Error;
  }

  checkErr(rsRead(instrSession, (ViPBuf)firstChar, 1, NULL));

  cnt = (ViUInt32)atol(firstChar);

  sprintf(format, "%%%sllu", firstChar);

  checkErr(rsRead(instrSession, (ViPBuf)length_of_length, cnt, NULL));

  sscanf(length_of_length, format, &length);

  if (length)
  {
    if (length > bufferSize)
    {
      error = VI_ERROR_INV_SIZE;
      goto Error;
    }

    // read tag+terminating LF
    checkErr(rsRead(instrSession, (ViPBuf)buffer, (ViUInt32)length + 1, NULL));
    // strip LF
    buffer[length] = 0;
    checkErr(rssgt_CheckStatus(instrSession));

  }

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN Queries the content of the specified tag of the selected Waveform file.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR tag/Specifies the tag to be queried. For list of valid tags, see operating manual.
/// HIPAR bufferSize/Length of the buffer.
/// HIPAR buffer/Tag value.
ViStatus _VI_FUNC rssgt_ReadARBTag(ViSession instrSession, ViString tag,
  ViInt32 bufferSize, ViChar buffer[])
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(tag);
    LogAddInt32Arg(bufferSize);
    LogAddCharArrOut(buffer, &bufferSize);
  }
  ViStatus    error = VI_SUCCESS;
  ViChar*     p2buf = NULL;
  ViInt32  path = 1;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  checkErr(rsPrintf(instrSession, "SOUR%ld:BB:ARB:WAV:TAG? '%s'\n", path, tag));
  //checkErr (Rs_ReadDataUnknownLength (instrSession, &p2buf, &len));

  checkErr(rsRead(instrSession, (ViPBuf)buffer, (ViUInt32)bufferSize, NULL));

  checkErr(rssgt_CheckStatus(instrSession));

  /* if (p2buf)
      strncpy (buffer, p2buf, (size_t) bufferSize);
  */
Error:
  if (p2buf)
    free(p2buf);

  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  Sets all ARB generator parameters to their default values.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_SET_TO_DEFAULT
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:PRESet
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
ViStatus _VI_FUNC rssgt_ARBSetToDefault(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_SET_TO_DEFAULT, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Sets the clock of the ARB for files created in HP format.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR waveformFilename/File name.
/// HIPAR clock/Sets the clock of the ARB for files created in HP format.
ViStatus _VI_FUNC rssgt_ConfigureARBFileClock(ViSession instrSession,
  ViString waveformFilename, ViReal64 clock)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(waveformFilename);
    LogAddReal64Arg(clock);
  }
  ViStatus error = VI_SUCCESS;
  ViInt32  path = 1;

  checkErr(Rs_LockSession(instrSession, NULL));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  viCheckParm(rssgt_invalidViReal64Range(clock, 400.0, 100.0e6), 3, "Clock");

  viCheckErr(rsPrintf(instrSession, ":SOUR%ld:BB:ARB:WAV:CLOC '%s',%lf\n", path, waveformFilename, clock));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  Adds a blank segment to the multi segment file.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     -
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce<hw>]:BB:ARBitrary:WSEGment:CONFigure:BLANk:APPend
/// HIFN     <SampCount>, <Frequency>
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR sampleCount/Specifies the number of samples.
/// HIPAR clockRate/Determines the clock rate.
ViStatus _VI_FUNC rssgt_ConfigureARBMultisegmentBlank(
  ViSession instrSession,
  ViReal64 sampleCount,
  ViReal64 clockRate
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(sampleCount);
    LogAddReal64Arg(clockRate);
  }
  ViStatus error = VI_SUCCESS;
  ViInt32  path = 1;

  checkErr(Rs_LockSession(instrSession, NULL));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  viCheckErr(rsPrintf(instrSession, ":SOUR%ld:BB:ARB:WSEG:CONF:BLAN:APP %lf,%lf\n", path, sampleCount, clockRate));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Adds a blank segment to the multi segment file.
/// HIFN     
/// HIFN     Note(s):
/// HIFN     Control User Clock Rate is valid only for Clock User.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_LEVEL_MODE
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_CLOCK_MODE
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_CLOCK
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:WSEGment:CONFigure:LEVel[:MODE]
/// HIFN     [SOURce]:BB:ARBitrary:WSEGment:CONFigure:CLOCk:MODE
/// HIFN     [SOURce]:BB:ARBitrary:WSEGment:CONFigure:CLOCk
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR level/Selects the level mode for the multi segment waveform.
/// HIPAR clock/Selects the clock rate mode for the multi segment waveform.
/// HIPAR userClockRate/Defines the clock rate used for multi segment waveform output in case 
/// HIPAR userClockRate/of Clock Mode "User".
ViStatus _VI_FUNC rssgt_ConfigureARBMultisegmentLevelClockMarker(
  ViSession instrSession,
  ViInt32 level,
  ViInt32 clock,
  ViReal64 userClockRate
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(level, ARBMultisegmentLevelMode);
    LogAddSelectArg(clock, ARBMultisegmentClockMode);
    LogAddReal64Arg(userClockRate);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_LEVEL_MODE,
    level), 2, "Level");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_CLOCK_MODE,
    clock), 3, "Clock");
  if (clock == RSSGT_VAL_ARB_CLOCK_MODE_USER)
  {
    viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
      RSSGT_ATTR_ARB_MULTISEGMENT_CLOCK,
      userClockRate), 4, "User Clock Rate");
  }

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Appends a new segment to the selected sequencing play list.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     -
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce<hw>]:BB:ARBitrary:WSEGment:SEQuence:APPend  <State>, 
/// HIFN  <Segment>, <Count>, <Next>
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR state/Activates/deactivates the appended segment.
/// HIPAR segment/Indicates the number of the segment as in the multi segment waveform 
/// HIPAR segment/file.
/// HIPAR repetition/Defines how many times this segment is repeated.
/// HIPAR next/Determines the action after completing the current segment, like for 
/// HIPAR next/instance which segment is processed after the processing of the current 
/// HIPAR next/one is finished.
ViStatus _VI_FUNC rssgt_ARBMultisegmentAppendSegmentToPlayList(
  ViSession instrSession,
  ViBoolean state,
  ViInt32 segment,
  ViInt32 repetition,
  ViInt32 next
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(state);
    LogAddInt32Arg(segment);
    LogAddInt32Arg(repetition);
    LogAddInt32Arg(next);
  }
  ViStatus error = VI_SUCCESS;
  ViInt32  path = 1;

  checkErr(Rs_LockSession(instrSession, NULL));
  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_PATH, &path));

  viCheckErr(rsPrintf(instrSession, "SOUR%ld:BB:ARB:WSEG:SEQ:APP %s,%ld,%ld,%s\n", path, enabledArr[state], segment, repetition, ARBNextAction[next]));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Enters a comment for the configuration file. The configuration file 
/// HIFN  must be specified with command :BB:ARB:WSEG:CONF:SEL.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_COMMENT
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:WSEGment:CONFigure:COMMent
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR comment/Enters a comment for the configuration file. The configuration file 
/// HIPAR comment/must be specified with command :BB:ARB:WSEG:CONF:SEL.
ViStatus _VI_FUNC rssgt_ARBMultisegmentFileComment(
  ViSession instrSession,
  ViString comment
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(comment);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_COMMENT,
    comment), 2, "Comment");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function queries the available configuration files in the 
/// HIFN  specified default directory. The configuration files are used to create 
/// HIFN  multi segment waveform files.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_CATALOG
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:WSEGment:CONFigure:CATalog?
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR arraySize/Array size parameter.
/// HIPAR catalog/Queries the available configuration files in the specified default 
/// HIPAR catalog/directory. The configuration files are used to create multi segment 
/// HIPAR catalog/waveform files.
ViStatus _VI_FUNC rssgt_ARBMultisegmentCatalog(
  ViSession instrSession,
  ViInt32 arraySize,
  ViChar _VI_FAR catalog[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(arraySize);
    LogAddCharArrOut(catalog, &arraySize);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_GetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_CATALOG, arraySize,
    catalog), 3, "Catalog");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function queries the segments of the currently selected 
/// HIFN  configuration file.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_SEGMENT_CATALOG
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:WSEGment:CONFigure:SEGMent:CATalog?
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR arraySize/Array size parameter.
/// HIPAR catalog/Queries the segments of the currently selected configuration file.
ViStatus _VI_FUNC rssgt_ARBMultisegmentSegmentCatalog(
  ViSession instrSession,
  ViInt32 arraySize,
  ViChar _VI_FAR catalog[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(arraySize);
    LogAddCharArrOut(catalog, &arraySize);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_GetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_SEGMENT_CATALOG, arraySize,
    catalog), 3, "Catalog");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Creates a multi segment waveform using the current settings of the 
/// HIFN  specified configuration file. The multi segment waveform is saved into 
/// HIFN  the file defined in the configuration file. The file extension is *. 
/// HIFN  wv.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_CREATE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARB:WSEGment:CREate
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
ViStatus _VI_FUNC rssgt_ARBMultisegmentCreate(
  ViSession instrSession,
  ViString file
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(file);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_CREATE,
    file), 2, "File");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Creates a and loads a multi segment waveform using the current entries of the specified 
/// HIFN  configuration file. This multi segment waveform is saved with the file name specified 
/// HIFN  in the configuration file. The file extension is *.wv. 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR file/File Name parameter.
ViStatus _VI_FUNC rssgt_ARBMultisegmentLoadFromConfigurationFile(
  ViSession instrSession,
  ViString file
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(file);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_LOAD_FROM_CONFIGURATION_FILE,
    file), 2, "File");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Selects the configuration file in the default directory A path can 
/// HIFN  also be specified, in which case the files in the specified directory 
/// HIFN  are selected. If a configuration file with the specified name does not 
/// HIFN  yet exist, it is created. The file extension may be omitted. Only files 
/// HIFN  with the file extension *.inf_mswv will be selected or created.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_CONFIGURATION_FILE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:WSEGment:CONFigure:SELect
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR file/File Name parameter.
ViStatus _VI_FUNC rssgt_ARBMultisegmentConfigurationFile(
  ViSession instrSession,
  ViString file
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(file);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_CONFIGURATION_FILE,
    file), 2, "File");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Deletes the configuration file. The configuration files are used to 
/// HIFN  create multi segment waveform files.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_DELETE_CONFIGURATION_FILE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:WSEGment:CONFigure:DELete
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR file/File Name parameter.
ViStatus _VI_FUNC rssgt_ARBMultisegmentConfigurationFileDelete(
  ViSession instrSession,
  ViString file
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(file);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_DELETE_CONFIGURATION_FILE,
    file), 2, "File");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Appends the specified waveform to the configuration file.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTISEGMENT_APPEND_FILE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:WSEGment:CONFigure:SEGMent:APPend
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR file/File Name parameter.
ViStatus _VI_FUNC rssgt_ARBMultisegmentConfigurationFileAppend(
  ViSession instrSession,
  ViString file
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(file);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTISEGMENT_APPEND_FILE,
    file), 2, "File");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Sets all ARB multi carrier parameters to their default values.
/// HIFN     
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_RESET
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:PRESet
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
ViStatus _VI_FUNC rssgt_ARBMultiCarrierSetToDefault(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_RESET, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the Multi Carrier.
/// HIFN     
/// HIFN     Note(s):
/// HIFN     (1) Control Carrier Spacing is valid only for Mode Equidistant.
/// HIFN     (2) Control Signal Period is valid only for Signal Period Mode User
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_MODE
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_COUNT
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_SPACING
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_CREST_FACTOR_MODE
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_SIGNAL_PERIOD_MODE
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_SIGNAL_PERIOD
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_POWER_REFERENCE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce<hw>]:BB:ARBitrary:MCARrier:CARRier:MODE
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CARRier:COUNt
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CARRier:SPACing
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CFACtor:MODE
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:TIME:MODE
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:TIME
/// HIFN     [:SOURce<hw>]:BB:ARBitrary:MCARrier:POWer:REFerence
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR mode/Sets the carrier frequency mode.
/// HIPAR numberOfCarriers/Sets the number of carriers in the ARB multi carrier waveform.
/// HIPAR carrierSpacing/Sets the frequency spacing between adjacent carriers of the multi 
/// HIPAR carrierSpacing/carrier waveform.
/// HIPAR crestFactorMode/Sets the mode for optimizing the crest factor by calculating the 
/// HIPAR crestFactorMode/carrier phases.
/// HIPAR signalPeriodMode/Selects the mode for calculating the resulting signal period of the 
/// HIPAR signalPeriodMode/multi carrier waveform. The resulting period is always calculated for 
/// HIPAR signalPeriodMode/all carriers in the carrier table irrespective of their state (ON/OFF).
/// HIPAR signalPeriod/The command sets the user-defined signal period. This setting is only 
/// HIPAR signalPeriod/possible for Signal Period Mode User (BB:ARB:MCAR:TIME:MODE USER).
/// HIPAR powerReference/Defines the way the individual carriers in a composed multi carrier 
/// HIPAR powerReference/signal are leveled.
ViStatus _VI_FUNC rssgt_ConfigureARBMultiCarrier(
  ViSession instrSession,
  ViInt32 mode,
  ViInt32 numberOfCarriers,
  ViReal64 carrierSpacing,
  ViInt32 crestFactorMode,
  ViInt32 signalPeriodMode,
  ViReal64 signalPeriod,
  ViInt32 powerReference
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(mode, ARBMultiCarrierMode);
    LogAddInt32Arg(numberOfCarriers);
    LogAddReal64Arg(carrierSpacing);
    LogAddSelectArg(crestFactorMode, ARBMultiCarrierCrestFactorMode);
    LogAddSelectArg(signalPeriodMode, ARBSignalPeriodMode);
    LogAddReal64Arg(signalPeriod);
    LogAddSelectArg(powerReference, ARBMCarPowerRef);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_MODE,
    mode), 2, "Mode");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_COUNT,
    numberOfCarriers), 2, "Number Of Carriers");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_SPACING,
    carrierSpacing), 4, "Carrier Spacing");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_CREST_FACTOR_MODE,
    crestFactorMode), 5, "Crest Factor Mode");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_SIGNAL_PERIOD_MODE,
    signalPeriodMode), 6, "Signal Period Mode");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_SIGNAL_PERIOD,
    signalPeriod), 7, "Signal Period");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_POWER_REFERENCE,
    powerReference), 8, "Power Reference");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the Multi Carrier Clipping.
/// HIFN     
/// HIFN     Note(s):
/// HIFN     Control Target Crest Factor and Filter Cut Off Frequency is valid only 
/// HIFN  for Clipping set to On.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_CLIPPING_ENABLED
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_CLIPPING_TARGET_CREST_FACTOR
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_CLIPPING_FILTER_CUT_OFF_FREQUENCY
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:MCARrier:CLIPping[:STATe]
/// HIFN     [SOURce<[1]|2>]:BB:ARB:MCARrier:CLIPping:CFACtor
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:MCARrier:CLIPping:CUToff
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR clipping/Switches baseband clipping on and off.
/// HIPAR targetCrestFactor/Sets the value of the desired crest factor, if baseband clipping is 
/// HIPAR targetCrestFactor/enabled (BB:ARB:MCAR:CLIP:STAT ON).
/// HIPAR filterCutOffFrequency/Sets the cut off frequency of the final lowpass filter, if baseband 
/// HIPAR filterCutOffFrequency/clipping is enabled (BB:ARB:MCAR:CLIP:STAT ON).
ViStatus _VI_FUNC rssgt_ConfigureARBMultiCarrierClipping(
  ViSession instrSession,
  ViBoolean clipping,
  ViReal64 targetCrestFactor,
  ViReal64 filterCutOffFrequency
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(clipping);
    LogAddReal64Arg(targetCrestFactor);
    LogAddReal64Arg(filterCutOffFrequency);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_CLIPPING_ENABLED,
    clipping), 2, "Clipping");

  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_CLIPPING_TARGET_CREST_FACTOR,
    targetCrestFactor), 3, "Target Crest Factor");

  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_CLIPPING_FILTER_CUT_OFF_FREQUENCY,
    filterCutOffFrequency), 4, "Filter Cut Off Frequency");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Sets the number of carriers in the ARB multi carrier waveform.
/// HIFN     
/// HIFN     Note(s):
/// HIFN     The total bandwidth (Number of carriers - 1) * Carrier spacing is 120 
/// HIFN  or 60 MHz depending on the installed options.
/// HIFN     
/// HIFN     The number of carriers entered therefore defines the maximum carrier 
/// HIFN  spacing (:BB:ARB:MCARrier:CARRier:SPACing).
/// HIFN     
/// HIFN     Example: ":BB:ARB:MCAR:CARR:COUN 10"
/// HIFN     'sets 10 carriers for the multi carrier waveform.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_COUNT
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CARRier:COUNt
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR numberOfCarriers/Sets the number of carriers in the ARB multi carrier waveform.
ViStatus _VI_FUNC rssgt_ARBMultiCarrierTableNumberOfCarriers(
  ViSession instrSession,
  ViInt32 numberOfCarriers
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(numberOfCarriers);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_COUNT,
    numberOfCarriers), 2, "Number Of Carriers");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Selects the input file with I/Q data to be modulated onto the carriers 
/// HIFN  in the defined carrier range.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_INPUT_IQ_FILE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARB:MCARrier:EDIT:CARRier:FILE
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR file/File Name parameter.
ViStatus _VI_FUNC rssgt_ARBMultiCarrierTableInputWaveformFile(
  ViSession instrSession,
  ViString file
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(file);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_INPUT_IQ_FILE,
    file), 2, "File");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Adopts the settings for the carrier range which has been defined using 
/// HIFN  the :BB:ARB:MCAR:EDIT:CARR:... commands.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_EXECUTE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:EDIT:CARRier:EXECute
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
ViStatus _VI_FUNC rssgt_ARBMultiCarrierTableApply(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_EXECUTE, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the Multi Carrier Table.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_START
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_STOP
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_ENABLED
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:EDIT:CARRier:STARt
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:EDIT:CARRier:STOP
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:EDIT:CARRier:STATe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR carrierStart/Selects the first carrier in the carrier range to which the settings 
/// HIPAR carrierStart/with the SOUR:BB:ARB:MCAR:EDIT:CARR:.. commands shall apply.
/// HIPAR carrierStop/Selects the last carrier in the carrier range to which the settings 
/// HIPAR carrierStop/with the :BB:ARB:MCAR:EDIT:CARR:.. commands shall apply.
/// HIPAR carrierState/Switches all the carriers in the selected carrier range on or off.
ViStatus _VI_FUNC rssgt_ConfigureARBMultiCarrierTableCarrier(
  ViSession instrSession,
  ViInt32 carrierStart,
  ViInt32 carrierStop,
  ViBoolean carrierState
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(carrierStart);
    LogAddInt32Arg(carrierStop);
    LogAddBoolArg(carrierState);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_START,
    carrierStart), 2, "Carrier Start");

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_STOP,
    carrierStop), 3, "Carrier Stop");

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_ENABLED,
    carrierState), 4, "Carrier State");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the Multi Carrier Table.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_POWER
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_POWER_STEP
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_PHASE
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_PHASE_STEP
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_DELAY
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_DELAY_STEP
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:EDIT:CARRier:POWer[:STARt]
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:EDIT:CARRier:POWer:STEP
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:EDIT:CARRier:PHASe[:STARt]
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:EDIT:CARRier:PHASe:STEP
/// HIFN     [SOURce]:BB:ARB:MCARrier:EDIT:CARRier:DELay[:STARt]
/// HIFN     [SOURce]:BB:ARB:MCARrier:EDIT:CARRier:DELay:STEP
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR gainStart/Sets the power for the individual carriers in the defined carrier 
/// HIPAR gainStart/range. If the command :BB:ARB:MCAR:EDIT:CARR:POW:STEP is used to define 
/// HIPAR gainStart/a step width, the power entered here applies only to the starting 
/// HIPAR gainStart/carrier. The power of the remaining carriers is stepped up or down by 
/// HIPAR gainStart/the power specified in the :BB:ARB:MCAR:EDIT:CARR:POW:STEP command.
/// HIPAR gainStep/Sets the step width by which the starting power of the carriers in the 
/// HIPAR gainStep/defined carrier range will be incremented.
/// HIPAR phaseStart/Sets the start phase for the individual carriers in the defined 
/// HIPAR phaseStart/carrier range. If the command :BB:ARB:MCAR:EDIT:CARR:PHAS:STEP is used 
/// HIPAR phaseStart/to define a step width, the phase entered here applies only to the 
/// HIPAR phaseStart/starting carrier. The phases of the remaining carriers are stepped up 
/// HIPAR phaseStart/or down by the phase value specified in the 
/// HIPAR phaseStart/:BB:ARB:MCAR:EDIT:CARR:PHAS:STEP command. The phase settings are only 
/// HIPAR phaseStart/valid if optimization of the crest factor is disabled 
/// HIPAR phaseStart/(:SOURce:BB:ARB:MCARrier:CFACtor:MODE OFF).
/// HIPAR phaseStep/Sets the step width by which the start phases of the carriers in the 
/// HIPAR phaseStep/defined carrier range will be incremented.
/// HIPAR delayStart/Sets the start delay for the individual carriers in the defined 
/// HIPAR delayStart/carrier range. If the command :BB:ARB:MCAR:EDIT:CARR:DEL:STEP is used 
/// HIPAR delayStart/to define a step width, the delay entered here applies only to the 
/// HIPAR delayStart/starting carrier. The delays of the remaining carriers are stepped up 
/// HIPAR delayStart/or down by the delay value specified in the 
/// HIPAR delayStart/:BB:ARB:MCAR:EDIT:CARR:DEL:STEP command.
/// HIPAR delayStep/Sets the step width by which the start delays of the carriers in the 
/// HIPAR delayStep/defined carrier range will be incremented.
ViStatus _VI_FUNC rssgt_ConfigureARBMultiCarrierTable(
  ViSession instrSession,
  ViReal64 gainStart,
  ViReal64 gainStep,
  ViReal64 phaseStart,
  ViReal64 phaseStep,
  ViReal64 delayStart,
  ViReal64 delayStep
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(gainStart);
    LogAddReal64Arg(gainStep);
    LogAddReal64Arg(phaseStart);
    LogAddReal64Arg(phaseStep);
    LogAddReal64Arg(delayStart);
    LogAddReal64Arg(delayStep);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_POWER,
    gainStart), 2, "Gain Start");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_POWER_STEP,
    gainStep), 3, "Gain Step");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_PHASE,
    phaseStart), 4, "Phase Start");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_PHASE_STEP,
    phaseStep), 5, "Phase Step");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_DELAY,
    delayStart), 6, "Delay Start");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_EDIT_DELAY_STEP,
    delayStep), 7, "Delay Step");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the selected Multi Carrier.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_ENABLED
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_FREQUENCY
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_POWER
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_PHASE
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_DELAY
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_DATA_FILE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CARRier<1...32>:STATe
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CARRier<1...32>:FREQuency
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CARRier<1...32>:POWer
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CARRier<1...32>:PHASe
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CARRier<1...32>:DELay
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:CARRier<1...32>:FILE
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR carrier/Selects the carrier.
/// HIPAR state/Switches the selected carrier on or off.
/// HIPAR frequency/Sets the carrier frequency, depending on the selected carrier 
/// HIPAR frequency/frequency mode.
/// HIPAR gain/Sets the gain of the selected carrier.
/// HIPAR phase/Sets the start phase of the selected carrier. The phase settings are 
/// HIPAR phase/only valid if optimization of the crest factor is disabled.
/// HIPAR delay/Sets the start delay of the selected carrier.
/// HIPAR file/Selects the file with I/Q data to be modulated onto the selected 
/// HIPAR file/carrier.
ViStatus _VI_FUNC rssgt_ConfigureARBMultiCarr(
  ViSession instrSession,
  ViInt32 carrier,
  ViBoolean state,
  ViReal64 frequency,
  ViReal64 gain,
  ViReal64 phase,
  ViReal64 delay,
  ViString file
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(carrier);
    LogAddBoolArg(state);
    LogAddReal64Arg(frequency);
    LogAddReal64Arg(gain);
    LogAddReal64Arg(phase);
    LogAddReal64Arg(delay);
    LogAddStringArg(file);
  }
  ViStatus error = VI_SUCCESS;
  ViChar rep_cap_carrier[15] = "";
  ViInt32   mode;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(carrier, 1, 32), 2, "Carrier");

  (void)snprintf(rep_cap_carrier, 15, "Carrier" _PERCD, carrier);

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, rep_cap_carrier,
    RSSGT_ATTR_ARB_MULTI_CARRIER_ENABLED,
    state), 3, "State");

  rssgt_GetAttributeViInt32(instrSession, "", RSSGT_ATTR_ARB_MULTI_CARRIER_MODE, &mode);

  // do not send a carrier frequency in equidistant mode!
  if (mode != RSSGT_VAL_ARB_MULTIC_MODE_EQU)
  {
    viCheckParm(rssgt_SetAttributeViReal64(instrSession, rep_cap_carrier,
      RSSGT_ATTR_ARB_MULTI_CARRIER_FREQUENCY,
      frequency), 4, "Frequency");
  }


  viCheckParm(rssgt_SetAttributeViReal64(instrSession, rep_cap_carrier,
    RSSGT_ATTR_ARB_MULTI_CARRIER_POWER,
    gain), 5, "Gain");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, rep_cap_carrier,
    RSSGT_ATTR_ARB_MULTI_CARRIER_PHASE,
    phase), 6, "Phase");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, rep_cap_carrier,
    RSSGT_ATTR_ARB_MULTI_CARRIER_DELAY,
    delay), 7, "Delay");
  viCheckParm(rssgt_SetAttributeViString(instrSession, rep_cap_carrier,
    RSSGT_ATTR_ARB_MULTI_CARRIER_DATA_FILE,
    file), 8, "File");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Determines whether the instrument performs an absolute or a 
/// HIFN  differential storing of the settings.
/// HIFN     
/// HIFN     Enable this function to accelerate the saving process by saving only 
/// HIFN  the settings with values different to the default ones.
/// HIFN     
/// HIFN     Note: This function is not affected by the "Preset" function.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_FILE_STORE_FAST
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce<hw>]:BB:ARBitrary:MCARrier:SETTing:STORe:FAST
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR storeFast/Determines whether the instrument performs an absolute or a 
/// HIPAR storeFast/differential storing of the settings.
ViStatus _VI_FUNC rssgt_ConfigureARBMultiCarrierStoreFast(
  ViSession instrSession,
  ViBoolean storeFast
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(storeFast);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_FILE_STORE_FAST,
    storeFast), 2, "Store Fast");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function queries the available settings files in the specified 
/// HIFN  default directory. The settings files are used to set the ARB multi 
/// HIFN  carrier submenu. Only files with the file extension *.arb_multcarr will 
/// HIFN  be listed.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_FILE_CATALOG
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:SETTing:CATalog?
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR arraySize/Array size parameter.
/// HIPAR catalog/Queries the available settings files in the specified default 
/// HIPAR catalog/directory. The settings files are used to set the ARB multi carrier 
/// HIPAR catalog/submenu. Only files with the file extension *.arb_multcarr will be 
/// HIPAR catalog/listed.
ViStatus _VI_FUNC rssgt_ARBMultiCarrierFileCatalog(
  ViSession instrSession,
  ViInt32 arraySize,
  ViChar _VI_FAR catalog[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(arraySize);
    LogAddCharArrOut(catalog, &arraySize);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_GetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_FILE_CATALOG, arraySize,
    catalog), 2, "Catalog");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Loads the settings file in the default directory. A path can also be 
/// HIFN  specified, in which case the files in the specified directory are 
/// HIFN  selected. If a settings file with the specified name does not yet 
/// HIFN  exist, it is created. The file extension may be omitted. Only files 
/// HIFN  with the file extension *.arb_multcarr will be loaded or created.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_FILE_LOAD
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:MCARrier:SETTing:LOAD
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR file/File Name parameter.
ViStatus _VI_FUNC rssgt_ARBMultiCarrierFileLoad(
  ViSession instrSession,
  ViString file
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(file);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_FILE_LOAD,
    file), 2, "File");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Stores the current settings of submenu ARB Multi Carrier in a file in 
/// HIFN  the default directory. A path can also be specified, in which case the 
/// HIFN  files are stored in the specified directory. The file extension may be 
/// HIFN  omitted, the files are stored with the file extension *.arb_multcarr.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MULTI_CARRIER_FILE_STORE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARB:MCARrier:SETTing:STORe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR file/File Name parameter.
ViStatus _VI_FUNC rssgt_ARBMultiCarrierFileStore(
  ViSession instrSession,
  ViString file
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(file);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_MULTI_CARRIER_FILE_STORE,
    file), 2, "File");
Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the sine test signal.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_SINE_FREQUENCY
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_SINE_SAMPLES
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_SINE_PHASE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:TSIGnal:SINE:FREQuency
/// HIFN     [SOURce]:BB:ARBitrary:TSIGnal:SINE:SAMPles
/// HIFN     [SOURce]:BB:ARBitrary:TSIGnal:SINE:PHASe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR frequency/Sets the frequency of the simple sinusoidal test signal. This signal 
/// HIPAR frequency/is used as output via the I channel. A sine wave of the same frequency 
/// HIPAR frequency/but optionally phase-shifted is generated on the Q path 
/// HIPAR frequency/(ARB:TSIGnal:SINE:POFFset).
/// HIPAR samples/Sets the sample rate for the sine signal in samples per period. The 
/// HIPAR samples/resulting clock rate must not exceed the maximum ARB clock rate of 150 
/// HIPAR samples/or 90 MHz depending on the installed options. The maximum value is 
/// HIPAR samples/automatically restricted by reference to the set frequency.
/// HIPAR phaseOffset/Sets the phase offset of the sine wave on the Q channel relative to 
/// HIPAR phaseOffset/the sine wave on the I channel.
ViStatus _VI_FUNC rssgt_ConfigureARBSineTestSignal(
  ViSession instrSession,
  ViReal64 frequency,
  ViInt32 samples,
  ViReal64 phaseOffset
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(frequency);
    LogAddInt32Arg(samples);
    LogAddReal64Arg(phaseOffset);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_SINE_FREQUENCY,
    frequency), 2, "Frequency");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_SINE_SAMPLES,
    samples), 3, "Samples");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_SINE_PHASE,
    phaseOffset), 4, "Phase Offset");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Selects the test signal waveform file name.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_SINE_NAME
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]:BB:ARBitrary:TSIGnal:SINE:CREate:NAMed
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR fileName/File Name parameter.
ViStatus _VI_FUNC rssgt_ConfigureARBSineTestSignalName(
  ViSession instrSession,
  ViString fileName
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(fileName);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_SINE_NAME,
    fileName), 2, "File Name");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Create a sine test signal waveform.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_SINE_CREATE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:TSIGnal:SINE:CREate
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
ViStatus _VI_FUNC rssgt_ARBSineTestSignalGenerate(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_SINE_CREATE, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the rectangle test signal.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_FREQUENCY
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_SAMPLES
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_AMPLITUDE
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_OFFSET_DC
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:TSIGnal:SINE:FREQuency
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:TSIGnal:RECTangle:SAMPles
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:TSIGnal:RECTangle:AMPLitude
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:TSIGnal:RECTangle:OFFSet
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR frequency/Sets the frequency of the test signal.
/// HIPAR samples/Sets the number of sample values required for the rectangular signal 
/// HIPAR samples/per period.
/// HIPAR amplitude/Sets the digital amplitude of the rectangular wave. 
/// HIPAR offsetDC/Sets the DC component.
ViStatus _VI_FUNC rssgt_ConfigureARBRectTestSignal(
  ViSession instrSession,
  ViReal64 frequency,
  ViInt32 samples,
  ViReal64 amplitude,
  ViReal64 offsetDC
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(frequency);
    LogAddInt32Arg(samples);
    LogAddReal64Arg(amplitude);
    LogAddReal64Arg(offsetDC);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_FREQUENCY,
    frequency), 2, "Frequency");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_SAMPLES,
    samples), 3, "Samples");

  // clear interdependent parameters amplitude and offset before setting new values 
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_AMPLITUDE,
    0), 4, "Amplitude");


  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_OFFSET_DC,
    0), 5, "Offset DC");


  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_AMPLITUDE,
    amplitude), 4, "Amplitude");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_OFFSET_DC,
    offsetDC), 5, "Offset DC");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Selects the test signal waveform file name.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_NAME
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:TSIGnal:RECTangle:CREate:NAMed
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR fileName/File Name parameter.
ViStatus _VI_FUNC rssgt_ConfigureARBRectTestSignalName(
  ViSession instrSession,
  ViString fileName
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(fileName);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_NAME,
    fileName), 3, "File Name");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Generates a signal and uses it as output straight away.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_CREATE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:TSIGnal:RECTangle:CREate
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
ViStatus _VI_FUNC rssgt_ARBRectTestSignalGenerate(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_RECTANGLE_CREATE, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the Constant IQ test signal.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_CIQ_I
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_CIQ_Q
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:TSIGnal:CIQ:I
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:TSIGnal:CIQ:Q
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR iValue/Sets the value for the I component of the test signal. 
/// HIPAR qValue/Sets the value for the Q component of the test signal. 
ViStatus _VI_FUNC rssgt_ConfigureARBCIQTestSignal(
  ViSession instrSession,
  ViReal64 iValue,
  ViReal64 qValue
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(iValue);
    LogAddReal64Arg(qValue);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_CIQ_I,
    iValue), 2, "I Value");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_CIQ_Q,
    qValue), 3, "Q Value");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Generates a signal and saves it to a waveform file.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_CIQ_NAME
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>]:BB:ARB:TSIGnal:CIQ:CREate:NAMed
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR fileName/File Name parameter.
ViStatus _VI_FUNC rssgt_ConfigureARBCIQTestSignalName(
  ViSession instrSession,
  ViString fileName
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(fileName);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_CIQ_NAME,
    fileName), 2, "File Name");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Create a sine test signal waveform.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TSIGNAL_CIQ_CREATE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>]:BB:ARBitrary:TSIGnal:CIQ:CREate
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
ViStatus _VI_FUNC rssgt_ARBCIQTestSignalGenerate(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_TSIGNAL_CIQ_CREATE, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Executes a trigger. The internal trigger source must be selected using 
/// HIFN  the command ARB:TRIGger:SOURce INTernal and a trigger mode other than 
/// HIFN  AUTO must be selected using the command :ARB:SEQuence.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_EXECUTE_TRIGGER
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:TRIGger:EXECute
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
ViStatus _VI_FUNC rssgt_ARBExecuteTrigger(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_EXECUTE_TRIGGER, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Queries the status of waveform output or all trigger modes with ARB 
/// HIFN  on.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TRIGGER_STATUS
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:BB:ARBitrary:TRIGger:RMODe?
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR status/Queries the status of waveform output or all trigger modes with ARB 
/// HIPAR status/on.
/// HIPAR arraySize/Length of the buffer.
ViStatus _VI_FUNC rssgt_ARBTriggerStatus(
  ViSession instrSession,
  ViChar _VI_FAR status[],
  ViInt32     arraySize
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddCharArrOut(status, &instrSession);
    LogAddInt32Arg(arraySize);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_GetAttributeViString(instrSession, "",
    RSSGT_ATTR_ARB_TRIGGER_STATUS, arraySize,
    status), 2, "Status");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function selects the trigger mode.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TRIGGER_MODE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:SEQuence
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR mode/Selects ARB trigger mode.
ViStatus _VI_FUNC rssgt_ConfigureARBTriggerMode(
  ViSession instrSession,
  ViInt32 mode
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(mode, TrgTriggerMode);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_TRIGGER_MODE,
    mode), 2, "Mode");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function selects the trigger source for legacy instruments.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TRIGGER_SOURCE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:TRIGger:SOURce
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR source/Sets the trigger source.
ViStatus _VI_FUNC rssgt_ConfigureARBTriggerSource(
  ViSession instrSession,
  ViInt32 source
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(source, TriggerSour);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_TRIGGER_SOURCE,
    source), 2, "Source");
Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the trigger single.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_TRIGGER_SINGLE_SIGNAL_DURATION_UNIT
/// HIFN     RSSGT_ATTR_ARB_TRIGGER_SINGLE_SEQUENCE_LENGTH
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARB:TRIGger:SLUNit
/// HIFN     [SOURce<[1]|2>:]BB:ARB:TRIGger:SLENgth
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR signalDurationUnit/Defines the unit for the entry of the length of the signal sequence to 
/// HIPAR signalDurationUnit/be output in the Single trigger mode.
/// HIPAR signalDuration/Defines the length of the signal sequence to be output in the Single 
/// HIPAR signalDuration/trigger mode. It is possible to output deliberately just part of the 
/// HIPAR signalDuration/waveform, an exact sequence of the waveform, or a defined number of 
/// HIPAR signalDuration/repetitions of the waveform.
ViStatus _VI_FUNC rssgt_ConfigureARBTriggerSingle(
  ViSession instrSession,
  ViInt32 signalDurationUnit,
  ViInt32 signalDuration
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(signalDurationUnit, ARBSignalDurationUnit);
    LogAddInt32Arg(signalDuration);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_TRIGGER_SINGLE_SIGNAL_DURATION_UNIT,
    signalDurationUnit), 2, "Signal Duration Unit");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_TRIGGER_SINGLE_SEQUENCE_LENGTH,
    signalDuration), 3, "Signal Duration");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function defines the signal for the selected marker output.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MARKER_MODE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:TRIGger:OUTPut<[1]...4>:MODE
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR marker/Sets the marker number.
/// HIPAR mode/Defines the signal for the selected marker output.
ViStatus _VI_FUNC rssgt_ConfigureARBMarkerMode(
  ViSession instrSession,
  ViInt32 marker,
  ViInt32 mode
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(marker);
    LogAddSelectArg(mode, ARBMarkerMode);
  }
  ViStatus error = VI_SUCCESS;
  ViChar rep_cap_marker[15] = "";

  checkErr(Rs_LockSession(instrSession, NULL));
  viCheckParm(rssgt_invalidViInt32Range(marker, 1, 2), 2, "Marker");

  (void)snprintf(rep_cap_marker, 15, "M" _PERCD, marker);

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, rep_cap_marker,
    RSSGT_ATTR_ARB_MARKER_MODE,
    mode), 3, "Mode");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function sets the divider for Pulse marker mode 
/// HIFN  (SOUR:BB:ARB:TRIG:OUTP:MODE PULSe). The resulting pulse frequency is 
/// HIFN  derived by dividing the symbol rate by the divider.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_MARKER_PULSE_DIVIDER
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:TRIGger:OUTPut<[1]...4>:PULSe:DIVider
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR marker/Sets the marker number.
/// HIPAR divider/Sets the divider for Pulse marker mode (SOUR:BB:ARB:TRIG:OUTP:MODE 
/// HIPAR divider/PULSe). The resulting pulse frequency is derived by dividing the symbol 
/// HIPAR divider/rate by the divider.
ViStatus _VI_FUNC rssgt_ConfigureARBMarkerDivider(
  ViSession instrSession,
  ViInt32 marker,
  ViInt32 divider
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(marker);
    LogAddInt32Arg(divider);
  }
  ViStatus error = VI_SUCCESS;
  ViChar rep_cap_marker[15] = "";

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(marker, 1, 2), 2, "Marker");

  (void)snprintf(rep_cap_marker, 15, "M" _PERCD, marker);

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, rep_cap_marker,
    RSSGT_ATTR_ARB_MARKER_PULSE_DIVIDER,
    divider), 3, "Divider");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the marker delay.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR marker/Sets the marker number.
/// HIPAR delay/Defines the delay between the signal on the marker outputs and the 
/// HIPAR delay/start of the frame/slot, expressed in terms of symbols.
/// HIPAR fixMarkerDelay/Unused
ViStatus _VI_FUNC rssgt_ConfigureARBMarkerDelay(
  ViSession instrSession,
  ViInt32 marker,
  ViInt32 delay,
  ViBoolean fixMarkerDelay
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(marker);
    LogAddInt32Arg(delay);
    LogAddBoolArg(fixMarkerDelay);
  }
  ViStatus error = VI_SUCCESS;
  ViChar rep_cap_marker[15] = "";

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(marker, 1, 2), 2, "Marker");

  (void)snprintf(rep_cap_marker, 15, "M" _PERCD, marker);

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, rep_cap_marker,
    RSSGT_ATTR_ARB_MARKER_DELAY,
    delay), 3, "Delay");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the ARB the clock source.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ARB_CLOCK_SOURCE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]BB:ARBitrary:CLOCk:SOURce
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR path/The path for which is this function performed.
/// HIPAR source/Sets the clock source. 
ViStatus _VI_FUNC rssgt_ConfigureARBClockSource(
  ViSession instrSession,
  ViInt32 source
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(source, ClockSource);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ARB_CLOCK_SOURCE,
    source), 2, "Source");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function activates or deactivates the white noise (AWGN = 
/// HIFN  Averaged White Gaussian Noise). The noise signal is either superimposed 
/// HIFN  on the baseband signal (Additive Noise mode (:SOUR:AWGN:MODE ADD)) or 
/// HIFN  is output as a pure noise signal (Noise Only mode (:SOUR:AWGN:MODE 
/// HIFN  ONLY)).
/// HIFN     
/// HIFN     Required Option(s):
/// HIFN     K62
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_AWGN_ENABLED
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<1|2>:]AWGN:STATe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR awgnEnabled/Activates or deactivates the white noise (AWGN = Averaged White 
/// HIPAR awgnEnabled/Gaussian Noise). The noise signal is either superimposed on the 
/// HIPAR awgnEnabled/baseband signal (Additive Noise mode (:SOUR:AWGN:MODE ADD)) or is 
/// HIPAR awgnEnabled/output as a pure noise signal (Noise Only mode (:SOUR:AWGN:MODE ONLY)).
ViStatus _VI_FUNC rssgt_ConfigureAWGNEnabled(
  ViSession instrSession,
  ViBoolean awgnEnabled
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(awgnEnabled);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_AWGN_ENABLED,
    awgnEnabled), 2, "AWGN Enabled");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function selects the mode for generating the noise signal.
/// HIFN     
/// HIFN     Required Option(s):
/// HIFN     K62
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_AWGN_MODE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<1|2>:]AWGN:MODE
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR mode/Selects the mode for generating the noise signal.
ViStatus _VI_FUNC rssgt_ConfigureAWGNMode(
  ViSession instrSession,
  ViInt32 mode
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(mode, AWGNMode);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_AWGN_MODE,
    mode), 2, "Mode");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the AWGN Noise Only.
/// HIFN     
/// HIFN     Required Option(s):
/// HIFN     K62
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_AWGN_BANDWIDTH
/// HIFN     RSSGT_ATTR_AWGN_BANDWIDTH_RATIO
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]AWGN:BWIDth
/// HIFN     [SOURce<[1]|2>:]AWGN:BWIDth:RATio
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR systemBandwidth/Sets the system bandwidth. The noise signal at the level which 
/// HIPAR systemBandwidth/corresponds to the specified carrier/noise ratio is generated in the 
/// HIPAR systemBandwidth/bandwidth specified here.
/// HIPAR bandwidthRatio/Sets the ratio of minimum noise bandwidth to system bandwidth.
ViStatus _VI_FUNC rssgt_ConfigureAWGNNoiseOnly(
  ViSession instrSession,
  ViReal64 systemBandwidth,
  ViReal64 bandwidthRatio
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(systemBandwidth);
    LogAddReal64Arg(bandwidthRatio);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_AWGN_BANDWIDTH,
    systemBandwidth), 2, "System Bandwidth");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_AWGN_BANDWIDTH_RATIO,
    bandwidthRatio), 3, "Bandwidth Ratio");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function selects the display mode to Bseband or RF.
/// HIFN     
/// HIFN     Required Option(s):
/// HIFN     K62
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_AWGN_DISPLAY_MODE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce<hw>]:AWGN:DISP:MODE
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR displayMode/Selects the display mode to Bseband or RF.
ViStatus _VI_FUNC rssgt_ConfigureAWGNAdditiveNoiseDisplayMode(
  ViSession instrSession,
  ViInt32 displayMode
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(displayMode, AWGNDisplayMode);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_AWGN_DISPLAY_MODE,
    displayMode), 2, "Display Mode");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures frequency and phase.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_RF_FREQUENCY
/// HIFN     RSSGT_ATTR_RF_DELTA_PHASE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<1|2>:]FREQuency[:CW|:FIXed]
/// HIFN     [SOURce<1|2>:]PHASe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR frequency/Sets the frequency of the RF output signal for CW mode 
/// HIPAR frequency/(:SOURce:FREQuency:MODE CW).. 
/// HIPAR deltaPhase/Specifies the phase variation relative to the current phase. The 
/// HIPAR deltaPhase/variation can be specified in RADians.
ViStatus _VI_FUNC rssgt_ConfigureFrequencyPhase(
  ViSession instrSession,
  ViReal64 frequency,
  ViReal64 deltaPhase
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(frequency);
    LogAddReal64Arg(deltaPhase);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_RF_FREQUENCY,
    frequency), 2, "Frequency");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_RF_DELTA_PHASE,
    deltaPhase), 3, "Delta Phase");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Adopts the phase set with SOURce:PHASe:ADJust as the current phase. 
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_RESET_RF_DELTA_PHASE_DISPLAY
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<1|2>:]PHASe:REFerence
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rssgt_ResetDeltaPhaseDisplay(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_RESET_RF_DELTA_PHASE_DISPLAY, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function selects the source of the local oscillator signal.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_LOSC_SOURCE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce]:LOSCillator:SOURce
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR source/Selects the source of the local oscillator signal.
ViStatus _VI_FUNC rssgt_ConfigureLocalOscillatorSource(
  ViSession instrSession,
  ViInt32 source
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(source, ReferenceOscillatorSource);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_LOSC_SOURCE,
    source), 2, "Source");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function selects between internal and external reference 
/// HIFN  frequency.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_REFERENCE_OSCILLATOR_SOURCE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce]:ROSCillator:SOURce
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR source/Selects the mode for generating the noise signal.
ViStatus _VI_FUNC rssgt_ConfigureReferenceOscillatorSource(
  ViSession instrSession,
  ViInt32 source
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(source, ReferenceOscillatorSource);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_REFERENCE_OSCILLATOR_SOURCE,
    source), 2, "Source");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the Reference Oscillator external.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_REFERENCE_OSCILLATOR_OUTPUT_FREQUENCY
/// HIFN     RSSGT_ATTR_REFERENCE_OSCILLATOR_SYNCHRONIZATION_BANDWIDTH
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce]:ROSCillator:OUTPut:FREQuency
/// HIFN     [SOURce:]ROSCillator:EXTernal:SBANdwidth
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR referenceInputFrequency/Selects the output for the reference oscillator signal.
/// HIPAR synchronisationBandwidth/Valid Values:
ViStatus _VI_FUNC rssgt_ConfigureReferenceOscillatorExternal(
  ViSession instrSession,
  ViInt32 referenceInputFrequency,
  ViInt32 synchronisationBandwidth
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(referenceInputFrequency, ROSCOutputFrequency);
    LogAddSelectArg(synchronisationBandwidth, ROSCSynchronizationBandwidth);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_REFERENCE_OSCILLATOR_OUTPUT_FREQUENCY,
    referenceInputFrequency), 2, "Reference Input Frequency");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_REFERENCE_OSCILLATOR_SYNCHRONIZATION_BANDWIDTH,
    synchronisationBandwidth), 3, "Synchronisation Bandwidth");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the Reference Oscillator Adjustment.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_REFERENCE_OSCILLATOR_ADJUSTMENT_ACTIVE
/// HIFN     RSSGT_ATTR_REFERENCE_OSCILLATOR_ADJUSTMENT_FREQUENCY
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce:]ROSCillator[:INTernal]:ADJust[:STATe]
/// HIFN     [SOURce:]ROSCillator[:INTernal]:ADJust:VALue
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR active/Determines whether the calibrated (OFF) or a user-defined (ON) 
/// HIPAR active/adjustment value is used for fine adjustment of the frequency. With 
/// HIPAR active/STATe ON, the instrument is no longer in the calibrated state.
/// HIPAR smallShift/Specifies the frequency correction value (adjustment value). A 
/// HIPAR smallShift/detailed definition is given in  section  RF Reference Frequency - 
/// HIPAR smallShift/Reference Oscillator.
ViStatus _VI_FUNC rssgt_ConfigureReferenceOscillatorAdjustment(
  ViSession instrSession,
  ViBoolean active,
  ViInt32 smallShift
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(active);
    LogAddInt32Arg(smallShift);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_REFERENCE_OSCILLATOR_ADJUSTMENT_ACTIVE,
    active), 2, "Active");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_REFERENCE_OSCILLATOR_ADJUSTMENT_SMALL_SHIFT,
    smallShift), 3, "Small Shift");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function selects the state which the RF output assumes when the 
/// HIFN  instrument is switched on.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_OUTPUT_ENABLED
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     OUTPut[:STATe]
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR state/Selects the state which the RF output assumes when the instrument is 
/// HIPAR state/switched on.
ViStatus _VI_FUNC rssgt_ConfigureLevelState(
  ViSession instrSession,
  ViBoolean state
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(state);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_OUTPUT_ENABLED,
    state), 2, "State");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the Level.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_RF_LEVEL_MODE
/// HIFN     RSSGT_ATTR_RF_LEVEL_SETTING_CHARACTERISTIC
/// HIFN     RSSGT_ATTR_RF_AMPLITUDE
/// HIFN     RSSGT_ATTR_RF_LEVEL_LIMIT
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce<HW>]:POWer:LMODe
/// HIFN     [:SOURce]:POWer:SCHaracteristic
/// HIFN     [SOURce<1|2>:]POWer[:LEVel][:IMMediate][:AMPLitude]
/// HIFN     [SOURce<1|2>:]POWer:LIMit[:AMPLitude]
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR mode/Sets the RF level mode.
/// HIPAR settingCharacteristic/Selects the characteristic for the level setting.
/// HIPAR level/Sets the RF output level in CW mode.
/// HIPAR limit/Limits the maximum RF output level in CW and SWEEP mode. It does not 
/// HIPAR limit/influence the LEVEL display or the response to the POW? query command.
ViStatus _VI_FUNC rssgt_ConfigureLevelSettings(
  ViSession instrSession,
  ViInt32 mode,
  ViInt32 settingCharacteristic,
  ViReal64 level,
  ViReal64 limit
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(mode, RFLevelMode);
    LogAddSelectArg(settingCharacteristic, RFLevSettCharacteristic);
    LogAddReal64Arg(level);
    LogAddReal64Arg(limit);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_RF_LEVEL_MODE,
    mode), 2, "Mode");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_RF_LEVEL_SETTING_CHARACTERISTIC,
    settingCharacteristic), 3, "Setting Characteristic");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_RF_AMPLITUDE,
    level), 4, "Level");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_RF_LEVEL_LIMIT,
    limit), 5, "Limit");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function queries the Level limits.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_RF_LEVEL_RANGE_MINIMUM
/// HIFN     RSSGT_ATTR_RF_LEVEL_RANGE_MAXIMUM
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce]:POWer:RANGe:LOWer?
/// HIFN     [SOURce]:POWer:RANGe:UPPer?
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR minimum/Queries the minimum level range in the current level mode.
/// HIPAR maximum/Queries the maximum level range in the current level mode.
ViStatus _VI_FUNC rssgt_QueryLevelRange(
  ViSession instrSession,
  ViReal64* minimum,
  ViReal64* maximum
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Out(minimum);
    LogAddReal64Out(maximum);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_GetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_RF_LEVEL_RANGE_MINIMUM,
    minimum), 2, "Minimum");
  viCheckParm(rssgt_GetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_RF_LEVEL_RANGE_MAXIMUM,
    maximum), 3, "Maximum");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function selects the state which the RF output assumes when the 
/// HIFN  instrument is switched on.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_POWER_ON_ENABLED
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     OUTPut<1|2>[:STATe]:PON
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR state/Selects the state which the RF output assumes when the instrument is 
/// HIPAR state/switched on.
ViStatus _VI_FUNC rssgt_ConfigureLevelPowerOn(
  ViSession instrSession,
  ViInt32 state
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(state, PowerOnState);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_POWER_ON_ENABLED,
    state), 2, "State");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the Automatic Level Control.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_ALC_ENABLED
/// HIFN     RSSGT_ATTR_ALC_DETECTOR_SENSITIVITY
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<1|2>:]POWer:ALC[:STATe]
/// HIFN     [:SOURce<HW>]:POWer:ALC:DSENsitivity
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR state/Activates/deactivates automatic level control. 
/// HIPAR detectorSensitivity/Sets the sensitivity of the ALC detector.
ViStatus _VI_FUNC rssgt_ConfigureLevelALC(
  ViSession instrSession,
  ViInt32 state,
  ViInt32 detectorSensitivity
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(state, ALCEnabled);
    LogAddSelectArg(detectorSensitivity, ALCDetectorSensitivity);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ALC_ENABLED,
    state), 2, "State");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ALC_DETECTOR_SENSITIVITY,
    detectorSensitivity), 3, "Detector Sensitivity");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function queries the Level Attenuator limits.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_RF_ATTENUATOR_LOWER_RANGE
/// HIFN     RSSGT_ATTR_RF_ATTENUATOR_UPPER_RANGE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     OUTPut<1|2>:AFIXed:RANGe:LOWer?
/// HIFN     OUTPut<1|2>:AFIXed:RANGe:UPPer?
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR minimum/Queries the minimum level which can be set without the attenuator 
/// HIPAR minimum/being adjusted (Attenuator FIXed).
/// HIPAR maximum/Queries the maximum level which can be set without the attenuator 
/// HIPAR maximum/being adjusted (Attenuator FIXed).
ViStatus _VI_FUNC rssgt_QueryLevelAttenuatorRange(
  ViSession instrSession,
  ViReal64* minimum,
  ViReal64* maximum
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Out(minimum);
    LogAddReal64Out(maximum);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_GetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_RF_ATTENUATOR_LOWER_RANGE,
    minimum), 2, "Minimum");
  viCheckParm(rssgt_GetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_RF_ATTENUATOR_UPPER_RANGE,
    maximum), 3, "Maximum");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the Attenuator Settings.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_RF_ATTENUATOR_MODE
/// HIFN     RSSGT_ATTR_RF_ATTENUATOR_SWITCH_OVER_OFFSET
/// HIFN     RSSGT_ATTR_RF_ATTENUATOR_RF_OFF_MODE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     OUTPut<1|2>:AMODe
/// HIFN     [:SOURce]:POWer:ATTenuation:SOVer[:OFFSet]
/// HIFN     [:SOURce<HW>]:POWer:ATTenuation:RFOFf:MODE
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR mode/Switches the mode of the attenuator at the RF output (Attenuator 
/// HIPAR mode/MODe).
/// HIPAR satt/Sets the switch-over offset value of the attenuator.
/// HIPAR rfOffMode/Selects the attenuator's state after the RF signal is switched off.
ViStatus _VI_FUNC rssgt_ConfigureLevelAttenuatorSettings(
  ViSession instrSession,
  ViInt32 mode,
  ViReal64 satt,
  ViInt32 rfOffMode
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(mode, RFAttenuatorMode);
    LogAddReal64Arg(satt);
    LogAddSelectArg(rfOffMode, RFAttRFOFFMode);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_RF_ATTENUATOR_MODE,
    mode), 2, "Mode");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_RF_ATTENUATOR_SWITCH_OVER_OFFSET,
    satt), 3, "Satt");
  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_RF_ATTENUATOR_RF_OFF_MODE,
    rfOffMode), 4, "RF Off Mode");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function activates and deactivates IQ modulation. 
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_IQ_ENABLED 
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     :SOURce<Path>:IQ:STATe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR enabled/Activates and deactivates IQ modulation. 
ViStatus _VI_FUNC rssgt_ConfigureIQEnabled(
  ViSession instrSession,
  ViBoolean enabled
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(enabled);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_IQ_ENABLED,
    enabled), 2, "Enabled");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures the IQ Impairments. 
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_IQ_IMPAIRMENTS_ENABLED
/// HIFN     RSSGT_ATTR_IQ_I_OFFSET
/// HIFN     RSSGT_ATTR_IQ_Q_OFFSET
/// HIFN     RSSGT_ATTR_IQ_GAIN_IMBALANCE
/// HIFN     RSSGT_ATTR_IQ_QUADRATURE_OFFSET 
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     :SOURce<Path>:IQ:IMPairment[:STATe]
/// HIFN     :SOURce<Path>:IQ:IMPairment:LEAKage:I
/// HIFN     :SOURce<Path>:IQ:IMPairment:LEAKage:Q
/// HIFN     :SOURce<Path>:IQ:IMPairment:IQRatio[:MAGNitude]
/// HIFN     :SOURce<Path>:IQ:IMPairment:QUADrature[:ANGLe]
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR enabled/Activates (ON) and deactivates (OFF) the three impairment or 
/// HIPAR enabled/correction values LEAKage, QUADrature and IQRatio for the analog signal 
/// HIPAR enabled/in the IQ modulator.
/// HIPAR iOffset/Sets the carrier leakage amplitude for the I-channel of the IQ 
/// HIPAR iOffset/modulation.
/// HIPAR qOffset/Sets the carrier leakage amplitude for the Q-channel of the IQ 
/// HIPAR qOffset/modulation.
/// HIPAR gainImbalance/Sets the ratio of I modulation to Q modulation (amplification 
/// HIPAR gainImbalance/"imbalance").
/// HIPAR quadratureOffset/Sets the quadrature offset for the IQ modulation.
ViStatus _VI_FUNC rssgt_ConfigureIQImpairments(
  ViSession instrSession,
  ViBoolean enabled,
  ViReal64 iOffset,
  ViReal64 qOffset,
  ViReal64 gainImbalance,
  ViReal64 quadratureOffset
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(enabled);
    LogAddReal64Arg(iOffset);
    LogAddReal64Arg(qOffset);
    LogAddReal64Arg(gainImbalance);
    LogAddReal64Arg(quadratureOffset);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_IQ_MODULATION_IMPAIRMENTS_ENABLED,
    enabled), 2, "Enabled");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_IQ_I_OFFSET,
    iOffset), 3, "I Offset");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_IQ_Q_OFFSET,
    qOffset), 4, "Q Offset");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_IQ_GAIN_IMBALANCE,
    gainImbalance), 5, "Gain Imbalance");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_IQ_QUADRATURE_OFFSET,
    quadratureOffset), 6, "Quadrature Offset");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function stores the correction values into a file with the 
/// HIFN  selected file name and loads it.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR filename/Sets the filename of the shaping table.
/// HIPAR arraySize/Sets the amount of elements to be set with the arrays Vcc/Vmax and 
/// HIPAR arraySize/Vin/Vmax.
/// HIPAR vin_vmax/Array of the Vin/Vmax values. Number of elements of this array should 
/// HIPAR vin_vmax/be the same as number of elements of Vcc/Vmax array.
/// HIPAR vcc_vmax/Array of the Vcc/Vmax values. Number of elements of this array should 
/// HIPAR vcc_vmax/be the same as number of elements of Vin/Vmax array.
ViStatus _VI_FUNC rssgt_CreateNewEnvelopeTrackingShapingTable(
  ViSession instrSession,
  ViString filename,
  ViInt32     arraySize,
  ViReal64 vin_vmax[],
  ViReal64 vcc_vmax[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(filename);
    LogAddInt32Arg(arraySize);
    LogAddReal64ArrArg(vin_vmax, &arraySize);
    LogAddReal64ArrArg(vcc_vmax, &arraySize);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar    command[50] = "";
  ViInt32   adaptation = 0;
  ViChar*   pbuffer = NULL;
  ViChar*   p2buf = NULL;
  ViInt32   i = 0;

  checkErr(Rs_LockSession(instrSession, NULL));

  if (arraySize <= 0)
    LogCloseCall(VI_SUCCESS);
    return VI_SUCCESS;

  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ENVELOPE_TRACKING_VOLTAGE_ADAPTATION,
    &adaptation));

  // If Adaptation is AUTO (0), use command with ...SHAPing:PV:FILE...
  snprintf(command, 50, "SOUR:IQ:OUTP:ANAL:ENV:SHAP:%sFILE:NEW", (adaptation == 0) ? "PV:" : "");

  p2buf = pbuffer = (ViChar*)malloc((size_t)arraySize * 50 + 200 + strlen(filename));

  if (pbuffer)
  {
    p2buf += snprintf(pbuffer, 300, "%s '%s',%.12G,%.12G", command, filename, vin_vmax[0], vcc_vmax[0]);

    for (i = 1; i < arraySize; i++)
      p2buf += snprintf(p2buf, 50, ",%.12G,%.12G", vin_vmax[i], vcc_vmax[i]);

    strcat(pbuffer, "\n");

    checkErr(rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)strlen(pbuffer), NULL));

    checkErr(rssgt_CheckStatus(instrSession));
  }

Error:
  if (pbuffer)
    free(pbuffer);
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function defines the shaping function in a raw data format.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR arraySize/Sets the amount of elements to be set with the arrays Vcc/Vmax and 
/// HIPAR arraySize/Vin/Vmax.
/// HIPAR vin_vmax/Array of the Vin/Vmax values. Number of elements of this array should 
/// HIPAR vin_vmax/be the same as number of elements of Vcc/Vmax array.
/// HIPAR vcc_vmax/Array of the Vcc/Vmax values. Number of elements of this array should 
/// HIPAR vcc_vmax/be the same as number of elements of Vin/Vmax array.
ViStatus _VI_FUNC rssgt_ConfigureEnvelopeTrackingShapingTableData(
  ViSession instrSession,
  ViInt32     arraySize,
  ViReal64 vin_vmax[],
  ViReal64 vcc_vmax[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(arraySize);
    LogAddReal64ArrArg(vin_vmax, &arraySize);
    LogAddReal64ArrArg(vcc_vmax, &arraySize);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar    command[50] = "";
  ViInt32   adaptation = 0;
  ViChar*   pbuffer = NULL;
  ViChar*   p2buf = NULL;
  ViInt32   i = 0;

  checkErr(Rs_LockSession(instrSession, NULL));

  if (arraySize <= 0)
    LogCloseCall(VI_SUCCESS);
    return VI_SUCCESS;

  viCheckErr(rssgt_GetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_ENVELOPE_TRACKING_VOLTAGE_ADAPTATION,
    &adaptation));

  // If Adaptation is AUTO (0), use command with ...SHAPing:PV:FILE...
  snprintf(command, 50, "SOUR:IQ:OUTP:ANAL:ENV:SHAP:%sFILE:DATA", (adaptation == 0) ? "PV:" : "");

  p2buf = pbuffer = (ViChar*)malloc((size_t)arraySize * 50 + 200);

  if (pbuffer)
  {
    p2buf += snprintf(pbuffer, 200, "%s %.12G,%.12G", command, vin_vmax[0], vcc_vmax[0]);

    for (i = 1; i < arraySize; i++)
      p2buf += snprintf(p2buf, 50, ",%.12G,%.12G", vin_vmax[i], vcc_vmax[i]);

    strcat(pbuffer, "\n");

    checkErr(rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)strlen(pbuffer), NULL));

    checkErr(rssgt_CheckStatus(instrSession));
  }

Error:
  if (pbuffer)
    free(pbuffer);
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function defines the polynomial coefficients.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR arraySize/Sets the amount of elements to be set with the arrays Real and 
/// HIPAR arraySize/Imaginary.
/// HIPAR polynomialCoefficients/Array of the real values of the polynomial coefficients. Amount
/// HIPAR polynomialCoefficients/of values used depends on the Array Size.
ViStatus _VI_FUNC rssgt_ConfigureEnvelopeTrackingPolynomialCoefficients(
  ViSession instrSession,
  ViInt32     arraySize,
  ViReal64 polynomialCoefficients[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(arraySize);
    LogAddReal64ArrArg(polynomialCoefficients, &arraySize);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar*   pbuffer = NULL;
  ViChar*   p2buf = NULL;
  ViInt32   i = 0;

  checkErr(Rs_LockSession(instrSession, NULL));

  if (arraySize <= 0)
    LogCloseCall(VI_SUCCESS);
    return VI_SUCCESS;

  p2buf = pbuffer = (ViChar*)malloc((size_t)arraySize * 25 + 200);

  if (pbuffer)
  {
    p2buf += snprintf(pbuffer, 200, "SOUR:IQ:OUTP:ANAL:ENV:SHAP:COEF %.12G", polynomialCoefficients[0]);

    for (i = 1; i < arraySize; i++)
      p2buf += snprintf(p2buf, 25, ",%.12G", polynomialCoefficients[i]);

    strcat(pbuffer, "\n");

    checkErr(rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)strlen(pbuffer), NULL));

    checkErr(rssgt_CheckStatus(instrSession));
  }

Error:
  if (pbuffer)
    free(pbuffer);
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function stores the correction values into a file with the 
/// HIFN  selected file name and loads it.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR distortionType/Selects the distortion type.
/// HIPAR filename/Sets the filename of the shaping table.
/// HIPAR arraySize/Sets the amount of elements to be set with the arrays Pin and Delta.
/// HIPAR pin/Array of the absolute input power Pin values. Number of elements of 
/// HIPAR pin/this array should be the same as number of elements of Delta array.
/// HIPAR delta/Array of the delta values. Number of elements of this array should be 
/// HIPAR delta/the same as number of elements of Pin array.
ViStatus _VI_FUNC rssgt_CreateNewDigitalPredistortionShapingTable(
  ViSession instrSession,
  ViInt32     distortionType,
  ViString filename,
  ViInt32     arraySize,
  ViReal64 pin[],
  ViReal64 delta[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(distortionType, DistortionType);
    LogAddStringArg(filename);
    LogAddInt32Arg(arraySize);
    LogAddReal64ArrArg(pin, &arraySize);
    LogAddReal64ArrArg(delta, &arraySize);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar    command[50] = "";
  ViChar*   pbuffer = NULL;
  ViChar*   p2buf = NULL;
  ViInt32   i = 0;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(distortionType, RSSGT_VAL_DISTORTION_AM_AM, RSSGT_VAL_DISTORTION_AM_PM), 2, "Distortion Type");

  if (arraySize <= 0)
    LogCloseCall(VI_SUCCESS);
    return VI_SUCCESS;

  snprintf(command, 50, "SOUR:IQ:DPD:SHAP:TABL:%s:FILE:NEW", (distortionType == RSSGT_VAL_DISTORTION_AM_AM) ? "AMAM" : "AMPM");

  p2buf = pbuffer = (ViChar*)malloc((size_t)arraySize * 50 + 200 + strlen(filename));

  if (pbuffer)
  {
    p2buf += snprintf(pbuffer, 300, "%s '%s',%.12G,%.12G", command, filename, pin[0], delta[0]);

    for (i = 1; i < arraySize; i++)
      p2buf += snprintf(p2buf, 50, ",%.12G,%.12G", pin[i], delta[i]);

    strcat(pbuffer, "\n");

    checkErr(rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)strlen(pbuffer), NULL));

    checkErr(rssgt_CheckStatus(instrSession));
  }

Error:
  if (pbuffer)
    free(pbuffer);
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function defines the predistortion function in a raw data format.
/// HIFN     
/// HIFN     Note(s):
/// HIFN     
/// HIFN     (1) Select a file to be defined with 
/// HIFN  RSSGT_ATTR_DIGITAL_PREDISTORTION_SHAPING_TABLE_AM_AM_SELECT_FILE or 
/// HIFN  RSSGT_ATTR_DIGITAL_PREDISTORTION_SHAPING_TABLE_AM_PM_SELECT_FILE 
/// HIFN  attributes before using this function.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR distortionType/Selects the distortion type.
/// HIPAR arraySize/Sets the amount of elements to be set with the arrays Pin and Delta.
/// HIPAR pin/Array of the absolute input power Pin values. Number of elements of 
/// HIPAR pin/this array should be the same as number of elements of Delta array.
/// HIPAR delta/Array of the delta values. Number of elements of this array should be 
/// HIPAR delta/the same as number of elements of Pin array.
ViStatus _VI_FUNC rssgt_ConfigureDigitalPredistortionShapingTableData(
  ViSession instrSession,
  ViInt32     distortionType,
  ViInt32     arraySize,
  ViReal64 pin[],
  ViReal64 delta[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(distortionType);
    LogAddInt32Arg(arraySize);
    LogAddReal64ArrArg(pin, &arraySize);
    LogAddReal64ArrArg(delta, &arraySize);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar    command[50] = "";
  ViChar*   pbuffer = NULL;
  ViChar*   p2buf = NULL;
  ViInt32   i = 0;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(distortionType, RSSGT_VAL_DISTORTION_AM_AM, RSSGT_VAL_DISTORTION_AM_PM), 2, "Distortion Type");

  if (arraySize <= 0)
    LogCloseCall(VI_SUCCESS);
    return VI_SUCCESS;

  snprintf(command, 50, "SOUR:IQ:DPD:SHAP:TABL:%s:FILE:DATA", (distortionType == RSSGT_VAL_DISTORTION_AM_AM) ? "AMAM" : "AMPM");

  p2buf = pbuffer = (ViChar*)malloc((size_t)arraySize * 50 + 200);

  if (pbuffer)
  {
    p2buf += snprintf(pbuffer, 200, "%s %.12G,%.12G", command, pin[0], delta[0]);

    for (i = 1; i < arraySize; i++)
      p2buf += snprintf(p2buf, 50, ",%.12G,%.12G", pin[i], delta[i]);

    strcat(pbuffer, "\n");

    checkErr(rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)strlen(pbuffer), NULL));

    checkErr(rssgt_CheckStatus(instrSession));
  }

Error:
  if (pbuffer)
    free(pbuffer);
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function defines the polynomial coefficients.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR arraySize/Sets the amount of elements to be set with the arrays Real and 
/// HIPAR arraySize/Imaginary.
/// HIPAR real/Array of the real values of the polynomial coefficients. Number of 
/// HIPAR real/elements of this array should be the same as number of elements of 
/// HIPAR real/Imaginary array.
/// HIPAR imaginary/Array of the imaginary values of the polynomial coefficients. Number 
/// HIPAR imaginary/of elements of this array should be the same as number of elements of 
/// HIPAR imaginary/Real array.
ViStatus _VI_FUNC rssgt_ConfigureDigitalPredistortionPolynomialCoefficients(
  ViSession   instrSession,
  ViInt32     arraySize,
  ViReal64    real[],
  ViReal64    imaginary[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(arraySize);
    LogAddReal64ArrArg(real, &arraySize);
    LogAddReal64ArrArg(imaginary, &arraySize);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar*   pbuffer = NULL;
  ViChar*   p2buf = NULL;
  ViInt32   i = 0;

  checkErr(Rs_LockSession(instrSession, NULL));

  if (arraySize <= 0)
    LogCloseCall(VI_SUCCESS);
    return VI_SUCCESS;

  p2buf = pbuffer = (ViChar*)malloc((size_t)arraySize * 50 + 200);

  if (pbuffer)
  {
    p2buf += snprintf(pbuffer, 200, "SOUR:IQ:DPD:SHAP:POLY:COEF %.12G,%.12G", real[0], imaginary[0]);

    for (i = 1; i < arraySize; i++)
      p2buf += snprintf(p2buf, 50, ",%.12G,%.12G", real[i], imaginary[i]);

    strcat(pbuffer, "\n");

    checkErr(rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)strlen(pbuffer), NULL));

    checkErr(rssgt_CheckStatus(instrSession));
  }

Error:
  if (pbuffer)
    free(pbuffer);
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function activates and deactivates pulse modulation. 
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_PULM_ENABLED 
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]PULM:STATe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR enabled/Activates and deactivates pulse modulation. 
ViStatus _VI_FUNC rssgt_ConfigurePULMEnabled(
  ViSession instrSession,
  ViBoolean enabled
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(enabled);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, "",
    RSSGT_ATTR_PULM_ENABLED,
    enabled), 2, "Enabled");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function selects the source for pulse modulation. 
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_PULM_SOURCE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [SOURce<[1]|2>:]PULM:SOURce
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR source/Selects the source for pulse modulation.
ViStatus _VI_FUNC rssgt_ConfigurePULMSource(
  ViSession instrSession,
  ViInt32 source
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(source, PULMSource);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_PULM_SOURCE,
    source), 2, "Source");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function configures Pulse Generator. 
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_PULM_MODE
/// HIFN     RSSGT_ATTR_PULM_PERIOD
/// HIFN     RSSGT_ATTR_PULM_DOUBLE_WIDTH
/// HIFN     RSSGT_ATTR_PULM_DOUBLE_DELAY
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     [:SOURce<HW>]:PULM:MODE
/// HIFN     [SOURce<HW>:]PULM:PERiod
/// HIFN     [SOURce:]PULM:DOUBle:WIDTh
/// HIFN     [SOURce:]PULM:DOUBle:DELay
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR pulseMode/Sets the operating mode of the pulse generator.
/// HIPAR pulsePeriod/Sets the period of the generated pulse. The period determines the 
/// HIPAR pulsePeriod/repetition frequency of the internal signal.
/// HIPAR pulseWidth/Sets the width of the generated pulse. The width determines the pulse 
/// HIPAR pulseWidth/length. The pulse width must be at least 20 ns less than the set pulse 
/// HIPAR pulseWidth/period.
/// HIPAR doublePulseWidth/Sets the width of the second pulse in case of double pulse generation.
/// HIPAR doublePulseDelay/Sets the delay from the start of the first pulse to the start of the 
/// HIPAR doublePulseDelay/second pulse.
ViStatus _VI_FUNC rssgt_ConfigurePULM(
  ViSession instrSession,
  ViInt32 pulseMode,
  ViReal64 pulsePeriod,
  ViReal64 pulseWidth,
  ViReal64 doublePulseWidth,
  ViReal64 doublePulseDelay
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(pulseMode, PULMMode);
    LogAddReal64Arg(pulsePeriod);
    LogAddReal64Arg(pulseWidth);
    LogAddReal64Arg(doublePulseWidth);
    LogAddReal64Arg(doublePulseDelay);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, "",
    RSSGT_ATTR_PULM_MODE,
    pulseMode), 2, "Pulse Mode");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_PULM_PERIOD,
    pulsePeriod), 3, "Pulse Period");
  viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
    RSSGT_ATTR_PULM_WIDTH,
    pulseWidth), 4, "Pulse Width");

  if (pulseMode == RSSGT_VAL_PULSE_MODE_DOUBLE)
  {
    viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
      RSSGT_ATTR_PULM_DOUBLE_WIDTH,
      doublePulseWidth), 5, "Double Pulse Width");
    viCheckParm(rssgt_SetAttributeViReal64(instrSession, "",
      RSSGT_ATTR_PULM_DOUBLE_DELAY,
      doublePulseDelay), 6, "Double Pulse Delay");
  }

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN Transfers the frequency data to the selected table.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR arraySize/Defines size of array <frequencies>.
/// HIPAR frequencies/Defines array of frequencies.
ViStatus _VI_FUNC rssgt_ConfigureUserCorrectionFrequencies(ViSession instrSession,
  ViInt32 arraySize,
  ViReal64 frequencies[])
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(arraySize);
    LogAddReal64ArrArg(frequencies, &arraySize);
  }
  ViStatus    error = VI_SUCCESS;
  ViChar*     pbuffer = NULL;
  ViChar*     p2buf = NULL;
  ViInt32     i = 0;

  if (arraySize <= 0)
    LogCloseCall(VI_SUCCESS);
    return VI_SUCCESS;

  p2buf = pbuffer = (ViChar*)malloc((size_t)arraySize * 50);

  if (pbuffer)
  {
    p2buf += snprintf(pbuffer, 50, ":CORR:CSET:DATA:FREQ %.12G",
      frequencies[0]);

    for (i = 1; i < arraySize; i++)
      p2buf += snprintf(p2buf, 20, ",%.12G", frequencies[i]);
    strcat(pbuffer, "\n");
    checkErr(rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)strlen(pbuffer), NULL));

    checkErr(rssgt_CheckStatus(instrSession));
  }
Error:
  if (pbuffer)
    free(pbuffer);

  LogCloseCall(error);
  return error;
}


/// HIFN Transfers the power data to the selected table.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR arraySize/Defines size of array <powers>.
/// HIPAR powers/Defines array of powers.
ViStatus _VI_FUNC rssgt_ConfigureUserCorrectionPowers(ViSession instrSession,
  ViInt32 arraySize,
  ViReal64 powers[])
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(arraySize);
    LogAddReal64ArrArg(powers, &arraySize);
  }
  ViStatus    error = VI_SUCCESS;
  ViChar*     pbuffer = NULL;
  ViChar*     p2buf = NULL;
  ViInt32     i = 0;

  if (arraySize <= 0)
    LogCloseCall(VI_SUCCESS);
    return VI_SUCCESS;

  p2buf = pbuffer = (ViChar*)malloc((size_t)arraySize * 50);

  if (pbuffer)
  {
    p2buf += snprintf(pbuffer, 50, ":CORR:CSET:DATA:POW %.12G",
      powers[0]);

    for (i = 1; i < arraySize; i++)
      p2buf += snprintf(p2buf, 20, ",%.12G", powers[i]);
    strcat(pbuffer, "\n");
    checkErr(rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)strlen(pbuffer), NULL));

    checkErr(rssgt_CheckStatus(instrSession));
  }

Error:
  if (pbuffer)
    free(pbuffer);

  LogCloseCall(error);
  return error;
}


/// HIFN  Configures the power measurement display.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_POWER_DISPLAY_PERMANENT_ENABLED
/// HIFN     RSSGT_ATTR_POWER_DISPLAY_PERMANENT
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     :SENSe<ch>[:POWer]:DISPlay:PERManent:STATe
/// HIFN     :SENSe<ch>[:POWer]:DISPlay:PERManent:PRIority
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR permanent/Activates the permanent display of the measured power level results in 
/// HIPAR permanent/the winbar. The instrument additionally indicates the sensor type, the 
/// HIPAR permanent/connection, the measurement source and the offset if set.
/// HIPAR priority/Selects average or peak power for permanent display.
ViStatus _VI_FUNC rssgt_ConfigurePowerMeasurementDisplay(
  ViSession instrSession,
  ViInt32 sensor,
  ViBoolean permanent,
  ViInt32 priority
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(sensor);
    LogAddBoolArg(permanent);
    LogAddSelectArg(priority, PowerDisplayPermanent);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar    rep_cap[5] = "";

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(sensor, 1, 4), 2, "Sensor");

  sprintf(rep_cap, "S" _PERCD, sensor);

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, rep_cap,
    RSSGT_ATTR_POWER_DISPLAY_PERMANENT_ENABLED,
    permanent), 3, "Permanent");

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, rep_cap,
    RSSGT_ATTR_POWER_DISPLAY_PERMANENT,
    priority), 4, "Priority");


Error:
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Performs zeroing of the sensor.
/// HIFN     
/// HIFN     Zeroing is required after warm-up, i.e. after connecting the sensor. 
/// HIFN  Additionally, it is recommended that you zero in regular intervals (at 
/// HIFN  least once a day), if the temperature has varied more than about 5 C, 
/// HIFN  if the sensor has been replaced, or if very lowpower are to be 
/// HIFN  measured.
/// HIFN     
/// HIFN     Note: The RF power source must be switched off or disconnected from 
/// HIFN  the sensor before zeroing.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_POWER_ZERO
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     :SENSe<ch>[:POWer]:ZERO
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR sensor/Sets the sensor number.
ViStatus _VI_FUNC rssgt_ConfigurePowerMeasurementZero(
  ViSession instrSession,
  ViInt32 sensor
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(sensor);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar    rep_cap[5] = "";

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(sensor, 1, 4), 2, "Sensor");

  sprintf(rep_cap, "S" _PERCD, sensor);

  viCheckErr(rssgt_SetAttributeViString(instrSession, rep_cap,
    RSSGT_ATTR_POWER_ZERO, NULL));

Error:
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Determines the signal to be measured. You can select the RF signal of 
/// HIFN  path A, path B or a user defined source.
/// HIFN     
/// HIFN     Note:
/// HIFN     When measuring the RF signal, the sensor considers the corresponding 
/// HIFN  correction factor at that frequency, and uses the level setting of the 
/// HIFN  instrument as reference level.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_POWER_SOURCE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     :SENSe<ch>[:POWer]:SOURce
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR source/Determines the signal to be measured. You can select the RF signal of 
/// HIPAR source/path A, path B or a user defined source.
ViStatus _VI_FUNC rssgt_ConfigurePowerMeasurementSource(
  ViSession instrSession,
  ViInt32 sensor,
  ViInt32 source
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(sensor);
    LogAddSelectArg(source, PowerSource);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar    rep_cap[5] = "";

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(sensor, 1, 4), 2, "Sensor");

  sprintf(rep_cap, "S" _PERCD, sensor);

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, rep_cap,
    RSSGT_ATTR_POWER_SOURCE,
    source), 3, "Source");

Error:
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the power measurement.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_POWER_OFFSET_ENABLED
/// HIFN     RSSGT_ATTR_POWER_OFFSET
/// HIFN     RSSGT_ATTR_POWER_FILTER_MODE
/// HIFN     RSSGT_ATTR_POWER_USE_SPARAMETERS
/// HIFN     RSSGT_ATTR_POWER_RECORDER_STATE
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     :SENSe<ch>[:POWer]:OFFSet:STATe
/// HIFN     :SENSe<ch>[:POWer]:OFFSet
/// HIFN     :SENSe<ch>[:POWer]:FILTer:TYPE
/// HIFN     :SENSe<ch>[:POWer]:CORRection:SPDevice:STATe
/// HIFN     :SENSe<ch>[:POWer]:LOGGing:STATe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR levelOffsetState/Activates the addition of the level offset to the measured value. The 
/// HIPAR levelOffsetState/level offset value is set with command :SENSe<ch>[:POWer]:OFFSet.
/// HIPAR levelOffset/Sets a level offset which is added to the measured level value after 
/// HIPAR levelOffset/activation with command SENSe[:POWer]:OFFSet:STATe ON. This allows e.g. 
/// HIPAR levelOffset/an attenuator in the signal path to be taken into account.
/// HIPAR filter/Selects the filter mode. The filter length is the multiplier for the 
/// HIPAR filter/time window and thus directly affects the measurement time.
/// HIPAR sparameter/Activates or deactivates the use of the s-parameters correction data 
/// HIPAR sparameter/of the connected power sensor. 
/// HIPAR logging/Activates the recording of the power values, measured by a connected 
/// HIPAR logging/R&S NRP-Z power sensor. 
ViStatus _VI_FUNC rssgt_ConfigurePowerMeasurement(
  ViSession instrSession,
  ViInt32 sensor,
  ViBoolean levelOffsetState,
  ViReal64 levelOffset,
  ViInt32 filter,
  ViBoolean sparameter,
  ViBoolean logging
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(sensor);
    LogAddBoolArg(levelOffsetState);
    LogAddReal64Arg(levelOffset);
    LogAddSelectArg(filter, PowerFilterMode);
    LogAddBoolArg(sparameter);
    LogAddBoolArg(logging);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar    rep_cap[5] = "";

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(sensor, 1, 4), 2, "Sensor");

  sprintf(rep_cap, "S" _PERCD, sensor);


  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, rep_cap,
    RSSGT_ATTR_POWER_OFFSET_ENABLED,
    levelOffsetState), 3, "Level Offset State");
  if (levelOffsetState == VI_TRUE)
  {
    viCheckParm(rssgt_SetAttributeViReal64(instrSession, rep_cap,
      RSSGT_ATTR_POWER_OFFSET,
      levelOffset), 4, "Level Offset");
  }

  viCheckParm(rssgt_SetAttributeViInt32(instrSession, rep_cap,
    RSSGT_ATTR_POWER_FILTER_MODE,
    filter), 5, "Filter");
  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, rep_cap,
    RSSGT_ATTR_POWER_USE_SPARAMETERS,
    sparameter), 6, "SParameter");
  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, rep_cap,
    RSSGT_ATTR_POWER_RECORDER_STATE,
    logging), 7, "Logging");

Error:
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Configures the power measurement aperture time.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_POWER_DEFAULT_APERTURE_TIME_ENABLED
/// HIFN     RSSGT_ATTR_POWER_APERTURE_TIME
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     :SENSe<ch>[:POWer]:APERture:DEFault:STATe
/// HIFN     :SENSe<ch>[:POWer]:APERture:TIMe
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR default/Deactivates the default aperture time of the respective sensor.
/// HIPAR aperturetime/Defines the aperture time (size of the acquisition interval) for the 
/// HIPAR aperturetime/corresponding sensor.
ViStatus _VI_FUNC rssgt_ConfigurePowerMeasurementApertureTime(
  ViSession instrSession,
  ViInt32 sensor,
  ViBoolean defaultTime,
  ViReal64 apertureTime
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(sensor);
    LogAddBoolArg(defaultTime);
    LogAddReal64Arg(apertureTime);
  }
  ViStatus  error = VI_SUCCESS;
  ViChar    rep_cap[5] = "";

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(sensor, 1, 4), 2, "Sensor");

  sprintf(rep_cap, "S" _PERCD, sensor);

  viCheckParm(rssgt_SetAttributeViBoolean(instrSession, rep_cap,
    RSSGT_ATTR_POWER_DEFAULT_APERTURE_TIME_ENABLED,
    defaultTime), 3, "Default Time");

  if (defaultTime == VI_FALSE)
  {
    viCheckParm(rssgt_SetAttributeViReal64(instrSession, rep_cap,
      RSSGT_ATTR_POWER_APERTURE_TIME,
      apertureTime), 4, "Aperture Time");
  }

Error:
  Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  Triggers power measurement and displays the results. The sensor 
/// HIFN  returns the result in the unit set with command :SENSe<ch>:UNIT[:POWer]
/// HIFN     Certain power sensors, such as the R&S NRP-Z81, return two values, 
/// HIFN  first the value of the average level and - separated by a comma - the 
/// HIFN  peak value.
/// HIFN     
/// HIFN     Note: This command does not affect the local state, i.e. you can get 
/// HIFN  results with local state on or off. For long measurement times it is 
/// HIFN  recommended that you use an SRQ for command synchronization (MAV bit).
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     :READ<ch>[:POWer]?
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR level_Peak/Indicates the measured peak level value.
/// HIPAR level_Average/Indicates the measured average level value.
ViStatus _VI_FUNC rssgt_ReadPowerMeasurement(ViSession instrSession,
  ViInt32 sensor, ViReal64 *level_Peak,
  ViReal64 *level_Average)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(sensor);
    LogAddReal64Out(level_Peak);
    LogAddReal64Out(level_Average);
  }
  ViStatus error = VI_SUCCESS;
  ViChar      *pbuffer = NULL;
  ViChar*     p2buf = NULL;
  ViChar      cmd[100];

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(sensor, 1, 4), 2, "Sensor");

  sprintf(cmd, ":READ" _PERCD ":POW?\n", sensor);
  checkErr(rsPrintf(instrSession, cmd));

  checkErr(Rs_ReadDataUnknownLength(instrSession, &pbuffer, NULL));

  p2buf = strtok(pbuffer, ",");

  if (p2buf != NULL)
  {
    *level_Peak = *level_Average = atof(p2buf);
    p2buf = strtok(NULL, ",\n");
  }

  if (p2buf != NULL)
  {
    *level_Peak = atof(p2buf);
  }

  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function sends a command to trigger the RF signal generator.  
/// HIFN  Call this function if you set a trigger source to a software trigger 
/// HIFN  value.  Below is a list of attributes and functions which can set the 
/// HIFN  trigger source to a software trigger.
/// HIFN     
/// HIFN     Functions: 
/// HIFN     rssgt_ConfigureArbTriggerSource
/// HIFN     rssgt_ConfigureSweep
/// HIFN     rssgt_ConfigureCDMATriggerSource
/// HIFN     rssgt_ConfigureTDMATriggerSource
/// HIFN     
/// HIFN     Attributes:
/// HIFN     RSSGT_ATTR_ARB_TRIGGER_SOURCE
/// HIFN     RSSGT_ATTR_SWEEP_TRIGGER_SOURCE
/// HIFN     RSSGT_ATTR_CDMA_TRIGGER_SOURCE
/// HIFN     RSSGT_ATTR_TDMA_TRIGGER_SOURCE
/// HIFN     
/// HIFN     Notes:
/// HIFN     
/// HIFN     This function does not check the instrument status.   Typically, you 
/// HIFN  call this function only in a sequence of calls to other low-level 
/// HIFN  driver functions.  The sequence performs one operation.  You use the 
/// HIFN  low-level functions to optimize one or more aspects of interaction with 
/// HIFN  the instrument.  If you want to check the instrument status, call the 
/// HIFN  rssgt_error_query function at the conclusion of the sequence.
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rssgt_SendSoftwareTrigger(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rssgt_SetAttributeViString(instrSession, "", RSSGT_ATTR_INIT_SW_TRIGGER, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN Sends a lock request ID which uniquely identifies the controller to the instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR requestID/Request lock from the controller with the specified Controller ID.
/// HIPAR requestResult/Returns the lock request.
ViStatus _VI_FUNC rssgt_Lock(ViSession instrSession, ViInt32 requestID,
  ViInt32 *requestResult)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(requestID);
    LogAddInt32Out(requestResult);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  checkErr(rsQueryf(instrSession, "LOCK? %ld\n", "%ld", requestID, requestResult));

  checkErr(rssgt_CheckStatus(instrSession));

Error:
  (void)(void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN Activates/deactivates the specified protection level.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR level/Selects the level of protection.
/// HIPAR enabled/Sets the protection state.
/// HIPAR key/Sets the key.
ViStatus _VI_FUNC rssgt_SystemProtect(ViSession instrSession,
  ViInt32 level, ViBoolean enabled, ViInt32 key)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(level);
    LogAddBoolArg(enabled);
    LogAddInt32Arg(key);
  }
  ViStatus error = VI_SUCCESS;
  ViChar      command[RSSGT_IO_BUFFER_SIZE] = "";

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(level, 1, 4), 2, "Level");

  if (enabled != VI_FALSE)
  {
    (void)snprintf(command, (size_t)RSSGT_IO_BUFFER_SIZE, ":SYST:PROT" _PERCD " ON\n", level);
  }
  else
  {
    (void)snprintf(command, (size_t)RSSGT_IO_BUFFER_SIZE, ":SYST:PROT" _PERCD " OFF," _PERCD "\n", level, key);
  }
  checkErr(rsWrite(instrSession, (ViBuf)command, (ViUInt32)strlen(command), NULL));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN Returns the used and available memory in Kb.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR totalPhysicalMemory/Total Physical Memory
/// HIPAR applicationMemory/Application Memory
/// HIPAR usedHeapMemory/Used Heap Memory
/// HIPAR availableHeapMemory/Available Heap Memory
ViStatus _VI_FUNC rssgt_GetMemoryInformation(ViSession instrSession,
  ViInt32 *totalPhysicalMemory,
  ViInt32 *applicationMemory,
  ViInt32 *usedHeapMemory,
  ViInt32 *availableHeapMemory)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Out(totalPhysicalMemory);
    LogAddInt32Out(applicationMemory);
    LogAddInt32Out(usedHeapMemory);
    LogAddInt32Out(availableHeapMemory);
  }
  ViStatus    error = VI_SUCCESS;
  ViChar*     p2buf = NULL;

  if (totalPhysicalMemory == NULL)
  {
    viCheckParm(RS_ERROR_NULL_POINTER, 2, "Total Physical Memory");
  }
  else if (applicationMemory == NULL)
  {
    viCheckParm(RS_ERROR_NULL_POINTER, 3, "Application Memory");
  }
  else if (usedHeapMemory == NULL)
  {
    viCheckParm(RS_ERROR_NULL_POINTER, 4, "Used Heap Memory");
  }
  else if (availableHeapMemory == NULL)
  {
    viCheckParm(RS_ERROR_NULL_POINTER, 5, "Available Heap Memory");
  }
  else
  {
    checkErr(rsWrite(instrSession, (ViBuf) ":MEM:HFR?\n", 10, NULL));
    checkErr(Rs_ReadDataUnknownLength(instrSession, &p2buf, NULL));

    if (p2buf)
      (void)sscanf(p2buf,
        "" _PERCD "," _PERCD "," _PERCD "," _PERCD,
        totalPhysicalMemory,
        applicationMemory,
        usedHeapMemory,
        availableHeapMemory);
  }

Error:
  if (p2buf)
    free(p2buf);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function reads out the subdirectories and files in the specified 
/// HIFN  directory. If no directory is specified, the default directory selected 
/// HIFN  with the MMEM:CDIR command is read out on the default drive selected 
/// HIFN  with the MMEM:MSIS command. 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR arraySize/Array size of the parameter labeled 'Catalog List'.
/// HIPAR catalogList/Reads back catalog list. Size of the list is limited by the 'Array 
/// HIPAR catalogList/Size' parameter.
ViStatus _VI_FUNC rssgt_MMEMoryCatalog(
  ViSession instrSession,
  ViString directoryPath,
  ViInt32 arraySize,
  ViChar catalogList[])
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(directoryPath);
    LogAddInt32Arg(arraySize);
    LogAddCharArrOut(catalogList, &arraySize);
  }
  ViStatus    error = VI_SUCCESS;
  ViChar      *pbuffer = NULL;
  ViUInt32    count = 0;

  checkErr(Rs_LockSession(instrSession, NULL));
  if (!directoryPath)
  {
    viCheckParm(RS_ERROR_INVALID_PARAMETER, 2, "Directory Path");
  }
  else
  {
    viCheckErr(rsPrintf(instrSession, ":MMEM:CAT? '%s'\n", directoryPath));
  }

  viCheckErr(Rs_ReadDataUnknownLength(instrSession, &pbuffer, &count));
  viCheckAlloc(pbuffer);

  checkErr(rssgt_CheckStatus(instrSession));
  if (count >= (ViUInt32)arraySize)
  {
    strncpy(catalogList, pbuffer, (size_t)(arraySize - 1));
    catalogList[arraySize - 1] = '\0';
    error = (ViStatus)(count + 1);
  }
  else
  {
    strncpy(catalogList, pbuffer, (size_t)count);
    catalogList[count] = '\0';
  }
Error:
  if (pbuffer) free(pbuffer);
  (void)(void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function reads out the number of files in the specified 
/// HIFN  directory. If no directory is specified, the default directory selected 
/// HIFN  with the MMEM:CDIR command is read out on the default drive selected 
/// HIFN  with the MMEM:MSIS command. 
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:CATalog:LENGth? <path>
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR catalogLength/Returns the number of files in the specified directory. If no 
/// HIPAR catalogLength/directory is specified, the default directory is used.
ViStatus _VI_FUNC rssgt_MMEMoryCatalogLength(
  ViSession instrSession,
  ViString directoryPath,
  ViInt32* catalogLength
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(directoryPath);
    LogAddInt32Out(catalogLength);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  if (directoryPath == NULL)
  {
    viCheckErr(rsQueryf(instrSession, "MMEM:CAT:LENG?\n", "%ld", catalogLength));
  }
  else
  {
    ViChar cmdbuffer[1024] = "";
    sprintf(cmdbuffer, "MMEM:CAT:LENG? '%s'\n", directoryPath);
    viCheckErr(rsQueryf(instrSession, cmdbuffer, "%ld", catalogLength));
  }

  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function changes the default directory. This directory is used 
/// HIFN  for all subsequent MMEM commands if no path is specified with them. It 
/// HIFN  is also possible to change to a higher directory using two dots '..'.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_MMEMORY_CHANGE_DIRECTORY
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:CDIRectory <directory_name>
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR directoryName/Directory name parameter.
ViStatus _VI_FUNC rssgt_MMEMoryChangeDirectory(
  ViSession instrSession,
  ViString directoryName
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(directoryName);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_MMEMORY_CHANGE_DIRECTORY,
    directoryName), 2, "Directory Name");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function copies the first specified file to the second specified 
/// HIFN  file. Instead of just a file, this function can also be used to copy a 
/// HIFN  complete directory together with all its files. 
/// HIFN     
/// HIFN     If <destination> is not specified, <source> is copied to the MMEM:MSIS 
/// HIFN  drive and the MMEM:CDIR directory. Files which already exist with the 
/// HIFN  same name in the destination directory are overwritten without an error 
/// HIFN  message.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:COPY <source>[,<destination>]
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR sourcePath/Source path of the file or a complete directory together with all its 
/// HIPAR sourcePath/files.
/// HIPAR destinationPath/Destination path of the file or a directory. If 'Destination Path' is 
/// HIPAR destinationPath/not specified, 'Source Path' is copied to the MMEM:MSIS drive and the 
/// HIPAR destinationPath/MMEM:CDIR directory.
ViStatus _VI_FUNC rssgt_MMEMoryCopy(
  ViSession instrSession,
  ViString sourcePath,
  ViString destinationPath
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(sourcePath);
    LogAddStringArg(destinationPath);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rsPrintf(instrSession, "MMEM:COPY '%s','%s'\n", sourcePath, destinationPath));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function writes the block data <binary block data> to the file 
/// HIFN  identified by <file_name>. The IEC/IEEE-bus terminator should be set to 
/// HIFN  EOI in order to ensure correct data transfer.
/// HIFN     
/// HIFN     This function can be used to transfer stored instrument settings or 
/// HIFN  waveforms directly to the instrument.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:DATA <file_name>[,<binary block data>]
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR fileName/File name of the binary block data.
/// HIPAR binaryBlockLength/Defines the number of bytes in the 'Binary Block Data' array.
/// HIPAR binaryBlockData/Binary block data array to be writen to the file defined by 'File 
/// HIPAR binaryBlockData/Name' parameter with the size limited with 'Binary Block Length' 
/// HIPAR binaryBlockData/parameter.
ViStatus _VI_FUNC rssgt_MMEMoryBlockDataWrite(
  ViSession instrSession,
  ViString fileName,
  ViInt32 binaryBlockLength,
  ViChar _VI_FAR binaryBlockData[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(fileName);
    LogAddInt32Arg(binaryBlockLength);
    LogAddCharArrArg(binaryBlockData, &binaryBlockLength);
  }
  ViStatus    rssgt_status = VI_SUCCESS;
  ViChar*     pbuffer;
  ViChar*     p2buf;
  size_t      size = strlen(fileName) + binaryBlockLength + 1000;

  p2buf = pbuffer = (ViChar*)malloc(size);

  if (pbuffer)
  {
    p2buf += snprintf(pbuffer, size, ":MMEM:DATA '%s',#" _PERCD "" _PERCD,
      fileName,
      (ViInt32)(floor(log10((double)binaryBlockLength)) + 1),
      binaryBlockLength);

    memcpy(p2buf, binaryBlockData, (size_t)binaryBlockLength);
    p2buf += binaryBlockLength;
    *p2buf++ = '\n';
    *p2buf = '\0';  // for safety reason
    rssgt_status = rsWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)(p2buf - pbuffer), NULL);

    free(pbuffer);
  }
  else
    rssgt_status = RS_ERROR_OUT_OF_MEMORY;

  LogCloseCall(rssgt_status);
  return rssgt_status;
}



/// HIFN  This function transfers the specified file from the R&S SMU to the 
/// HIFN  IEC/IEEE bus and then on to the control computer. It is important to 
/// HIFN  ensure that the intermediate memory on the control computer is large 
/// HIFN  enough to take the file. In this case, the setting for the IEC/IEEE-bus 
/// HIFN  terminator is irrelevant.
/// HIFN     
/// HIFN     This function can be used to read stored instrument settings or 
/// HIFN  waveforms directly from the instrument.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:DATA? <file_name>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR fileName/File name of the binary block data.
/// HIPAR binaryBlockLength/Length of the binary block data array.
/// HIPAR binaryBlockData/Binary block data to be retrieved from the file defined by the 'File 
/// HIPAR binaryBlockData/Name' parameter. Array size of this parameter must be large enough to 
/// HIPAR binaryBlockData/store all the retrieved data.
/// HIPAR bytesRead/Returns the number of bytes read from the file.
ViStatus _VI_FUNC rssgt_MMEMoryBlockDataRead(
  ViSession instrSession,
  ViString fileName,
  ViInt32 binaryBlockLength,
  ViChar _VI_FAR binaryBlockData[],
  ViUInt32    *bytesRead
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(fileName);
    LogAddUInt32Arg(binaryBlockLength);
    LogAddCharArrOut(binaryBlockData, &binaryBlockLength);
    LogAddUInt32Out(bytesRead);
  }
  ViStatus    error = VI_SUCCESS;
  ViUInt32    count = 0;
  ViUInt32    countInBinBlock = 0;
  int         lengthOfLength = 0;
  ViChar      buffer[BUFFER_SIZE] = "",
    format[32] = "",
    *pbuffer = NULL, *p2buffer = NULL;

  (void)snprintf(buffer, (size_t)BUFFER_SIZE, ":MMEM:DATA? '%s'\n", fileName);

  checkErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));
  checkErr(rssgt_readDataUnknownLength(instrSession, &pbuffer, &count));

  checkAlloc(pbuffer);

  p2buffer = pbuffer;

  if (sscanf(p2buffer, "#%1d", &lengthOfLength) != 1)
  {
    free(pbuffer);
    LogCloseCall(VI_ERROR_INSTR_INTERPRETING_RESPONSE);
    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
  }

  (void)snprintf(format, 10, "%%%dld", lengthOfLength);
  p2buffer += 2;

  if (sscanf(p2buffer, format, &countInBinBlock) != 1)
  {
    free(pbuffer);
    LogCloseCall(VI_ERROR_INSTR_INTERPRETING_RESPONSE);
    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
  }

  if (count != countInBinBlock + lengthOfLength + 3)
  {
    free(pbuffer);
    LogCloseCall(VI_ERROR_INSTR_INTERPRETING_RESPONSE);
    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
  }


  p2buffer += lengthOfLength;

  if ((error = rssgt_CheckStatus(instrSession)) & _VI_ERROR)
  {
    free(pbuffer);
    LogCloseCall(error);
    return error;
  }


  if (countInBinBlock > (ViUInt32)binaryBlockLength)
  {
    error = RSSGT_WARNING_SMALL_BUFFER;
    countInBinBlock = (ViUInt32)binaryBlockLength;
    memcpy(binaryBlockData, p2buffer, (size_t)countInBinBlock);
  }
  else
  {
    error = VI_SUCCESS;
    memcpy(binaryBlockData, p2buffer, (size_t)countInBinBlock);
  }

  if (bytesRead)
    *bytesRead = countInBinBlock;

Error:
  if (pbuffer)
    free(pbuffer);

  LogCloseCall(error);
  return error;
}


/// HIFN  This function is used to read data from the instrument and write
/// HIFN  it to a user specified file on the host computer.     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR source/This control selects file for the data transfer from instrument
/// HIPAR source/to control computer.
/// HIPAR destination/This control defines destination file to which the data transfer
/// HIPAR destination/from instrument to control computer applies.
ViStatus _VI_FUNC rssgt_MMEMoryReadToFileFromInstrument(ViSession instrSession,
  ViString source,
  ViString destination)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(source);
    LogAddStringArg(destination);
  }
  ViStatus    error = VI_SUCCESS;
  ViUInt16    intfType = VI_INTF_GPIB;
  ViUInt32    count = 0,
    ret_cnt = 0;
  ViInt64     length = 0;
  ViChar      buffer[BUFFER_SIZE] = "",
    format[32] = "",
    length_of_length[10] = "",
    *pbuffer = NULL;
  FILE        *targetFile;
  ViChar      firstChar[20] = "";
  ViUInt32    block_length = 1000000;
  ViUInt32    cnt = 0;
  int nTotalCount = 0;

  checkErr(rsGetAttribute(instrSession, VI_ATTR_INTF_TYPE, &intfType));

  if (intfType == VI_INTF_GPIB)
    checkErr(rsWrite(instrSession, (ViBuf) ":SYST:COMM:GPIB:LTER EOI\n", 25, NULL));

  sprintf(buffer, ":MMEM:DATA? '%s'\n", source);

  checkErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));

  // read out #
  checkErr(rsRead(instrSession, (ViBuf)firstChar, 1, &ret_cnt));
  if (*firstChar != '#')
  {
    error = RS_ERROR_READING_FILE;
    goto Error;
  }

  checkErr(rsRead(instrSession, (ViPBuf)firstChar, 1, &ret_cnt));

  firstChar[1] = '\0';
  if (strncmp(firstChar, "(", 1) == 0)// result is like this #(length)data...
  {
    int i = 0;
    for (i = 0; i < 12; i++)
    {
      unsigned char c = 0;
      rsRead(instrSession, &c, 1, &ret_cnt);
      if (ret_cnt == 0 || c == ')')
        break;
      firstChar[i] = c;
    }
    firstChar[i] = '\0';
    length = atol(firstChar);

    if (length < 0)
    {
      // the current implementation of MMEM:DATA? is faulty
      // and returns a negative file length
      error = RS_ERROR_READING_FILE;
      goto Error;
    }
  }
  else
  {
    sprintf(format, "%%%sllu", firstChar); // results is like this #lenghtoflengthLENGTHdata...
    cnt = (ViUInt32)atol(firstChar);

    checkErr(rsRead(instrSession, (ViPBuf)length_of_length, cnt, NULL));
    sscanf(length_of_length, format, &length);
  }

  pbuffer = (ViChar*)malloc(block_length);

  /* Store data from instrument to file on host computer */
  if ((targetFile = fopen(destination, "wb")) == NULL)
  {
    if (pbuffer)
      free(pbuffer);
    LogCloseCall(RS_ERROR_CANNOT_OPEN_FILE);
    return RS_ERROR_CANNOT_OPEN_FILE;
  }

  do
  {
    checkErr(rsRead(instrSession, (ViPBuf)pbuffer, block_length, &count));

    nTotalCount += count;
    if (nTotalCount > length)
    {
      fwrite(pbuffer, sizeof(ViByte), (size_t)(count - 1), targetFile);
    }
    else
    {
      if (count != (ViUInt32)fwrite(pbuffer, sizeof(ViByte), (size_t)(count), targetFile))
      {
        fclose(targetFile);
        if (pbuffer)
          free(pbuffer);
        LogCloseCall(RS_ERROR_WRITING_FILE);
        return RS_ERROR_WRITING_FILE;
      }
    }
  } while (nTotalCount <= length); // block_length == count); // rsRead may return patially filled buffer not only at the end of a block read!

  if (intfType == VI_INTF_GPIB)
    checkErr(rsWrite(instrSession, (ViBuf) ":SYST:COMM:GPIB:LTER STAN\n", 26, NULL));

  checkErr(rssgt_CheckStatus(instrSession));

  fclose(targetFile);

Error:
  if (pbuffer != NULL)
    free(pbuffer);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function is used to read data from the host computer and
/// HIFN  write it to a user specified file in the instrument.     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR source/This control selects file for the data transfer from control
/// HIPAR source/computer to instrument.
/// HIPAR destination/This control defines destination file to which the data transfer
/// HIPAR destination/from control computer to instrument applies.
ViStatus _VI_FUNC rssgt_MMEMoryWriteFromFileToInstrument(ViSession instrSession,
  ViString source,
  ViString destination)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(source);
    LogAddStringArg(destination);
  }
  ViStatus  error = VI_SUCCESS;
  ViUInt32    block_length = 1000000;
  ViChar      *pbuffer = NULL;
  size_t      read = 0;
  size_t   file_size = 0;
  size_t   remaining = 0;

  FILE        *source_file = 0;
  ViUInt16    intfType = VI_INTF_GPIB;
  ViChar      buffer[BUFFER_SIZE] = "";

  if ((source_file = fopen(source, "rb")) == NULL)
  {
    LogCloseCall(RS_ERROR_CANNOT_OPEN_FILE);
    return RS_ERROR_CANNOT_OPEN_FILE;
  }

  /* Get size of file */
  fseek(source_file, 0, SEEK_END);
  remaining = file_size = ftell(source_file);
  fseek(source_file, 0, SEEK_SET);

  checkErr(rsGetAttribute(instrSession, VI_ATTR_INTF_TYPE, &intfType));
  if (intfType == VI_INTF_GPIB)
    rsWrite(instrSession, (ViBuf) ":SYST:COMM:GPIB:LTER STAN\n", 26, NULL);

  checkErr(rsSetAttribute(instrSession, VI_ATTR_SEND_END_EN, VI_FALSE));

  if (file_size <= 999999999)
    sprintf(buffer, ":MMEM:DATA '%s',#" _PERCD "%ld", destination, (ViInt32)(floor(log10((double)file_size)) + 1), (long)file_size);
  else
    sprintf(buffer, ":MMEM:DATA '%s',#(%ld)", destination, (long)file_size);

  checkErr(rsWrite(instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));

  pbuffer = (ViChar*)malloc(block_length);
  if (pbuffer == NULL)
    error = VI_ERROR_ALLOC;

  while (error == VI_SUCCESS && remaining > 0)
  {
    read = fread(pbuffer, 1, block_length, source_file);
    if (read == 0)
    {
      error = VI_ERROR_FILE_IO;
      break;
    }
    remaining -= read;
    if (remaining < block_length)
      block_length = (ViUInt32)remaining;

    error = rsWrite(instrSession, (ViBuf)pbuffer, (ViUInt32)read, NULL);

  } // while


  rsSetAttribute(instrSession, VI_ATTR_SEND_END_EN, VI_TRUE);
  rsWrite(instrSession, (ViBuf) "\n", 1, NULL);

  if (intfType == VI_INTF_GPIB)
    rsWrite(instrSession, (ViBuf) ":SYST:COMM:GPIB:LTER STAN\n", 26, NULL);

Error:
  fclose(source_file);

  if (pbuffer != NULL)
    free(pbuffer);

  LogCloseCall(error);
  return error;
}


/// HIFN  This function reads out the subdirectories of the specified directory. 
/// HIFN  If no directory is specified, the default directory selected with the 
/// HIFN  MMEM:CDIR command is read out. The directories are output in a list 
/// HIFN  (the list entries are separated by commas). 
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:DCATalog? <path>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR arraySize/Array size of the parameter labeled 'Catalog List'.
/// HIPAR catalogList/Reads back catalog list. Size of the list is limited by the 'Array 
/// HIPAR catalogList/Size' parameter.
ViStatus _VI_FUNC rssgt_MMEMoryDirectoryCatalog(
  ViSession instrSession,
  ViString directoryPath,
  ViInt32 arraySize,
  ViChar _VI_FAR catalogList[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(directoryPath);
    LogAddInt32Arg(arraySize);
    LogAddCharArrOut(catalogList, &arraySize);
  }
  ViStatus    error = VI_SUCCESS;
  ViChar      *pbuffer = NULL;
  ViUInt32    count = 0;

  checkErr(Rs_LockSession(instrSession, NULL));
  if (!directoryPath)
  {
    viCheckParm(RS_ERROR_INVALID_PARAMETER, 2, "Directory Path");
  }
  else
  {
    viCheckErr(rsPrintf(instrSession, ":MMEM:DCAT? '%s'\n", directoryPath));
  }

  viCheckErr(Rs_ReadDataUnknownLength(instrSession, &pbuffer, &count));
  checkAlloc(pbuffer);
  checkErr(rssgt_CheckStatus(instrSession));
  //strncpy(output, pbuffer, (size_t) bufferSize);
  if (count >= (ViUInt32)arraySize)
  {
    strncpy(catalogList, pbuffer, (size_t)(arraySize - 1));
    catalogList[arraySize - 1] = '\0';
    error = (ViStatus)(count + 1);
  }
  else
  {
    strncpy(catalogList, pbuffer, (size_t)count);
    catalogList[count] = '\0';
  }
Error:
  if (pbuffer) free(pbuffer);
  (void)(void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function reads out the number of subdirectories in the specified 
/// HIFN  directory. If no directory is specified, the directory selected with 
/// HIFN  the MMEM:CDIR command is read out.  
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:DCATalog:LENGth? <path>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR catalogLength/Returns the number of subdirectories in the specified directory. If no 
/// HIPAR catalogLength/directory is specified, the default directory is used.
ViStatus _VI_FUNC rssgt_MMEMoryDirectoryCatalogLength(
  ViSession instrSession,
  ViString directoryPath,
  ViInt32* catalogLength
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(directoryPath);
    LogAddInt32Out(catalogLength);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  if (directoryPath == NULL)
  {
    viCheckErr(rsQueryf(instrSession, "MMEM:DCAT:LENG?\n", "%ld", catalogLength));
    viCheckErr(rssgt_CheckStatus(instrSession));
  }
  else
  {
    viCheckErr(rsQueryf(instrSession, "MMEM:DCAT:LENG? '%s'\n", "%ld", directoryPath, catalogLength));
    viCheckErr(rssgt_CheckStatus(instrSession));
  }

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function deletes the specified file.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:DELete <file_name>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR fileName/File name parameter.
ViStatus _VI_FUNC rssgt_MMEMoryDelete(
  ViSession instrSession,
  ViString fileName
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(fileName);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_MMEMORY_DELETE,
    fileName), 2, "File Name");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function loads the specified file stored under the specified name 
/// HIFN  in an internal memory.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:LOAD:STATe 0,<file_name>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR instrumentSetting/Instrument setting number.
/// HIPAR fileName/File name parameter.
ViStatus _VI_FUNC rssgt_MMEMoryLoad(
  ViSession instrSession,
  ViInt32 instrumentSetting,
  ViString fileName
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(instrumentSetting);
    LogAddStringArg(fileName);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rsPrintf(instrSession, "MMEM:LOAD:STAT %ld,'%s'\n", instrumentSetting, fileName));
  viCheckErr(rssgt_CheckStatus(instrSession));

  viCheckErr(rsPrintf(instrSession, "*RCL %d\n", instrumentSetting));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function creates a new subdirectory in the specified directory. 
/// HIFN  If no directory is specified, a subdirectory is created in the default 
/// HIFN  directory. This function can also be used to create a directory tree.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:MDIRectory <directory_name>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR directoryName/Directory name parameter.
ViStatus _VI_FUNC rssgt_MMEMoryMakeDirectory(
  ViSession instrSession,
  ViString directoryName
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(directoryName);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_MMEMORY_MAKE_DIRECTORY,
    directoryName), 2, "Directory Name");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function renames an existing file ('File Source') if no path is 
/// HIFN  specified for 'File Destination'. Otherwise the file is moved to the 
/// HIFN  specified path and stored under the original file name or, if 
/// HIFN  specified, a new file name.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:MOVE <file_source>,<file_destination>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR fileSource/Define file source.
/// HIPAR fileDestination/Define file destination.
ViStatus _VI_FUNC rssgt_MMEMoryMove(
  ViSession instrSession,
  ViString fileSource,
  ViString fileDestination
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(fileSource);
    LogAddStringArg(fileDestination);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rsPrintf(instrSession, "MMEM:MOVE '%s','%s'\n", fileSource, fileDestination));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function deletes the specified subdirectory in the specified 
/// HIFN  directory. If no directory is specified, the subdirectory with the 
/// HIFN  specified name is deleted in the default directory.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:RDIRectory <directory_name>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR directoryName/Directory name parameter.
ViStatus _VI_FUNC rssgt_MMEMoryRemoveDirectory(
  ViSession instrSession,
  ViString directoryName
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(directoryName);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_MMEMORY_REMOVE_DIRECTORY,
    directoryName), 2, "Directory Name");

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function stores the current instrument setting in the specified 
/// HIFN  file.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     MMEMory:STORe:STATe 0,<file_name>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR instrumentSetting/Instrument setting number.
/// HIPAR fileName/File name parameter.
ViStatus _VI_FUNC rssgt_MMEMoryStore(
  ViSession instrSession,
  ViInt32 instrumentSetting,
  ViString fileName
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(instrumentSetting);
    LogAddStringArg(fileName);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rsPrintf(instrSession, "*SAV %d\n", instrumentSetting));
  viCheckErr(rssgt_CheckStatus(instrSession));

  viCheckErr(rsPrintf(instrSession, "MMEM:STOR:STAT %ld,'%s'\n", instrumentSetting, fileName));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function triggers voltage measurement at the specified test point 
/// HIFN  and returns the measured voltage. A detailed description of the test 
/// HIFN  points can be found in Chapter 3 of the Service Manual (supplied with 
/// HIFN  the instrument). 
/// HIFN     
/// HIFN     A list of the available test points for the respective path can be 
/// HIFN  queried using the :DIAG<[1]|2>:POIN:CAT? command .   
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     DIAGnostic[:MEASure:]POINt? <test point name>
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR testPointName/Test point name.
/// HIPAR measuredVoltage/Returns the voltage measured at the specified test point.
ViStatus _VI_FUNC rssgt_DiagnosticTestPointVoltages(
  ViSession instrSession,
  ViString testPointName,
  ViReal64* measuredVoltage
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(testPointName);
    LogAddReal64Out(measuredVoltage);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckErr(rsQueryf(instrSession, "DIAG:POIN? '%s'\n", "%lf", testPointName, measuredVoltage));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function sets the value of selected status register.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     STATus:OPERation:PTRansition
/// HIFN     STATus:OPERation:NTRansition
/// HIFN     STATus:OPERation:ENABle
/// HIFN     STATus:QUEStionable:PTRansition
/// HIFN     STATus:QUEStionable:NTRansition
/// HIFN     STATus:QUEStionable:ENABle
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR registerGroup/Selects the status register group.
/// HIPAR registerType/Selects the type of status register.
/// HIPAR value/Sets the value to be set to selected register.
ViStatus _VI_FUNC rssgt_SetRegister(
  ViSession instrSession,
  ViInt32 registerGroup,
  ViInt32 registerType,
  ViInt32 value
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(registerGroup);
    LogAddInt32Arg(registerType);
    LogAddInt32Arg(value);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(registerGroup, 0, 1), 2, "Register Group");
  viCheckParm(rssgt_invalidViInt32Range(registerType, 2, 4), 3, "Register Type");

  viCheckErr(rsPrintf(instrSession, ":STAT:%s:%s %ld\n", regGroupArr[registerGroup], regTypeArr[registerType], value));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function sets the value of selected status register.
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     STATus:OPERation:EVENt?
/// HIFN     STATus:OPERation:CONDition?
/// HIFN     STATus:OPERation:PTRansition?
/// HIFN     STATus:OPERation:NTRansition?
/// HIFN     STATus:QUEStionable:EVENt?
/// HIFN     STATus:QUEStionable:CONDition?
/// HIFN     STATus:QUEStionable:PTRansition?
/// HIFN     STATus:QUEStionable:NTRansition?
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR registerGroup/Selects the status register group.
/// HIPAR registerType/Selects the type of status register.
/// HIPAR value/Returns the value of selected register.
ViStatus _VI_FUNC rssgt_GetRegister(
  ViSession instrSession,
  ViInt32 registerGroup,
  ViInt32 registerType,
  ViInt32* value
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(registerGroup);
    LogAddInt32Arg(registerType);
    LogAddInt32Out(value);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(registerGroup, 0, 1), 2, "Register Group");
  viCheckParm(rssgt_invalidViInt32Range(registerType, 0, 4), 3, "Register Type");

  viCheckErr(rsQueryf(instrSession, ":STAT:%s:%s?\n", "%ld", regGroupArr[registerGroup], regTypeArr[registerType], value));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function resets the status registers. All PTRansition parts are 
/// HIFN  set to FFFFh (32767), i.e. all transitions from 0 to 1 are detected. 
/// HIFN  All NTRansition parts are set to 0, i.e. a transition from 1 to 0 in a 
/// HIFN  CONDition bit is not detected. The ENABle parts of STATus:OPERation and 
/// HIFN  STATus:QUEStionable are set to 0, i.e. all events in these registers 
/// HIFN  are not passed on.
/// HIFN     
/// HIFN     Attribute(s):
/// HIFN     RSSGT_ATTR_STATUS_PRESET
/// HIFN     
/// HIFN     Remote-control command(s):
/// HIFN     STATus:PRESet
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgt_STATusPRESet(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));
  viCheckErr(rssgt_SetAttributeViString(instrSession, "",
    RSSGT_ATTR_STATUS_PRESET, NULL));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}



/// HIFN Sets the date for the instrument-internal calendar.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR year/Year
/// HIPAR month/Month
/// HIPAR day/Day
ViStatus _VI_FUNC rssgt_SetDate(ViSession instrSession, ViInt32 year,
  ViInt32 month, ViInt32 day)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(year);
    LogAddInt32Arg(month);
    LogAddInt32Arg(day);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(month, 1, 12), 3, "Month");
  viCheckParm(rssgt_invalidViInt32Range(day, 1, 31), 4, "Day");

  checkErr(rsPrintf(instrSession, ":SYST:DATE %ld,%ld,%ld\n", year, month, day));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN Gets the date for the instrument-internal calendar.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR year/Year
/// HIPAR month/Month
/// HIPAR day/Day
ViStatus _VI_FUNC rssgt_GetDate(ViSession instrSession, ViInt32 *year,
  ViInt32 *month, ViInt32 *day)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Out(year);
    LogAddInt32Out(month);
    LogAddInt32Out(day);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  checkErr(rsQueryf(instrSession, ":SYST:DATE?\n", "%ld,%ld,%ld", year, month, day));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN Sets the time for the instrument-internal clock.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR hour/Hour
/// HIPAR minute/Minute
/// HIPAR second/Second
ViStatus _VI_FUNC rssgt_SetTime(ViSession instrSession, ViInt32 hour,
  ViInt32 minute, ViInt32 second)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(hour);
    LogAddInt32Arg(minute);
    LogAddInt32Arg(second);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  viCheckParm(rssgt_invalidViInt32Range(hour, 0, 23), 3, "Hour");
  viCheckParm(rssgt_invalidViInt32Range(second, 0, 59), 4, "Minute");
  viCheckParm(rssgt_invalidViInt32Range(second, 0, 59), 5, "Second");

  checkErr(rsPrintf(instrSession, ":SYST:TIME %ld,%ld,%ld\n", hour, minute, second));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN Gets the time for the instrument-internal clock.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR hour/Hour
/// HIPAR minute/Minute
/// HIPAR second/Second
ViStatus _VI_FUNC rssgt_GetTime(ViSession instrSession, ViInt32 *hour,
  ViInt32 *minute, ViInt32 *second)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Out(hour);
    LogAddInt32Out(minute);
    LogAddInt32Out(second);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  checkErr(rsQueryf(instrSession, ":SYST:TIME?\n", "%ld,%ld,%ld", hour, minute, second));
  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function clears status.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
ViStatus _VI_FUNC rssgt_ClearStatus(ViSession instrSession)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  checkErr(rssgt_WriteInstrData(instrSession, "*CLS\n"));

  checkErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function returns the ID Query response string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR IDQueryResponse/Returns the ID Query response string.
ViStatus _VI_FUNC rssgt_IDQueryResponse(ViSession instrSession,
  ViChar IDQueryResponse[])
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringOut(IDQueryResponse);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  checkErr(rsQueryf(instrSession, "*IDN?\n", "%s", IDQueryResponse));

  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  Stops further command processing until all commands sent before
/// HIFN  *WAI have been executed.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
ViStatus _VI_FUNC rssgt_ProcessAllPreviousCommands(ViSession instrSession)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  checkErr(rssgt_WriteInstrData(instrSession, "*WAI\n"));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function queries the OPC.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a particular 
/// HIPAR instrSession/instrument session.
/// HIPAR OPC/Queries the OPC.
ViStatus _VI_FUNC rssgt_QueryOPC(ViSession instrSession, ViInt32 *OPC)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Out(OPC);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  checkErr(rsQueryf(instrSession, "*OPC?\n", "%ld", OPC));

  viCheckErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function retrieves and then clears the IVI error information for 
/// HIFN  the session or the current execution thread. One exception exists: If 
/// HIFN  the BufferSize parameter is 0, the function does not clear the error 
/// HIFN  information. By passing 0 for the buffer size, the caller can ascertain 
/// HIFN  the buffer size required to get the entire error description string and 
/// HIFN  then call the function again with a sufficiently large buffer.
/// HIFN     
/// HIFN     If the user specifies a valid IVI session for the instrSession 
/// HIFN  parameter, Get Error retrieves and then clears the error information 
/// HIFN  for the session.  If the user passes NULL for the instrSession 
/// HIFN  parameter, this function retrieves and then clears the error 
/// HIFN  information for the current execution thread.  If the instrSession 
/// HIFN  parameter is an invalid session, the function does nothing and returns 
/// HIFN  an error. Normally, the error information describes the first error 
/// HIFN  that occurred since the user last called rssgt_GetError or 
/// HIFN  rssgt_ClearError.
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR code/Returns the error code for the session or execution thread.
/// HIPAR buffersize/Pass the number of bytes in the ViChar array you specify for the 
/// HIPAR buffersize/Description parameter.
/// HIPAR description/Returns the error description for the IVI session or execution thread. 
/// HIPAR description/ If there is no description, the function returns an empty string.
ViStatus _VI_FUNC rssgt_GetError(
  ViSession instrSession,
  ViStatus* code,
  ViInt32 buffersize,
  ViChar description[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Out(code);
    LogAddInt32Arg(buffersize);
    LogAddCharArrOut(description, &buffersize);
  }
  ViStatus    error = VI_SUCCESS;
  ViStatus    primary = VI_SUCCESS,
    secondary = VI_SUCCESS;
  ViChar      elaboration[RS_MAX_MESSAGE_BUF_SIZE] = "";
  ViChar      errorMessage[RS_MAX_MESSAGE_BUF_SIZE + 256] = "";
  ViChar     *appendPoint = errorMessage;
  ViInt32     actualSize = 0;
  ViBoolean   locked = VI_FALSE;

  /* Lock Session */
  checkErr(Rs_LockSession(instrSession, &locked));

  /* Test for nulls and acquire error data */
  if (buffersize != 0)
  {
    if (code == NULL)
    {
      viCheckParm(RS_ERROR_INVALID_PARAMETER, 2, "Null address for Error");
    }
    checkErr(Rs_GetErrorInfo(instrSession, &primary, &secondary, elaboration));
  }

  else
  {
    checkErr(Rs_GetAttribute(instrSession, NULL, RS_ATTR_ERROR_ELABORATION, 0, 256, elaboration));
    checkErr(Rs_GetAttribute(instrSession, NULL, RS_ATTR_SECONDARY_ERROR, 0, (unsigned int) sizeof(secondary), &secondary));
    checkErr(Rs_GetAttribute(instrSession, NULL, RS_ATTR_PRIMARY_ERROR, 0, (unsigned int) sizeof(primary), &primary));
  }

  if (description == NULL)
  {
    viCheckParm(RS_ERROR_INVALID_PARAMETER, 4, "Null address for Description");
  }

  /* Format data */
  if (primary != VI_SUCCESS)
  {
    ViChar msg[256] = "";
    checkErr(rssgt_error_message(instrSession, primary, msg));
    appendPoint += snprintf(appendPoint, 288, "Primary Error: (Hex 0x%08X) %s\n", (unsigned int)primary, msg);
  }

  if (secondary != VI_SUCCESS)
  {
    ViChar msg[256] = "";
    checkErr(rssgt_error_message(instrSession, secondary, msg));
    appendPoint += snprintf(appendPoint, 290, "Secondary Error: (Hex 0x%08X) %s\n", (unsigned int)secondary, msg);
  }

  if (strlen(elaboration) > 0)
  {
    (void)snprintf(appendPoint, RS_MAX_MESSAGE_BUF_SIZE + 15, "Elaboration: %s", elaboration);
  }

  actualSize = (ViInt32)(strlen(errorMessage) + 1);

  /* Prepare return values */
  if (buffersize == 0)
  {
    error = (ViStatus)actualSize;
  }
  else
  {
    if (buffersize > 0)
    {
      if (actualSize > buffersize)
      {
        error = actualSize;
        actualSize = buffersize;
      }
    }
    if (description != NULL)
    {
      memcpy(description, errorMessage, (size_t)(actualSize - 1));
      description[actualSize - 1] = '\0';
    }
  }

  if (code)
  {
    *code = primary;
  }

Error:
  /* Unlock Session */
  (void)Rs_UnlockSession(instrSession, &locked);
  LogCloseCall(error);
  return error;
}



/// HIFN  This function clears the error code and error description for the IVI 
/// HIFN  session. If the user specifies a valid IVI session for the 
/// HIFN  instrSession parameter, this function clears the error information 
/// HIFN  for the session. If the user passes NULL for the Vi parameter, this 
/// HIFN  function clears the error information for the current execution thread. 
/// HIFN  If the Vi parameter is an invalid session, the function does nothing 
/// HIFN  and returns an error.
/// HIFN     The function clears the error code by setting it to VI_SUCCESS.  If 
/// HIFN  the error description string is non-NULL, the function de-allocates the 
/// HIFN  error description string and sets the address to NULL.
/// HIFN     
/// HIFN     Maintaining the error information separately for each thread is useful 
/// HIFN  if the user does not have a session handle to pass to the 
/// HIFN  rssgt_GetError function, which occurs when a call to 
/// HIFN  rssgt_init or rssgt_InitWithOptions fails.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rssgt_ClearError(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = Rs_ClearErrorInfo(instrSession);
  LogCloseCall(error);
  return error; 
}



/// HIFN  This function writes data to the instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR writeBuffer/Pass the string to be written to the instrument.
ViStatus _VI_FUNC rssgt_WriteInstrData(ViSession instrSession,
  ViChar writeBuffer[])
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(writeBuffer);
  }
#if 0
  return Rs_WriteInstrData(instrSession, writeBuffer);
#else
  ViStatus    retval = VI_SUCCESS;

  if (writeBuffer != NULL)
  {  /* Only transmit really existing data */
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE] = "";
    ViUInt32    slen = (ViUInt32)strlen(writeBuffer);

    if (slen < RS_MAX_MESSAGE_BUF_SIZE - 2)
    {  /* Append a '\n' to the command string & take into account the terminating '0' */
      if (slen > 0)
        memcpy(buffer, writeBuffer, (size_t)slen);
      buffer[slen++] = '\n';
      buffer[slen] = '\0';
      retval = Rs_WriteInstrData(instrSession, buffer);
    }
    else
    {  /* Large data transfer - issue a second call */
      retval = Rs_WriteInstrData(instrSession, writeBuffer);
      (void)Rs_WriteInstrData(instrSession, (ViChar*) "\n");
    }
  }  /* end if (writeBuffer != NULL) */

  LogCloseCall(retval);
  return retval;
#endif
}


/// HIFN  This function reads data from the instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR numberofBytesToRead/Pass the maximum number of bytes to read from the instruments.  
/// HIPAR readBuffer/After this function executes, this parameter contains the data that 
/// HIPAR readBuffer/was read from the instrument.
/// HIPAR numBytesRead/Returns the number of bytes actually read from the instrument and 
/// HIPAR numBytesRead/stored in the Read Buffer.
ViStatus _VI_FUNC rssgt_ReadInstrData(
  ViSession instrSession,
  ViInt32 numberofBytesToRead,
  ViChar readBuffer[],
  ViUInt32* numBytesRead
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(numberofBytesToRead);
    LogAddCharArrOut(readBuffer, &numberofBytesToRead);
    LogAddInt32Out(numBytesRead);
  }
  ViStatus error = Rs_ReadInstrData(instrSession, numberofBytesToRead, readBuffer, numBytesRead);
  LogCloseCall(error);
  return error;
}



/*****************************************************************************
 * Function: rssgt_SetAttribute<type> Functions
 * Purpose:  These functions enable the instrument driver user to set
 *           attribute values directly.  There are typesafe versions for
 *           ViInt32, ViReal64, ViString, ViBoolean, and ViSession datatypes.
 *****************************************************************************/
 /* ViInt32 */
ViStatus _VI_FUNC rssgt_SetAttributeViInt32(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViInt32 value)
{
  if (LogLoggingEnabled())
  {
    ViInt32 idx = -1;
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    Rs_GetAttributeIndex(instrSession, attributeId, &idx);
    if (idx >=0 )
    {
      RsSessionPropertiesPtr sessionProperties = Rs_ViSession(instrSession);
      LogAddSelectArg2("value", value, sessionProperties->attr[idx].constantName);
    }
    else    
    {
      LogAddInt32Arg(value);
    }
  }
  ViStatus error = Rs_SetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
  LogCloseCall(error);
  return error;
}


/* ViInt64 */
ViStatus _VI_FUNC rssgt_SetAttributeViInt64(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViInt64 value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddInt64Arg(value);
  }
  ViStatus error = Rs_SetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
  LogCloseCall(error);
  return error;
}


/* ViReal64 */
ViStatus _VI_FUNC rssgt_SetAttributeViReal64(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViReal64 value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddReal64Arg(value);
  }
  ViStatus error = Rs_SetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
  LogCloseCall(error);
  return error;
}


/* ViString */
ViStatus _VI_FUNC rssgt_SetAttributeViString(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViString value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddStringArg(value);
  }
  ViStatus error = Rs_SetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);
  LogCloseCall(error);
  return error;
}


/* ViBoolean */
ViStatus _VI_FUNC rssgt_SetAttributeViBoolean(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViBoolean value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddBoolArg(value);
  }

  ViStatus error = Rs_SetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
  LogCloseCall(error);
  return error;
}


/* ViSession */
ViStatus _VI_FUNC rssgt_SetAttributeViSession(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViSession value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddSessionArg(value);
  }

  ViStatus error = Rs_SetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
  LogCloseCall(error);
  return error;
}



/*****************************************************************************
 * Function: rssgt_GetAttribute<type> Functions
 * Purpose:  These functions enable the instrument driver user to get
 *           attribute values directly.  There are typesafe versions for
 *           ViInt32, ViReal64, ViString, ViBoolean, and ViSession attributes.
 *****************************************************************************/

 /* ViInt32 */
ViStatus _VI_FUNC rssgt_GetAttributeViInt32(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViInt32 *value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddInt32Out(value);
  }

  ViStatus error = Rs_GetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
    (unsigned int) sizeof(ViInt32), (ViInt32 *)value);
  LogCloseCall(error);
  return error;
}


/* ViInt64 */
ViStatus _VI_FUNC rssgt_GetAttributeViInt64(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViInt64 *value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddInt64Out(value);
  }

  ViStatus error = Rs_GetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
    (unsigned int) sizeof(ViInt64), (ViInt64 *)value);
  LogCloseCall(error);
  return error;
}


/* ViReal64 */
ViStatus _VI_FUNC rssgt_GetAttributeViReal64(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViReal64 *value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddReal64Out(value);
  }

  ViStatus error = Rs_GetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
    (unsigned int) sizeof(ViReal64), (ViReal64 *)value);
  LogCloseCall(error);
  return error;
}


/* ViString */
ViStatus _VI_FUNC rssgt_GetAttributeViString(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViInt32 bufSize,
  ViChar value[])
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddInt32Arg(bufSize);
    LogAddCharArrOut(value, &bufSize);
  }
  ViStatus error = Rs_GetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
    (unsigned int)bufSize, value);
  LogCloseCall(error);
  return error;
}


/* ViBoolean */
ViStatus _VI_FUNC rssgt_GetAttributeViBoolean(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViBoolean *value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddBoolOut(value);
  }
  ViStatus error = Rs_GetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
    (unsigned int) sizeof(ViBoolean), (ViBoolean *)value);
  LogCloseCall(error);
  return error;
}


/* ViSession */
ViStatus _VI_FUNC rssgt_GetAttributeViSession(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViSession *value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddSessionOut(value);
  }

  ViStatus error = Rs_GetAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
    (unsigned int) sizeof(ViSession), (ViSession *)value);
  LogCloseCall(error);
  return error;
}



/*****************************************************************************
 * Function: rssgt_SetAttributeFlags
 * Purpose:  This function sets the flags of an attribute to new values. It
 *           always sets all of the flags. If you want to change one flag,
 *           use rssgt_GetAttributeFlags to obtain the current values of all
 *           the flags, modify the bit for the flag you want to change, and
 *           then call rssgt_SetAttributeFlags.
 *           You cannot modify the value of the RS_VAL_MULTI_CHANNEL flag.
 *****************************************************************************/
ViStatus _VI_FUNC rssgt_SetAttributeFlags(ViSession instrSession, ViUInt32 attributeId,
  ViInt32 flags)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(attributeId, Attributes);
    LogAddInt32Arg(flags);
  }

  ViStatus error = Rs_SetAttributeFlags(instrSession, attributeId, flags);
  LogCloseCall(error);
  return error;
}


/*****************************************************************************
 * Function: rssgt_GetAttributeFlags
 * Purpose:  This function obtains the current values of the flags for the
 *           attribute you specify.
 *****************************************************************************/
ViStatus _VI_FUNC rssgt_GetAttributeFlags(ViSession instrSession, ViUInt32 attributeId,
  ViInt32 *flags)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddSelectArg(attributeId, Attributes);
    LogAddInt32Out(flags);
  }
  ViStatus error = Rs_GetAttributeFlags(instrSession, attributeId, flags);
  LogCloseCall(error);
  return error;
}


/*****************************************************************************
 * Function: rssgt_CheckAttribute<type> Functions
 * Purpose:  These functions enable the instrument driver user to check
 *           attribute values directly.  These functions check the value you
 *           specify even if you set the RSSGT_ATTR_RANGE_CHECK
 *           attribute to VI_FALSE.  There are typesafe versions for ViInt32,
 *           ViReal64, ViString, ViBoolean, and ViSession datatypes.
 *****************************************************************************/

 /* ViInt32 */
ViStatus _VI_FUNC rssgt_CheckAttributeViInt32(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViInt32 value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddInt32Arg(value);
  }
  ViStatus error = Rs_CheckAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
  LogCloseCall(error);
  return error;
}


/* ViReal64 */
ViStatus _VI_FUNC rssgt_CheckAttributeViReal64(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViReal64 value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddReal64Arg(value);
  }
  ViStatus error = Rs_CheckAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
  LogCloseCall(error);
  return error;
}


/* ViString */
ViStatus _VI_FUNC rssgt_CheckAttributeViString(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViString value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddStringArg(value);
  }
  ViStatus error = Rs_CheckAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);
  LogCloseCall(error);
  return error;
}


/* ViBoolean */
ViStatus _VI_FUNC rssgt_CheckAttributeViBoolean(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViBoolean value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddBoolArg(value);
  }
  ViStatus error = Rs_CheckAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
  LogCloseCall(error);
  return error;
}


/* ViSession */
ViStatus _VI_FUNC rssgt_CheckAttributeViSession(ViSession instrSession, ViString channelName,
  ViUInt32 attributeId, ViSession value)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringArg(channelName);
    LogAddSelectArg(attributeId, Attributes);
    LogAddSessionArg(value);
  }
  ViStatus error = Rs_CheckAttribute(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function resets the instrument to a known state and sends 
/// HIFN  initialization commands to the instrument.  The initialization commands 
/// HIFN  set instrument settings such as Headers Off, Short Command form, and 
/// HIFN  Data Transfer Binary to the state necessary for the operation of the 
/// HIFN  instrument driver.
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rssgt_reset(
  ViSession instrSession
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  //viCheckErr (rsPrintf (instrSession, "*RST;*CLS;*WAI\n"));
  viCheckErr(rsPrintf(instrSession, "*RST\n"));
  viCheckErr(rssgt_DefaultInstrSetup(instrSession));

  checkErr(rssgt_CheckStatus(instrSession));

Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function converts a status code returned by an instrument driver 
/// HIFN  function into a user-readable string. 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR errorCode/Pass the Status parameter that is returned from any of the instrument 
/// HIPAR errorCode/driver functions.
/// HIPAR errorMessage/Returns the user-readable message string that corresponds to the 
/// HIPAR errorMessage/status code you specify.
ViStatus _VI_FUNC rssgt_error_message(
  ViSession instrSession,
  ViStatus errorCode,
  ViChar errorMessage[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Arg(errorCode);
    LogAddStringOut(errorMessage);
  }
  ViStatus    error = VI_SUCCESS;

  static      RsStringValueTable errorTable =
  {
      RSSGT_ERROR_CODES_AND_MSGS,
      {0, NULL}
  };

  (void)Rs_LockSession(instrSession, NULL);

  /* all VISA and RS error codes are handled as well as codes in the table */
  if (errorMessage == NULL)
  {
    viCheckParm(RS_ERROR_INVALID_PARAMETER, 3, "Null address for Error Message");
  }
  else
  {
    checkErr(Rs_GetSpecificDriverStatusDesc(instrSession, errorCode, errorMessage, errorTable));
  }

Error:
  (void)(void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function returns the revision numbers of the instrument driver 
/// HIFN  and instrument firmware.
/// HIFN     
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR instrumentDriverRevision/Returns the instrument driver software revision numbers in the form of 
/// HIPAR instrumentDriverRevision/a string.
/// HIPAR firmwareRevision/Returns the instrument firmware revision numbers in the form of a 
/// HIPAR firmwareRevision/string.
ViStatus _VI_FUNC rssgt_revision_query(
  ViSession instrSession,
  ViChar instrumentDriverRevision[],
  ViChar firmwareRevision[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddStringOut(instrumentDriverRevision);
    LogAddStringOut(firmwareRevision);
  }
  ViStatus    error = VI_SUCCESS;

  checkErr(Rs_LockSession(instrSession, NULL));

  if (instrumentDriverRevision == NULL)
  {
    viCheckParm(RS_ERROR_INVALID_PARAMETER, 2, "Null address for Instrument Driver Revision");
  }
  else
  {
    checkErr(Rs_GetAttribute(instrSession, NULL, RS_ATTR_SPECIFIC_DRIVER_REVISION,
      0, 256, instrumentDriverRevision));
  }

  if (firmwareRevision == NULL)
  {
    viCheckParm(RS_ERROR_INVALID_PARAMETER, 3, "Null address for Instrument Revision");
  }
  else
  {
    checkErr(Rs_GetAttribute(instrSession, "", RS_ATTR_INSTRUMENT_FIRMWARE_REVISION,
      0, 256, firmwareRevision));
  }

  checkErr(rssgt_CheckStatus(instrSession));

Error:
  (void)(void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/// HIFN  This function reads an error code and a message from the instrument's 
/// HIFN  error queue.   
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
/// HIPAR errorCode/Returns the error code read from the instrument's error queue.
/// HIPAR errorMessage/Returns the error message string read from the instrument's error 
/// HIPAR errorMessage/message queue.
ViStatus _VI_FUNC rssgt_error_query(
  ViSession instrSession,
  ViInt32* errorCode,
  ViChar _VI_FAR errorMessage[]
)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddInt32Out(errorCode);
    LogAddStringOut(errorMessage);
  }
  ViStatus    error = VI_SUCCESS;
  ViChar      buffer[RSSGT_IO_BUFFER_SIZE] = "";

  checkErr(Rs_LockSession(instrSession, NULL));

  if (errorCode == NULL)
  {
    viCheckParm(RS_ERROR_INVALID_PARAMETER, 2, "Null address for Error Code");
  }
  else if (errorMessage == NULL)
  {
    viCheckParm(RS_ERROR_INVALID_PARAMETER, 3, "Null address for Error Message");
  }
  else
  {
    if (Rs_Simulating(instrSession) == 0)                /* call only when locked */
    {
      viCheckErr(rssgt_GetAttributeViString(instrSession, "", RSSGT_ATTR_SYSTEM_ERROR, (ViInt32)RSSGT_IO_BUFFER_SIZE, buffer));
      if (sscanf(buffer, "" _PERCD ",\"%256[^\"]", errorCode, errorMessage) != 2)
        viCheckErr(RS_ERROR_UNEXPECTED_RESPONSE);
    }
    else
    {
      /* Simulate Error Query */
      *errorCode = 0;
      strcpy(errorMessage, "No error.");
    }
  }

Error:
  (void)(void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}


/*****************************************************************************
 * Function: Rs_CheckStatusCallback
 * Purpose:  This function queries the instrument to determine if it has
 *           encountered an error. If the instrument has encountered an
 *           error, this function returns the RS_ERROR_INSTRUMENT_SPECIFIC
 *           error code. This function is called by the
 *           Rs_CheckStatus utility function. The
 *           Rs_SetAttribute and Rs_GetAttribute functions invoke this
 *           function when the optionFlags parameter includes the
 *           RS_VAL_DIRECT_USER_CALL flag.
 *
 *           The user can disable calls to this function by setting the
 *           RSCMWDAU_QUERY_INSTRUMENT_STATUS attribute to VI_FALSE.  The driver can
 *           disable the check status callback for a particular attribute by
 *           setting the RS_VAL_DONT_CHECK_STATUS flag.
 *
 *****************************************************************************/
ViStatus rssgt_CheckStatusCallback(ViSession instrSession)
{
  ViStatus    error = VI_SUCCESS;
  ViInt16     esr = 0;
  ViChar      buffer[50] = "";
  ViUInt32    ret_cnt = 0;

  memset(buffer, 0, 50);
  viCheckErr(rsWrite(instrSession, (ViBuf) "*STB?\n", 6, NULL));
  viCheckErr(rsRead(instrSession, (ViBuf)buffer, 49, &ret_cnt));
  buffer[ret_cnt] = '\0';
  esr = (ViInt16)atol(buffer);

  if ((esr & IEEE_488_2_ERROR_MASK) != 0)
  {
    viCheckErr(RS_ERROR_INSTR_SPECIFIC);
  }

Error:

  return error;
}

/*****************************************************************************
 * Function: Rs_WaitForOPCCallback
 * Purpose:  This function waits until the instrument has finished processing
 *           all pending operations. This function is called by the
 *           Rs_WaitForOPC utility function. The RS engine invokes
 *           this function in the following two cases:
 *           - Before invoking the read callback for attributes for which the
 *             RS_VAL_WAIT_FOR_OPC_BEFORE_READS flag is set.
 *           - After invoking the write callback for attributes for which the
 *             RS_VAL_WAIT_FOR_OPC_AFTER_WRITES flag is set.
 *
 *****************************************************************************/
ViStatus rssgt_WaitForOPCCallback(ViSession instrSession)
{
  ViStatus    error = VI_SUCCESS;
  int res = 0;
  error = rsQueryf(instrSession, "*OPC?\n", "%d", &res);

  return error;
}

/*****************************************************************************
 * Function: rssgt_RsClose
 * Purpose:  This function performs all of the drivers clean-up operations
 *           except for closing the RS session.  This function is called by
 *           rscmwdau_close or by an RS class driver.
 *
 *           Note:  This function must close the I/O session and set
 *           RS_ATTR_IO_SESSION to 0.
 *****************************************************************************/
ViStatus _VI_FUNC rssgt_RsClose(ViSession instrSession)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus    error = VI_SUCCESS;

  /* Do not lock here. The caller manages the lock. */

  if (instrSession != 0)
  {
    ViSession   rmSession = 0;

    checkErr(Rs_GetAttribute(instrSession, NULL, RS_ATTR_VISA_RM_SESSION, 0, (unsigned int) sizeof(rmSession), &rmSession));

    (void)rsPrintf(instrSession, "&GTL\n");

    (void)rsClose(instrSession);
    (void)rsClose(rmSession);
  }
Error:
  LogCloseCall(error);
  return error;
}


/// HIFN  This function performs the following operations:
/// HIFN     
/// HIFN     - Closes the instrument I/O session.
/// HIFN     
/// HIFN     - Destroys the instrument driver session and all of its attributes.
/// HIFN     
/// HIFN     - Deallocates any memory resources the driver uses.
/// HIFN     
/// HIFN     Notes:
/// HIFN     
/// HIFN     (1) You must unlock the session before calling rssgt_close.
/// HIFN     
/// HIFN     (2) After calling rssgt_close, you cannot use the instrument 
/// HIFN  driver again until you call rssgt_init or 
/// HIFN  rssgt_InitWithOptions.
/// HIFN     
/// HIFN     
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rssgt_init or 
/// HIPAR instrSession/rssgt_InitWithOptions function.  The handle identifies a 
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rssgt_close(ViSession instrSession)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus rssgs_status = rsClose(instrSession);

  (void)Rs_Dispose(instrSession);

  LogCloseCall(rssgs_status);
  return rssgs_status;
}



/// HIFN This function enables and disables special PCI express features.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR enableFast/Activates/deactivates the fast mode.
/// HIPAR enableAsynchronous/Activates/deactivates the asynchronous mode.
ViStatus _VI_FUNC rssgt_UseFastSettings(ViSession instrSession, ViBoolean enableFast, ViBoolean enableAsynchronous)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolArg(enableFast);
    LogAddBoolArg(enableAsynchronous);
  }
  ViStatus rssgs_status = rsSetAttribute(instrSession, SGS_ATTR_FAST, enableFast);

  if ((rssgs_status == VI_SUCCESS) && (enableFast != VI_FALSE))
  {
    ViUInt32 fastEnabled = VI_FALSE;
    rssgs_status = rsGetAttribute(instrSession, SGS_ATTR_FAST, &fastEnabled);
    if ((rssgs_status == VI_SUCCESS) && (fastEnabled != VI_FALSE))
      rssgs_status = rsSetAttribute(instrSession, SGS_ATTR_FASTASYNCHRONOUS, enableAsynchronous);
  }
  LogCloseCall(rssgs_status);
  return rssgs_status;
}


/// HIFN This function returns the states of parameters PCIe enable and PCIe asynchronous mode enable.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR enableFast/Returns the state of the fast mode.
/// HIPAR enableAsynchronous/Returns the state of the asynchronous mode.
ViStatus _VI_FUNC rssgt_GetFastSettings(ViSession instrSession, ViPBoolean bEnableFast, ViPBoolean bEnableAsynchronous)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddBoolOut(bEnableFast);
    LogAddBoolOut(bEnableAsynchronous);
  }
  ViInt32 enableFast = 0, enableAsynchronous = 0;
  ViStatus rssgs_status = rsGetAttribute(instrSession, SGS_ATTR_FAST, &enableFast);

  if ((rssgs_status == VI_SUCCESS) && (enableFast))
  {
    //ViBoolean fastEnabled  = VI_FALSE;
    rssgs_status = rsGetAttribute(instrSession, SGS_ATTR_FASTASYNCHRONOUS, &enableAsynchronous);
  }
  *bEnableFast = enableFast != 0;
  *bEnableAsynchronous = enableAsynchronous != 0;
  LogCloseCall(rssgs_status);
  return rssgs_status;
}


/// HIFN In fast mode this function sends an empty command to instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgt_FastSync(ViSession instrSession)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
  }
  ViStatus rssgs_status = rsFastSync(instrSession);
  if (rssgs_status == VI_ERROR_NSUP_OPER)
  {
    int res = 0;
    rssgs_status = rsQueryf(instrSession, "*OPC?\n", "%d", &res);
  }
  LogCloseCall(rssgs_status);
  return rssgs_status;
}


/// HIFN Find devices on the PCIe bus and return their serial numbers
/// HIRET Error code
/// HIPAR vendorId/PCI id of the device's manufacturer
/// HIPAR deviceId/PCI id of the device
/// HIPAR serialNumbers/array with returned serial numbers. Allocated by caller
/// HIPAR serialNumberCount/allocated size of array. On return: number of found devices
ViStatus _VI_FUNC rssgt_enumerate_pcie_devices(ViInt32 vendorId,
  ViInt32 deviceId,
  ViInt32 serialNumbers[],
  ViInt32 *serialNumbersCount)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddInt32Arg(vendorId);
    LogAddInt32Arg(deviceId);
    LogAddInt32ArrOut(serialNumbers, serialNumbersCount);
    LogAddInt32Out(serialNumbersCount);
  }
  ViStatus status = viEnumerate_pcie_devices(vendorId, deviceId, serialNumbers, serialNumbersCount);
  LogCloseCall(status);
  return status;
}


/// HIFN Find fast socket devices and return their types, serial numbers
/// HIFN and IP addresses. Call function with infoCount=0/info = NULL,
/// HIFN allocate buffer and call again with valid buffers.  
/// HIRET Returns the status code of this operation.
/// HIPAR info/Array of pointers to string buffers. Each buffer must be at least 64 bytes long.
/// HIPAR info/Each string returned by the function is a comma-separated list of instrument type, 
/// HIPAR info/serial number and IP address e.g. SGT100A,100027,169.254.45.172. 
/// HIPAR infoCount/Number of buffers in array info. On return: the number of found devices.
ViStatus _VI_FUNC rssgt_enumerate_socket_devices(ViPString *info, ViPInt32 infoCount)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddStringArrOut(info, infoCount);
    LogAddInt32Out(infoCount);
  }
  ViStatus status = viEnumerate_socket_devices(info, infoCount);
  LogCloseCall(status);
  return status;
}


//=====================================================================================
/// HIFN This function configures frequency and level.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR frequency/RF frequency.
/// HIPAR level/RF level.
ViStatus _VI_FUNC rssgt_SetFrequencyAndLevel(ViSession instrSession,
  ViReal64  frequency, ViReal64  level)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(frequency);
    LogAddReal64Arg(level);
  }
  ViUInt32 supported = 0;
  ViStatus error = VI_SUCCESS;
  ViChar rep_cap[5] = "";

  (void)viFastSettingSupported(instrSession, IdSgtFreqFreq, &supported);

  if (supported != 0)
  {

    ParaItem ParaList[2] = {
      {IdSgtFreqFreq, PARATYPE_REAL64, 0},
      {IdSgtPowPow,  PARATYPE_REAL64, 0} };
    ParaList[0].pData = (unsigned char*)&frequency;
    ParaList[1].pData = (unsigned char*)&level;

    error = viFastSetMultiple(instrSession, ParaList, 2);
  }
  else
  {
    checkErr(Rs_LockSession(instrSession, NULL));

    checkErr(rssgt_pathRepCap(instrSession, 0, rep_cap));

    viCheckParm(rssgt_SetAttributeViReal64(instrSession, rep_cap,
      RSSGT_ATTR_RF_FREQUENCY,
      frequency), 2, "Frequency");

    viCheckParm(rssgt_SetAttributeViReal64(instrSession, rep_cap,
      RSSGT_ATTR_RF_AMPLITUDE,
      level), 3, "Level");
  }
Error:
  (void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;

}



/// HIFN  This function sets the power servoing target power to target pow and starts 
/// HIFN  power servoing. 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR targetpow/The desired power at the power sensor input
/// HIPAR result/Result of the operation(0: OK, 1: fail) 
ViStatus _VI_FUNC rssgt_PowerServoingSet(ViSession instrSession, ViReal64 targetpow, ViPInt32 result)
{
  if (LogLoggingEnabled())
  {
    LogAddCall(__func__);
    LogAddSessionArg(instrSession);
    LogAddReal64Arg(targetpow);
    LogAddInt32Out(result);
  }
  ViUInt32 supported = 0;
  ViStatus error = VI_SUCCESS;

  (void)viFastSettingSupported(instrSession, IdSgtPowServExec, &supported);

  if (supported)
  {
    viFastSetReal64(instrSession, IdSgtPowServTarget, targetpow);
    viFastSetInt32(instrSession, IdSgtPowServExec, 0);

    //unsigned int DummyInt = 1;
    //ParaItem ParaList[2] = {
    //  {IdSgtPowServTarget, PARATYPE_REAL64, 0},
    //  {IdSgtPowServExec,  PARATYPE_INT32, 0} };
    //ParaList[0].pData = (unsigned char*)&targetpow;
    //ParaList[1].pData = (unsigned char*)&DummyInt;
    //viFastSetMultiple(instrSession, ParaList, 2);

    error = viFastQueryInt32(instrSession, IdSgtPowServSet, result);
  }
  else
  {
    checkErr(Rs_LockSession(instrSession, NULL));
    viCheckErr(rsQueryf(instrSession, "POW:SERV:SET? %f\n", "%ld", targetpow, result));
  }

Error:
  (void)(void)Rs_UnlockSession(instrSession, NULL);
  LogCloseCall(error);
  return error;
}

/*===========================================================================*/
/* Function: rssgt_openLog                                                   */
/* Purpose:  start recording function calls                                  */
/* historydepth: max number of function calls in the record                  */
/* enablewrap  : if not set the logger stops recording after historydepth    */
/*               calls have been recorded else it overwrites old log entries */
/*               if the number of records exceeds historydepth               */
/* stoponerror : stop recording if the return status of a call has the error */
/*               flag(_VI_ERROR) set.                                        */
/*===========================================================================*/

void _VI_FUNC rssgt_openLog(ViInt32 historydepth, ViBoolean enablewrap, ViBoolean stoponerror)
{
  LogOpenLog("rssgt", historydepth, enablewrap, stoponerror);
  LogSetDecodeErrorCallback((DecodeErrorCallback_f)get_error_name_and_message);
}

/*===========================================================================*/
/* Function: rssgt_closeLog                                                   */
/* Purpose:  stop recording function calls                                  */
/*===========================================================================*/
void _VI_FUNC rssgt_closeLog(void)
{
  LogCloseLog();
}

/*===========================================================================*/
/* Function: rssgt_saveLog                                                   */
/* Purpose:  start recording function calls                                  */
/* filename: name of the xml file into which to write the log                */
/*===========================================================================*/
ViStatus _VI_FUNC rssgt_saveLog(const ViString filename)
{
  return LogSaveLog(filename);
}

/*===========================================================================*/
/* Function: rssgt_setMaxLogCallDepth                                        */
/* Purpose:  maximum call stack depth to record                              */
/* depth:    call stack depth (default = 1)                                  */
/*===========================================================================*/
void _VI_FUNC rssgt_setMaxLogCallDepth(ViUInt32 depth)
{
  LogSetMaxCallDepth(depth);
}

/*===========================================================================*/
/* Function: rssgt_setWrapStartIndex                                         */
/* Purpose: set start of wrap section                                        */
/* wrapstartindex: number of calls at the beginning of the record list which */
/*                 will not be overwritten when the recording wraps around   */
/*                 (default=20). Preserves any initialization sequence.      */
/*===========================================================================*/
void _VI_FUNC rssgt_setLogWrapStartIndex(ViUInt32 wrapstartindex)
{
  LogSetWrapStartIndex(wrapstartindex);
}
